//
// fgraph.cc
//  
// Made by vicen
// Login   <vicen@localhost.localdomain>
// 
// Started on  Wed Apr 19 20:52:08 2006 vicen
// Started on  Wed Apr 19 20:52:08 2006 vicen
//

// VERBOSITY RULE:
// use cerr for m_verbosiy < 2
// for m_verbosity > 1, use both cout & cerr

#include "fgraph.hh"

/////////////////////////////////
// COMMENT DEFINES TO REMOVE THEM

typedef Real TYPE_MATRIX;
//typedef long double TYPE_MATRIX;

template<typename Graph, typename DiGraph>
struct output_visitor : public planar_face_traversal_visitor
{
	int m_verbosity;
	ExtVertexD target_prev, source_prev;
	ExtEdgeD edge_prev, first_edge, edge_face;
	bool is_first_edge;
	int nface;
	Graph g;
	DiGraph dg;
	map<ExtEdgeD, vector<int> > map_efaces;        // to store edges in Graph not in DiGraph
	vector<vector<ExtEdgeD> > clockwise_efaces;    // to store the clockwise traversals

	output_visitor(const Graph &_g, const DiGraph &_dg, const int &verb) {
		g = _g;
		dg = _dg;
		nface = 0;
		m_verbosity = verb;
	}

  void begin_face() {
		if (m_verbosity > 2) cout << "Face " << nface << "\t";
		is_first_edge = true;
		vector<ExtEdgeD> v;
		clockwise_efaces.push_back(v);
	}
 
	template <typename Edge> 
	void next_edge(Edge e) {

		if (m_verbosity>2) std::cout << e << " ";

		// check if the edge belongs to the spanning tree
		if ( ( edge(source(e, g), target(e, g), dg).second != true ) &&
				 ( edge(target(e, g), source(e, g), dg).second != true ) )
		{
			edge_face = e;
			if (map_efaces.find(e) == map_efaces.end()) {
				vector<int> v(1); v[0] = nface;
				map_efaces[e] = v;
			}
			else map_efaces[e].push_back(nface);
		}
	
		// add the edge to the traversal
		clockwise_efaces[nface].push_back(e);
	}

	void end_face() {
		if (m_verbosity>2) 	cout << " End Face" << endl; // The edge not found in the spanning tree was " << edge_face << endl;
		nface++;
	}
};
	

Real FGraph::getBPError() 
{
    m_error_bp = Real(-1.);
    for (VectorVertexIt itv = m_vars_labels.begin(); itv != m_vars_labels.end(); itv++) {
      for (uint val = 0; val < (*g)[*itv].p.size(); val++) {
        Real err = Abs( (*g)[*itv].bel[val] - (*g)[*itv].p[val] );
        if (err > m_error_bp)
          m_error_bp = err;
      }    
    }
    return m_error_bp;
}

template<typename Graph, typename EdgeIt, typename VertexIt>
void g2Dot(Graph g, string(filename))
{
  std::ofstream fout(filename.c_str());
  fout << "graph G {" << std::endl;
  fout << "graph[size=\"100,100\"];" << std::endl;

  // var nodes 
  fout << "node[shape=circle,width=0.25,height=0.25,fixedsize=true];" << std::endl;
	VertexIt vi, ve;
	for (tie(vi, ve) = vertices(g); vi != ve; vi++) {
    fout << "\t" << *vi;
    fout << " [style=\"setlinewidth(1)\",label=" << *vi << "];" << std::endl;
  }

	// edges 
  EdgeIt ei, ei_end;
  for (tie(ei, ei_end) = edges(g); ei != ei_end; ei++) {
		fout << "\t" << source(*ei, g) << " -- " << target(*ei, g);
		fout << "[style=\"setlinewidth(1)\",fontsize=15,label=\"";
		fout << setprecision(3) << "\"];" << std::endl;
  }
  fout << "}" << std::endl;
  fout.close();
}

void FGraph::forney2Dot(const char *filename, const ForneyGraph &forney)
{

  // create dot file
  std::ofstream fout(filename);
  fout << "graph G {" << std::endl;
  fout << "graph[size=\"100,100\"];" << std::endl;

  // var nodes 
  fout << "node[shape=circle,width=0.25,height=0.25,fixedsize=true];" << std::endl;
  std::pair<FVertexIt, FVertexIt> it = vertices(forney); 
  while (it.first != it.second) {
    fout << "\t" << *it.first;
    if (forney[*it.first].delta) {
      fout << " [style=\"setlinewidth(1)\",label=\"=" << *it.first << "\"];" << endl;
			FOutEdgeIt o_begin, o_end;
//			for (tie(o_begin, o_end) = out_edges(*it.first, forney); o_begin != o_end; o_begin++)
//				cout << "\t" << forney[*o_begin].id << "\t";
//			cout << endl;
    }
		else {
      fout << " [style=\"setlinewidth(1)\",label=" << *it.first << "];" << std::endl;
		}
    *it.first++;
  }

	// rest
  FEdgeIt ei, ei_end;
  for (tie(ei, ei_end) = edges((forney)); ei != ei_end; ei++) {
    FEdgeD e = *ei;

		// Factor to Var (m message)
		fout << "\t" << source(e, forney) << " -- " << target(e, forney);
		fout << "[style=\"setlinewidth(1)\",fontsize=15,label=" << forney[e].id << "];" << std::endl;
  }	
  fout << "}" << std::endl;
  fout.close();
}

void FGraph::ext2Dot(const ExtGraph &gext, const char *filename)
{

  // create dot file
  std::ofstream fout(filename);
  fout << "graph G {" << std::endl;
  fout << "graph[size=\"100,100\"];" << std::endl;

  // var nodes 
  fout << "node[shape=circle,width=0.25,height=0.25,fixedsize=true];" << std::endl;
  std::pair<FVertexIt, FVertexIt> it = vertices(gext); 
  while (it.first != it.second) {
    fout << "\t" << gext[*it.first].id;
    fout << " [style=\"setlinewidth(1)\",label=" << gext[*it.first].id << "];" << std::endl;
    *it.first++;
  }

	// rest
  ExtEdgeIt ei, ei_end;
  for (tie(ei, ei_end) = edges(gext); ei != ei_end; ei++) {
    Edge e = *ei;

		// edges 
		fout << "\t" << gext[source(e, gext)].id << " -- " << gext[target(e, gext)].id;
		fout << "[style=\"setlinewidth(1)\",fontsize=15,label=\"";

		fout << setprecision(3) << (double)(gext)[*ei].mu << "\"];" << std::endl;
  }
  fout << "}" << std::endl;
  fout.close();
}

void FGraph::diext2Dot(const ExtDiGraph &digraph, const char *filename)
{
  // create dot file
  std::ofstream fout(filename);
  fout << "digraph G {" << std::endl;
  fout << "digraph[size=\"100,100\"];" << std::endl;

  // var nodes 
  fout << "node[shape=circle,width=0.25,height=0.25,fixedsize=true];" << std::endl;
  std::pair<FVertexIt, FVertexIt> it = vertices(digraph); 
  while (it.first != it.second) {
    fout << "\t" << *it.first;
    fout << " [style=\"setlinewidth(1)\",label=" << digraph[*it.first].id << "];" << std::endl;
    *it.first++;
  }

	// rest
  ExtDiEdgeIt ei, ei_end;
  for (tie(ei, ei_end) = edges(digraph); ei != ei_end; ei++) {
    ExtDiEdgeD e = *ei;

		// edges 
		fout << "\t" << source(e, digraph) << " -> " << target(e, digraph);
		fout << "[style=\"setlinewidth(1)\",fontsize=15,label=\"";
		fout << setprecision(3) << (double)(digraph)[*ei].mu << "\"];" << std::endl;

  }
  fout << "}" << std::endl;
  fout.close();
}

void FGraph::graph2Dot(const char *filename)
{
  VectorVertex vars, factors;

  // create dot file
  std::ofstream fout(filename);
//  fout << "digraph G {" << std::endl;
//  fout << "digraph[size=\"100,100\"];" << std::endl;
  fout << "graph G {" << std::endl;
  fout << "graph[size=\"100,100\"];" << std::endl;

  // var nodes 
  fout << "node[shape=circle,width=0.25,height=0.25,fixedsize=true];" << std::endl;
  std::pair<VertexIt, VertexIt> it_var = vertices(*g); 
  while (it_var.first != it_var.second) {
    if ((*g)[*it_var.first].type == VAR) {
      fout << "\t" << *it_var.first;
      fout << "[style=\"setlinewidth(1)\",label=" << *it_var.first << "];" << std::endl;
//      fout << "\t" << (*g)[*it_var.first].label;
//      fout << "[style=\"setlinewidth(1)\",label=v" << (*g)[*it_var.first].label << "];" << std::endl;
    }
    it_var.first++;
  }

  // factor nodes
  fout << "node[shape=box,width=0.25,height=0.25,fixedsize=true];" << std::endl;
  std::pair<VertexIt, VertexIt>  sit = vertices(*g); 
  while (sit.first != sit.second) {
    if ((*g)[*sit.first].type == FACTOR) {
      fout << "\t" << *sit.first;
//cout << "\t" << *sit.first << " psi = ";
//for (uint ii=0; ii<(*g)[*sit.first].psi.size(); ii++) cout << (*g)[*sit.first].psi[ii] << " ";
//cout << endl;

      fout << "[style=\"setlinewidth(1)\",label=" << *sit.first << "];" << std::endl;
//      fout << "\tp" << (*g)[*sit.first].label;
//      fout << "[style=\"setlinewidth(1)\",label=f" << char(97+(*g)[*sit.first].label) << "];" << std::endl;
    }
    sit.first++;
  }

  // rest
  EdgeIt ei, ei_end;
  for (tie(ei, ei_end) = edges((*g)); ei != ei_end; ei++) {
    Edge e = *ei;
    if (((*g))[source(e, (*g))].type == FACTOR) {
      fout << "\t" << source(e, (*g)) << " -- " << target(e, (*g));
      fout << "[style=\"setlinewidth(1)\",fontsize=15,fontcolor=red,label=\"" << "\"];" << std::endl;

      // Factor to Var (m message)
//      VProb m = getBPm(e, false);
//      VProb n = getBPn(e, false);
//      fout << "\tp" << ((*g))[source(e, (*g))].label << " -> " << ((*g))[target(e, (*g))].label;
//      fout << "[style=\"setlinewidth(1)\",fontsize=15,fontcolor=red,label=\"" <<  std::setprecision(3) << m[0]-m[1] << "\"];" << std::endl;
//
//      // Var to Factor (n message)
//      fout << "\t" << ((*g))[target(e, (*g))].label << " -> p" << ((*g))[source(e, (*g))].label;
//      fout << "[style=\"setlinewidth(1)\",fontsize=15,fontcolor=blue,label=\"" <<  std::setprecision(3) << n[0]-n[1] << "\"];" << std::endl;

     }
    else std::cerr << "????" << std::endl;
  }
  fout << "}" << std::endl;
  fout.close();
}

FGraph::FGraph(std::string filename, TBPVariant bp_variant, int precision, int ran, int verbosity)
{
	m_verbosity = verbosity;
	
	// check the format by extension
	// .fg  : libDAI
	// .pot : bbcore
	// .uai : UAI'08
	m_max_factor_value = -1.;
	
	// inits
	m_bp_variant = bp_variant;
	m_nedges = 0;
	m_nvars = 0;
	m_nfactors = 0;

	// random inits
	m_random_init = ran;									// are bp msgs init randomly?
	m_r = gsl_rng_alloc(gsl_rng_mrg);			// allocate the random num gen
	m_gsl_seed = ran;											// keep the seed in a member var	
	gsl_rng_set(m_r, m_gsl_seed);					// set the rand seed
//	gsl_rng_env_setup();			// Alternative form to setup gsl seed using env var GSL_RNG_SEED
 
	m_filename = filename;
	m_logZ_bp = -1.;
	m_logZ_exact = -1.;
	m_ZPfaffian = 0.;
	m_ZPfaffian1 = 0.;
	m_negpot = false;
	
	// create the Graph
	g = new BGraph();
	
	// read from file
	std::ifstream file_in(filename.c_str());
	if (!file_in) throw "Factor Graph file " + filename + " not found!";
	
	// temporal set for the variables
	std::set<PVertex> svars;
	
	// start reading from stream
	int labelf = 0;
	char aline[100];
	int posdot = filename.find_last_of('.')+1;
	std::string extension(filename.substr(posdot, filename.length()-posdot));
	if (extension == "fg") { 
		if (m_verbosity > 1) std::cout << "READING in libDAI format" << std::endl;
	  do {
	    file_in.getline(aline, 100);
			if (file_in.eof()) throw "eof found!";
		} while ( (aline[0] == '#') || (atoi(aline) == 0) );
	
		if (m_verbosity > 1) std::cout << aline << " factors"; std::cout.flush();
		m_nfactors = atoi(aline);
	}
	else if (extension == "pot") {
		if (m_verbosity > 1) std::cout << "READING in bbcore format" << std::endl;
	}
	else if (extension == "uai") {
		if (m_verbosity > 1) std::cout << "READING in UAI'08 format" << std::endl;
	}
	
	// temporal vector for the variables of a factor
	std::vector<std::vector<PVertex> > vfactorvars;
	
	// read the factor variables
	int nvals; uint nvars;
	std::string s;
	if (extension == "fg")
	  file_in >> nvars;
	if (extension == "uai") {

		// lets read everything here...
		string network_type;
		getline(file_in, network_type);
		if (network_type != "MARKOV")
			throw "Only MARKOV networks supported";
//cout << "MARKOV" << endl;
		int tmp;
		getline(file_in, s);
		tmp = atoi(s.c_str()); // number of vars
//cout << tmp << endl;
		getline(file_in, s);
//cout << "cards " << s << endl;
		stringstream str(s);
		string card;
		while (str >> card) {
//cout << card << " ";
			if (atoi(card.c_str()) != 2) throw "Only binary variables supported";
		}
//cout << " fi card " << endl;
		getline(file_in, s);
		tmp = atoi(s.c_str()); 	// number of factors
//cout << tmp << " factors" << endl;
		for (int f = 0; f < tmp; f++) {
			getline(file_in, s);
			stringstream str(s);
			string fvar;
			std::vector<PVertex> vvar;
			str >> fvar;
			if (atoi(fvar.c_str()) > 2)
				cerr << fvar << "-wise interactions" << endl;
			while (str >> fvar) {
	      PVertex var(VAR, atoi(fvar.c_str()), 2);
	      svars.insert(var);
				vvar.push_back(var);
			}
			vfactorvars.push_back(vvar);			
		}
	}

	while (!file_in.eof()) {
	
	//     std::cout << "\t" << nvars << " vars" << std::endl;
	
	  // add the variables to the set
	  int totalsize = 1;
		std::vector<PVertex> vvar;
		if (extension == "fg") { 
			std::vector<int> vlabels(nvars), vsizes(nvars);
	    for (uint v=0; v<nvars; v++) file_in >> vlabels[v];
	    for (uint v=0; v<nvars; v++) file_in >> vsizes[v];
	
	    for (uint v=0; v<nvars; v++) {
	      PVertex var(VAR, vlabels[v], vsizes[v]);
	      svars.insert(var);
				vvar.push_back(var);
	      totalsize *= vsizes[v];
	    }
	    vfactorvars.push_back(vvar);
		} else if (extension == "pot") {
	
			getline(file_in, s);
			std::cerr << s << std::endl;
			std::cerr << "eof? " << file_in.eof() << std::endl;
			if ( file_in.eof() ) break;
			std::stringstream str(s);
	 		int label_var;
			nvars = 0;
			while (str >> label_var) {
	      PVertex var(VAR, label_var, 2);
				totalsize *= 2;
	      svars.insert(var);
				vvar.push_back(var);
				nvars++;
			}
	    vfactorvars.push_back(vvar);
		}
	
	  // read the factor values
	  Real p;
	  VProb prob(totalsize, 0.);
		if (extension == "fg") { 
	    file_in >> nvals;
			for (int v=0, i; v<nvals; v++) {
	      file_in >> i >> p;
	      prob[i] = p;
	      if (Abs(p) > m_max_factor_value)
	        m_max_factor_value = Abs(p);
				if (p < Real(0.)) {
					m_negpot = true;
					std::cerr << "NEG POT" << std::endl;
			  }
	    }
		}
		else if (extension == "pot") {
	
			if (!getline(file_in, s)) break;
			std::cerr << s << std::endl;
			std::stringstream str(s);
			int i=0;
			while (str >> p) {
				prob[i] = p;
				i++;
			}
			std::cerr << setprecision(14) << prob << std::endl;
			m_nfactors++;
		}
		else if (extension == "uai") {
	
//cout << "getting line... " << endl;
			if (!getline(file_in, s)) break;
			std::stringstream str(s);
//cout << "values are " << s << endl;			
			int nv;
			str >> nv;
			prob.resize(nv, 0.);
			int i=0;
			while (str >> p) {
				if (i<nv) {
					prob[i] = p;
				}
				i++;
			}
			m_nfactors++;
		}

	  // add the factor to the graph
	  BGraph::vertex_descriptor v = add_vertex(*g);
	  (*g)[v].type = FACTOR;
	  (*g)[v].label = labelf;
	  (*g)[v].size = prob.size();
	  (*g)[v].psi = prob;
	
		labelf++;
	
	  // next factor
		if (extension == "fg") file_in >> nvars;
	}
	
	// the total number of vars
	m_nvars = svars.size();
	
	// add the variables to the graph using the svars set
	// they are inserted in the labeling order
	for (std::set<PVertex>::iterator it = svars.begin(); it != svars.end(); it++) {
	  BGraph::vertex_descriptor v = add_vertex(*g);
	  (*g)[v].type = VAR;
	  (*g)[v].label = (*it).label;
	  (*g)[v].size = (*it).bel.size();   
	  (*g)[v].bel = (*it).bel;           // the belief is initialized to 1/nvals (see global.hh)
  //m_mapvar[(*it).label] = v;
	}
	
	// add the edges between factors and variables using vfactorvars
	std::pair<VertexIt, VertexIt> vpf = vertices(*g), vpv;
	
	// vpf iterates over factors and vpv iterates over variables
	for (uint f = 0; (f < m_nfactors); vpf.first++, f++) {
	
	  // Indices of the factor f
	  PIndex idx;
	
	  // use vfactorvars to get the variables of the factor
	  int nvar = 0;
	  for (std::vector<PVertex>::iterator i = vfactorvars[f].begin(); i != vfactorvars[f].end(); i++) {
	
	    // set the vpv iterator to the begining of the variable nodes
	    vpv = vertices(*g); vpv.first += m_nfactors;
	
	    // find the descriptor of the corresponding variable node
	    while ( ((*g)[*vpv.first].label != (*i).label) && (vpv.first != vpv.second) ) vpv.first++;

	    // add the edge
	    addEdge(*vpf.first, *vpv.first, *g, (*g)[*vpv.first].size, nvar);
	
	    // add the variable to the index
	    idx.addVar((*g)[*vpv.first].size);
	
	    m_nedges++;
	    nvar++;
	  }
	
	  // add the index to the factor
	  (*g)[*vpf.first].index = idx;
	}
	
	buildFactors();        // vector of factors
	buildVarsPos();        // vector of vars and map of positions
	buildVarsSizes();      // vector of vars sorted by size
	
	std::pair<VertexIt, VertexIt> iit = vertices(*g);
	m_max_degree = 0;
	while (iit.first != iit.second) {
		if (degree(*iit.first, *g) == 0)
			std::cerr << *iit.first << " has no neighbors!" << std::endl;
		if (m_max_degree < degree(*iit.first, *g))
			m_max_degree = degree(*iit.first, *g);
		iit.first++;
	}		  
	
	m_exact = false;
	m_error_bp = -1;
	
	if (m_verbosity > 1) cout << "fgraph readed" << std::endl;
	
	g2Dot<BGraph, EdgeIt, VertexIt>(*g, string("fgraph.dot"));
	graph2Dot("fgraph.dot");
}
	struct coord_t
	{
		std::size_t x;
		std::size_t y;
	};


void FGraph::planar2Matlab(ExtGraph &gext0, const ExtDiGraph &digraph, int verb, char *filename)
{
	ExtGraph gext = gext0;

  // Initialize the interior edge index
	property_map<ExtGraph, edge_index_t>::type e_index = get(edge_index, gext);
	graph_traits<ExtGraph>::edges_size_type edge_count = 0;
	graph_traits<ExtGraph>::edge_iterator ei, ei_end;
	for(tie(ei, ei_end) = edges(gext); ei != ei_end; ++ei) 
	  put(e_index, *ei, edge_count++);
					
  //Test for planarity; compute the planar embedding as a side-effect
  typedef std::vector< graph_traits<ExtGraph>::edge_descriptor > vec_t;
  std::vector<vec_t> embedding(num_vertices(gext));
  if (boyer_myrvold_planarity_test(boyer_myrvold_params::graph = gext,
				boyer_myrvold_params::embedding = &embedding[0])) {
		if (verb > 2) cout << "[Input graph is planar]" << endl;
	}
  else
		throw "Input graph is not planar";

	make_biconnected_planar(gext, &embedding[0]);

  // Re-initialize the edge index, since we just added a few edges
  edge_count = 0;
  for(tie(ei, ei_end) = edges(gext); ei != ei_end; ++ei)
    put(e_index, *ei, edge_count++);

  //Test for planarity again; compute the planar embedding as a side-effect
  if (boyer_myrvold_planarity_test(boyer_myrvold_params::graph = gext,
                                   boyer_myrvold_params::embedding = &embedding[0])) {
		if (verb > 2) cout << "After calling make_biconnected, the graph is still planar" << endl;
	}
	else
	  throw "After calling make_biconnected, the graph is not planar";

	make_maximal_planar(gext, &embedding[0]);

  // Re-initialize the edge index, since we just added a few edges
  edge_count = 0;
  for(tie(ei, ei_end) = edges(gext); ei != ei_end; ++ei) 
    put(e_index, *ei, edge_count++);

  // Define the new storage type for the planar embedding
  typedef std::vector< std::vector< graph_traits<ExtGraph>::edge_descriptor > > embedding_storage_t;
  typedef boost::iterator_property_map< embedding_storage_t::iterator, property_map<ExtGraph, vertex_index_t>::type > embedding_t;

	// Create the planar embedding
	embedding_storage_t embedding_storage(num_vertices(gext));
	embedding_t new_embedding(embedding_storage.begin(), get(vertex_index,gext));

	//Test for planarity again; compute a new planar embedding as a side-effect
  if (boyer_myrvold_planarity_test(boyer_myrvold_params::graph = gext,
                                   boyer_myrvold_params::embedding = new_embedding)){
		if (verb > 2) cout << "After calling make_biconnected, the graph is still planar" << endl;
	}
	else
	  throw "After calling make_biconnected, the graph is not planar";

	// Find a canonical ordering in the auxiliary graph
  std::vector<graph_traits<ExtGraph>::vertex_descriptor> ordering;
  planar_canonical_ordering(gext, new_embedding, std::back_inserter(ordering));
	if (verb > 2) cout << ordering.size() << " elements" << std::endl;

	// draw the graph
	if (verb > 2) cout << "\t[canonical ordering done] " << ordering.size() << endl;

  typedef vector<coord_t> straight_line_drawing_storage_t;
  typedef boost::iterator_property_map<straight_line_drawing_storage_t::iterator, property_map<ExtGraph, vertex_index_t>::type> straight_line_drawing_t;

  straight_line_drawing_storage_t straight_line_drawing_storage(num_vertices(gext));
  straight_line_drawing_t straight_line_drawing(straight_line_drawing_storage.begin(),get(vertex_index,gext));

	if (verb > 2) cout << "\t[drawing of the embedding maximal graph]" << endl;

	// Compute the straight line drawing
	chrobak_payne_straight_line_drawing(gext, 
                                      new_embedding, 
                                      ordering.begin(),
                                      ordering.end(),
                                      straight_line_drawing);

	// now each edge is a line between coord(get(straight_line_drawing,source)) and coord(get(straight_line_drawing,target))
	ofstream fout(filename);
  for(tie(ei, ei_end) = edges(gext); ei != ei_end; ++ei) {
	  ExtVertexD s = source(*ei, gext), t = target(*ei, gext);
		ExtEdgeC e = edge(s, t, gext0);
		ExtDiEdgeC e1 = edge(s,t,digraph);
		ExtDiEdgeC e2 = edge(t,s,digraph);

		bool found = e1.second;
		if (e2.second) {
			ExtVertexD aux = t;
			t = s;
			s = aux;
			found = true;
		}
		if (found) {
		// discard additional edges before drawing the graph
//		if (e.second) {
			coord_t coord_s(get(straight_line_drawing, s)), coord_t(get(straight_line_drawing, t));
			fout << gext[s].id << "\t" << gext[t].id << "\t" << coord_s.x << "\t" << coord_t.x << "\t" << coord_s.y << "\t" << coord_t.y << endl;
		}
	}
	fout.close();
	if (verb > 2) cout << "\t[drawing done. Matlab file : " << filename << "]" << endl;
}

using namespace std;

void FGraph::toForneyStyle()
{
	ForneyGraph forney;
  FEdgeIt fei, fee;

	// foreach factor node
	if (m_verbosity) cerr << "converting to Forney" << endl;

	vector<vector<FVertexD> > delta_nodes(num_vertices(*g));
	VertexIt vi, ve;
  for (tie(vi, ve) = vertices(*g); vi != ve; vi++) {
    Vertex i = *vi;
    if ((*g)[i].type == VAR) {

			// create degree(i)-2 identity-nodes
			int deg = degree(i, *g)-3;
			FVertexD delta = add_vertex(forney);
			forney[delta] = FVertex( VProb(), PIndex(), delta, true );
			delta_nodes[i].push_back(delta);
			while (deg > 0) {
				FVertexD delta_old = delta;
				delta = add_vertex(forney);
				forney[delta].delta = true;
				FEdgeC ed = add_edge(delta, delta_old, forney);
				forney[ed.first] = FEdge( i, 0 );
				forney[ed.first].bel = (*g)[i].bel;		// if BP result -> add belief
				forney[ed.first].msg = (*g)[i].bel;
				ed = add_edge(delta_old, delta, forney);
				forney[ed.first] = FEdge( i, 0 );
				forney[ed.first].bel = (*g)[i].bel;		// if BP result -> add belief
				forney[ed.first].msg = (*g)[i].bel;
				delta_nodes[i].push_back(delta);
				deg--;
			}
		}
	}

	for (tie(vi, ve) = vertices(*g); vi != ve; vi++) {
    Vertex i = *vi;
    if ((*g)[i].type == FACTOR) {

			// create "real" factors
	    FVertexD v = add_vertex(forney);
			forney[v] = FVertex( (*g)[i].psi, (*g)[i].index, (int)v );

			// if BP result -> add the belief
			forney[v].bel = (*g)[i].bel;

	    OutEdgeIt j_begin, j_end;
      for (tie(j_begin, j_end) = out_edges(i, *g); j_begin != j_end; j_begin++) {
				Vertex j = target(*j_begin, *g);
				uint ndeltas = delta_nodes[j].size()-1;
				FVertexD delta = delta_nodes[j][ndeltas];
				if (in_degree(delta, forney) == 3) {
					if (m_verbosity > 2) cout << "delta " << delta_nodes[j][ndeltas] << " is replaced by ";
					delta_nodes[j].pop_back();
					ndeltas--;
					delta = delta_nodes[j][ndeltas];
					if (m_verbosity > 2) cout << delta << endl;
				}
				FEdgeC ed = add_edge(v, delta, forney);
				forney[ed.first] = FEdge( j, (*g)[*j_begin].pos_index );  // this pos_index is overwrittenafterwards
				forney[ed.first].bel = (*g)[j].bel;		// if BP result -> add belief
				forney[ed.first].msg = (*g)[j].bel;
				ed = add_edge(delta, v, forney);
				forney[ed.first] = FEdge( j, (*g)[*j_begin].pos_index );  // this pos_index is overwrittenafterwards
				forney[ed.first].bel = (*g)[j].bel;		// if BP result -> add belief
				forney[ed.first].msg = (*g)[j].bel;
			}
		}
	}

	forney2Dot("forney_not_core.dot", forney);

	// fill delta nodes (phi, indexes, and pos of out_links)
	// pos_index always corresponds to the direction of the edge (target)
	if (m_verbosity > 2) cout << "Setting Phi values of delta nodes" << endl;
  pair<FVertexIt, FVertexIt> iitf = vertices(forney);
  while (iitf.first != iitf.second) {
    Vertex i = *iitf.first;
		if (forney[i].delta) {
			FInEdgeIt i_begin, i_end;
			int pos = 0;
			VProb bel_var(2);
			for (tie(i_begin, i_end) = in_edges(i, forney); i_begin != i_end; i_begin++) {
				forney[*i_begin].pos_index = pos;
				forney[i].index.addVar(2);
				bel_var = forney[*i_begin].bel;
				pos++;
			}
			if (m_verbosity > 2) { cout << "\t" << forney[*iitf.first].index.getNVals() << endl; }
		  forney[i].psi = VProb(forney[i].index.getNVals(), 0.);
			forney[i].psi[0] = 1.;
			forney[i].psi[forney[i].psi.size()-1] = 1.;
			// if BP result -> add the belief (zeros in all values which are not the same)
			forney[i].bel = VProb(forney[i].index.getNVals(), 0.);
			forney[i].bel[0] = bel_var[0];
			forney[i].bel[forney[i].bel.size()-1] = bel_var[1];
		}
		iitf.first++;
	}

	// set logZ of BP
	m_logZ_bp_forney = m_logZ_bp;
	
	// set the graph member of the class
	m_forney.clear();
	m_forney = forney;
}

void FGraph::doPfaffianSeries( size_t all_comb)
{	
	VertexIt vi, ve;
  FEdgeIt fei, fee;
	char buff[400];

	double init_ext_clock = ttoc(), final_ext_clock;

	// first check that the factor graph is planar
	if (boyer_myrvold_planarity_test(*g)) {
		if (m_verbosity > 2) cout << "factor graph is planar" << endl;
	}
	else throw "factor graph is not planar!";
	if (boyer_myrvold_planarity_test(m_forney)) {
		if (m_verbosity > 2) cout << "forney not core is planar" << endl;
	}
	else throw "forney not core is not planar!";

	// obtain 2-core
  // remove nodes with degree one 
  if (m_verbosity > 2) cout << "preprocessing graph ...";
  uint nvertices = num_vertices(m_forney);
  uint rem = 0, remp;
  do {
    remp = rem;
    pair<FVertexIt, FVertexIt> it = vertices(m_forney);
    while ( (it.first != it.second) && (rem == remp) ) {

			// we compare with 2 (remember forney is bidirectional)
      if (degree(*it.first, m_forney) == 2) {
        clear_vertex(*it.first, m_forney);

        // remove_vertex invalidates iterator
       	remove_vertex(*it.first, m_forney);
       	rem++;
      }
      it.first++;
    }
  } while (rem - remp);
	uint N = num_vertices(m_forney);
	if (m_verbosity > 2) cout << nvertices << " -> " << N << endl;
	if (N == 1) {
		m_ZPfaffian = Real(1.);
		m_ZPfaffian1 = Real(1.);
		m_cpu_pfaffian = 0;
		m_cpu_pfaffian1 = 0;
		return;
	}

	// rebuild indices and beliefs of the factors
	for (tie(vi, ve) = vertices(m_forney); vi != ve; vi++) 
		rebuildForneyIndex(*vi);
	forney2Dot("forney_core.dot", m_forney);

	if (boyer_myrvold_planarity_test(m_forney)) {
		if (m_verbosity > 2) cout << "forney is planar" << endl;
	}
	else throw "forney is not planar!";


	// find triplets in the forney core
	vector<Real> vtriplets;							// vector with mus of triplets
	vector<FVertexD> vtriplets_forney;		// vector of triplet nodes of the forney graph
	map<FEdgeD, int> map_trip;					  // map to get the triplet index given the edge
	uint total_triplets = 0, ntriplets = 0;
  for (tie(vi, ve) = vertices(m_forney); vi != ve; vi++) {
    FVertexD i = *vi;
		if (in_degree(i, m_forney) == 3) {

			// a triplet: compute the mu term and push into the triplets vector
			total_triplets++;
			VProb bela(m_forney[i].bel);
			PIndex *idx = &(m_forney)[i].index;
			FInEdgeIt i_begin, i_end;
			for (tie(i_begin, i_end) = in_edges(i, m_forney); i_begin != i_end; i_begin++) {
				int pos = m_forney[*i_begin].pos_index;
				Real m(m_forney[*i_begin].bel[0] - m_forney[*i_begin].bel[1]);
				Real sqrtm( Sqrt( Real(1.) - m*m ) );
				
				// compute the term of the triplet
				for (uint itf = 0; itf < bela.size(); itf++) {
					bela[itf] *= (idx->get(itf, pos) == 0) ? (Real(+1.) - m) : (Real(-1.) - m);
					bela[itf] /= sqrtm;
				}
			}
			// consider the triplet only if the term is different than zero
			Real mu_triplet(bela.normalizeL1());

			bool ok = true;
			if (mu_triplet == 0) ok = false;
			if (ok) {
				for (tie(i_begin, i_end) = in_edges(i, m_forney); i_begin != i_end; i_begin++) 
					map_trip[*i_begin] = ntriplets; 
//				vtriplets.push_back( (double)mu_triplet );
				vtriplets.push_back( mu_triplet );
				vtriplets_forney.push_back(i);
				ntriplets++;
			}
			else if (m_verbosity > 1) cout << "zero triplet " << mu_triplet << endl;
		}
	}

	// get the maximum value of triplet pairs and create vector of triplet pairs
	//double max_triplet = 0.;
	Real max_triplet = Real(0.);
	if (ntriplets != vtriplets.size()) throw "ntriplets does not fit";

	vector<TripletPair> vtriplet_pairs;
	for (uint t1 = 0; t1 < ntriplets; t1++) {
		for (uint t2 = t1+1; t2 < ntriplets; t2++) {
			Real mu_triplet_pair = vtriplets[t1]*vtriplets[t2];
//			double mu_triplet_pair = vtriplets[t1]*vtriplets[t2];
			vtriplet_pairs.push_back( TripletPair(t1, t2, mu_triplet_pair) ); 
//			if (fabs(mu_triplet_pair) > max_triplet)
			if (Abs(mu_triplet_pair) > max_triplet)
				max_triplet = mu_triplet_pair;
		}
	}
	// sort triplet pairs by absolute value
	sort( vtriplet_pairs.begin(), vtriplet_pairs.end(), LessAbsTripletPair() );

	// compute all distances between pairs of triplets
	Matrix<int> mat_dist(N, N, 0);
//	getDistancesForney(vtriplets_forney, mat_dist);
	vector<double> dist_triplets;

	if (m_verbosity) cerr << ntriplets << " of " << total_triplets << " triplets are not zero" << endl;

	// obtain the extended graph
	ExtGraph gext;
	vector<vector<FEdgeD> > vars_forney(N);							// vector with vars of G
	vector<vector<ExtVertexD> > vars_extended(N);				// vector with pairs or triples of Gext
	vector<vector<int> > vtriplets_nodes(ntriplets);		// vector with 3 ids for each triplet
	for (tie(fei, fee) = edges(m_forney); fei != fee; ++fei) {

		// for each edge in the original graph, create two nodes in the extended one
		FVertexD s = source(*fei, m_forney), t = target(*fei, m_forney);
		ExtVertexD node_src = add_vertex(gext), node_tgt = add_vertex(gext);
		gext[node_src] = ExtVertex(node_src);
		gext[node_tgt] = ExtVertex(node_tgt);
		vars_extended[s].push_back(node_src);
		vars_extended[t].push_back(node_tgt);

		// link the nodes
		ExtEdgeC ed = add_edge(node_src, node_tgt, gext);
		(gext)[ed.first].mu = Real(1.);

		vars_forney[t].push_back(*fei);

		// add the target node in Gext to the vector of triplets
		if ( map_trip.find(*fei) != map_trip.end() ) 
			vtriplets_nodes[map_trip[*fei]].push_back( gext[node_tgt].id );

		++fei;
		
		// check also the opposite edge
		// add the source node in Gext to the vector of triplets
		vars_forney[s].push_back(*fei);
		if ( map_trip.find(*fei) != map_trip.end() ) 
			vtriplets_nodes[map_trip[*fei]].push_back( gext[node_src].id );
	}

	// create triplets and pairs in the extended and set the mus
	for (uint u = 0; u < vars_extended.size(); u++) {
		if ( (vars_extended[u].size() != 2) && (vars_extended[u].size() != 3) ) {
			cerr << vars_extended[u].size() << endl;
			throw "Error in degrees";
		}

		// the belief of the factor
		for (uint v1 = 0; v1 < vars_extended[u].size(); v1++) {
			for (uint v2 = v1+1; v2 < vars_extended[u].size(); v2++) {
				if (m_verbosity > 1) {
					cout << vars_extended[u][v1] << " ~ " << vars_extended[u][v2] << "\t";
					cout << m_forney[vars_forney[u][v1]].id << " ~ " << m_forney[vars_forney[u][v2]].id << "\t";
				}
				ExtEdgeC ed = add_edge(vars_extended[u][v1], vars_extended[u][v2], gext);

				// compute the mu of the pair combination
				// numerator
				VProb B = m_forney[u].bel;
				PIndex *idx = &(m_forney)[u].index;
				int pos_v1 = m_forney[vars_forney[u][v1]].pos_index;
				int pos_v2 = m_forney[vars_forney[u][v2]].pos_index;
				Real m_v1 = m_forney[vars_forney[u][v1]].bel[0] - m_forney[vars_forney[u][v1]].bel[1];
				Real m_v2 = m_forney[vars_forney[u][v2]].bel[0] - m_forney[vars_forney[u][v2]].bel[1];
				for (uint itf = 0; itf < B.size(); itf++) {
	        B[itf] *= (idx->get(itf, pos_v1) == 0) ? (Real(+1.) - m_v1) : (Real(-1.) - m_v1);
	        B[itf] *= (idx->get(itf, pos_v2) == 0) ? (Real(+1.) - m_v2) : (Real(-1.) - m_v2);
				}
				Real mu = B.normalizeL1();

				// denominator
				Real denom = Sqrt((Real(1.) - m_v1*m_v1)*(Real(1.) - m_v2*m_v2));
				(gext)[ed.first].mu = mu/denom;
				if (m_verbosity > 1) cout << (gext)[ed.first].mu << endl;
			}
		}
	}

	final_ext_clock = ttoc() - init_ext_clock;
	if (m_verbosity>1) cout << endl << "Extended: Spent " << final_ext_clock << " scnds). " << endl;
	m_cpu_ext = final_ext_clock;

	// the full extended graph is constructed
	if (m_verbosity) cerr << num_vertices(gext) << " vertices in Gext ~ O(" << pow(num_vertices(gext),3.) << ")" << endl;
	if (m_verbosity > 1) {
		ext2Dot(gext, "extended.dot");
	}

	sprintf(buff, "triplets_contrib_%s.m", m_filename.c_str());
	m_fcorr.open(buff);

	// m_gext is the copy kept for getting higher order corrections
	m_gext = gext;
	m_contrib_pfaffian.clear();

	if (all_comb) {

			// all terms
			ulong total_comb = 0, total_loops = 0;
			for (int m = 0; m <= (int)ntriplets; m += 2) {
				vector<int> vcomb;
				allCombinations(-1, m, ntriplets, vcomb, total_comb, total_loops, vtriplets, vtriplets_nodes, true);
			}
			cerr << total_comb << " total combinations. " << total_loops << " total loops " << endl;
			m_fcorr.close();
	} else {
			// only Z_2
			ulong total_comb = 0, total_loops = 0;
			int m = 0;
			vector<int> vcomb;
			allCombinations(-1, m, ntriplets, vcomb, total_comb, total_loops, vtriplets, vtriplets_nodes, false);
			m_fcorr.close();
	}
}

void FGraph::outputRes() 
{
	// old code from loopgraph.cc
  m_error_bp = getBPError();

	// lets output the cumulated error of the pfaffian here ...
	if (m_verbosity > 1) {
		char buff[400];
		sprintf(buff, "cum_err_pfaffian_%s.m", m_filename.c_str());
		ofstream fcum(buff);
		Real acum(0.), logZ_pfaffian = m_logZ_bp_forney;
		fcum << setprecision(15) << Abs(m_logZ_exact - logZ_pfaffian) << endl;
		for (uint p = 0; p < m_contrib_pfaffian.size(); p++) {
			acum += m_contrib_pfaffian[p].first;
			logZ_pfaffian = m_logZ_bp_forney + Log(acum);
			fcum << setprecision(15) << Abs(m_logZ_exact - logZ_pfaffian) << endl;
		}
		fcum.close();
	}

	// -------------------------

	if ( m_verbosity ) {
		cerr << scientific << setprecision(14) << "m_ZPfaffian " << m_ZPfaffian << endl;
		if (Abs( m_ZPfaffian ) > Real(0)) {
		  cerr << "\tsqrt(det(P)) = " << m_ZPfaffian << endl;
		}

		cerr << "\tlogZ_exact   = " << m_logZ_exact << " " << expl(m_logZ_exact) << endl;
		cerr << "\tlogZ_bp      = " << m_logZ_bp << " " << expl(m_logZ_bp) << endl;

		if (Abs( m_ZPfaffian ) >  Real(0)) {
			cerr << "\tlogZ_Pfaff   = " << getLogZPfaffian() << " " << Exp( getLogZPfaffian() ) << endl;
		}
		cerr << endl;
    cerr << "\terr_Z_bp     = " << Abs(m_logZ_bp - m_logZ_exact)/m_logZ_exact << endl;

		if (Abs( m_ZPfaffian ) > Real(0)) {
			cerr << "\terr_Z0_Pfaff = " << Abs(m_logZ_exact - getLogZPfaffian1())/m_logZ_exact << endl;
			cerr << "\terr_Z_Pfaff  = " << Abs(m_logZ_exact - getLogZPfaffian())/m_logZ_exact << endl;
		}
    cerr << endl;
	}


}

void FGraph::allCombinations(	const int &i,
															const int &m,
															const int &n,
															vector<int> &vcomb,
															ulong &ncomb,
															ulong &total_loops,
															const vector<Real> &vtriplets,
															const vector<vector<int> > &vtriplets_nodes,
															bool sign_correction)
{
	if (m == 0) {
		// call Pfaffian
		for (uint x=0; x<vcomb.size(); x++)
			cout << vcomb[x] << " ";
		cout << endl;
		pfaffianTerm(ncomb, vcomb, vtriplets, vtriplets_nodes, total_loops, sign_correction);
		ncomb++;
	}
	else {
		for (int j = i+1; j < n; j++) {
			vcomb.push_back(j);
			allCombinations(j, m-1, n, vcomb, ncomb, total_loops, vtriplets, vtriplets_nodes, sign_correction);
			vcomb.pop_back();
		}
	}
}

void FGraph::pfaffianTerm(const int &pt,
													const vector<int> &vremoved,
													const vector<Real> &vtriplets,
													const vector<vector<int> > &vtriplets_nodes,
													ulong &total_loops,
													bool sign_correction)
{
	VertexIt vi, ve;
  FEdgeIt fei, fee;
	char buff[400];
	uint total_vertices = num_vertices(m_gext);

	// start pfaffian cpu-clock
	double init_pfaffian_clock = ttoc(), final_pfaffian_clock;

	//double mu_triplets = 1.;
	Real mu_triplets = Real(1.);
	ExtGraph gext = m_gext;

	// for each triplet
	for (uint ri = 0; ri < vremoved.size(); ri++) {
		if (m_verbosity > 1) cout << "\tremoving triplet " << vremoved[ri] << " : " << vtriplets[vremoved[ri]] << "( ";

		// multiply by weights of the triplet
		mu_triplets *= vtriplets[vremoved[ri]];

		// for each of the 3 nodes in Gext which belong to the triplet
		for (uint ni = 0; ni < vtriplets_nodes[vremoved[ri]].size(); ni++) {

			// find it in Gext
			bool found = false;
			for (tie(vi, ve) = vertices(gext); (vi != ve) && (!found); vi++)
				found = (gext[*vi].id == vtriplets_nodes[vremoved[ri]][ni]);
			if (!found)	throw "Error removing nodes";
			vi--;
			if (m_verbosity > 1) cout << vtriplets_nodes[vremoved[ri]][ni] << " "; //=" << gext[*vi].id << ")";
      clear_vertex(*vi, gext);
     	remove_vertex(*vi, gext);
		}
		if (m_verbosity > 1) cout << ")" << endl;
	}

	if (m_verbosity > 2) {
		if (vremoved.size()) {
			sprintf(buff, "extended_subgraph_notbiconnected%d.dot", pt);
			ext2Dot(gext, buff);
		}
	}

	// make it biconnected again to set the orientation
	makeBiconnected(gext);
	if (m_verbosity > 2) {
		sprintf(buff, "extended_subgraph_%d.dot", pt); 
		ext2Dot(gext, buff);
	}

	Real pf(1.);
	ulong nloops = 0;

	// check null graph
	if ((vremoved.size()*3) < total_vertices) {
		
		property_map<ExtGraph, edge_index_t>::type e_index = get(edge_index, gext);
		graph_traits<ExtGraph>::edges_size_type	edge_count = 0;
	  graph_traits<ExtGraph>::edge_iterator eei, eee;
		for(tie(eei, eee) = edges(gext); eei != eee; ++eei)
			put(e_index, *eei, edge_count++);

		typedef std::vector< std::vector< graph_traits<ExtGraph>::edge_descriptor > > embedding_storage_t;
		typedef boost::iterator_property_map< embedding_storage_t::iterator, property_map<ExtGraph, vertex_index_t>::type> embedding_t;
		embedding_storage_t embedding_storage(num_vertices(gext));
		embedding_t embedding2(embedding_storage.begin(), get(vertex_index,gext));
		if (boyer_myrvold_planarity_test(boyer_myrvold_params::graph = gext, boyer_myrvold_params::embedding = embedding2)) {
			if (m_verbosity > 2) cout << "gext is planar" << endl;
		}
		else throw "gext is not planar!";

		// Now we set the orientation. First obtain a DFS spanning tree
		ExtDiGraph digraph;
		map<ExtDiVertexD, bool> vmarked;
	  for (tie(vi, ve) = vertices(gext); vi != ve; vi++) {
			ExtDiVertexD node = add_vertex(digraph);
			digraph[node] = ExtDiVertex(gext[*vi].id);
			vmarked[*vi] = false;
		}
		makeDirectedDFS(*vertices(gext).first, gext, digraph, vmarked);
		
		if (m_verbosity > 2) {
			cout << "marked " << endl;
			for (map<ExtDiVertexD, bool>::iterator ii = vmarked.begin(); ii != vmarked.end(); ii++)
				cout << (*ii).first << " " << (*ii).second << endl;
			cout << "undirected " << num_vertices(gext) << " vertices and " << num_edges(gext) << endl;
			cout << "  directed " << num_vertices(digraph) << " vertices and " << num_edges(digraph) << endl;
			graph_traits<ExtGraph>::edge_iterator eei, eee;
			for (tie(eei, eee) = edges(gext); eei != eee; eei++) {
				if ( edge(source(*eei, gext), target(*eei, gext), digraph).second == true ) cout << "same or "<< *eei << endl;
				else if ( edge(target(*eei, gext), source(*eei, gext), digraph).second == true ) cout << "oppos or "<< *eei << endl;
				else cout << "not exists" << endl;
			}
		}
		
		// set the orientation traversing faces clockwise 
		// this traversal obtains the edges in gext not in digraph (e_vis.map_efaces)
		// and also builds the list of vertices for each face in clockwise order (e_vis.clock_faces)
		if (m_verbosity>2) cout << std::endl << "Edges on the faces: " << endl;
	  output_visitor<ExtGraph, ExtDiGraph> e_vis(gext, digraph, m_verbosity);
	  planar_face_traversal(gext, embedding2, e_vis);

		// build the Faces Tree (note that we count that the external face corresponds to the first one,
		// so we can root the faces tree on it
		map<BaEdgeD, ExtEdgeD> map_faces_edges;
		BasicGraph gfaces(e_vis.nface-1);
		for (map<ExtEdgeD, vector<int> >::iterator iif = e_vis.map_efaces.begin(); iif != e_vis.map_efaces.end(); iif++) {
			vector<int> faces = (*iif).second;
			if (faces.size() != 2) throw "An edge was found which separates more than 2 faces!!!!";
			if (faces[1] < faces[0]) throw "Error assigning faces...";
	
			// Link the two faces in gfaces
			BaEdgeC ed = add_edge(faces[0], faces[1], gfaces);
	
			// store the edge of gext (key of map_efaces) corresponding to the new edge in gfaces
			map_faces_edges[ed.first] = (*iif).first;
		}
		
		// traverse the faces tree in post-order and set the orientation of the remaining edges using e_vis.clock_faces
		map<BaVertexD, bool> vmarked_faces;
		BaVertexIt vif, vef;
	  for (tie(vif, vef) = vertices(gfaces); vif != vef; vif++) vmarked[*vif] = false;
		BaOutEdgeIt ebb, ebe;
		BaVertexD root = *vertices(gfaces).first;
		for (tie(ebb, ebe) = out_edges(root, gfaces); ebb != ebe; ebb++)
			orientPostOrder(target(*ebb, gfaces), root, gfaces, e_vis.clockwise_efaces, map_faces_edges, gext, digraph, vmarked_faces);
		
		// output the orientation
		if (m_verbosity > 2) {
			cout << "FINAL ORIENTATION:" << endl;
			ExtDiEdgeIt edi, ede;
			for (tie(edi, ede) = edges(digraph); edi != ede; edi++) 
				cout << "(" << digraph[source(*edi, digraph)].id << "," << digraph[target(*edi, digraph)].id << ")" << endl;
		}

		if (m_verbosity > 2) {
			sprintf(buff, "diextended_%d.dot", pt);
			diext2Dot(digraph, buff);
			sprintf(buff, "faces_graph_%d.dot", pt);
			g2Dot<BasicGraph, BaEdgeIt, BaVertexIt>(gfaces, string(buff));

			// NOW PLOT THE PLANAR ORIENTED GRAPH (AFTER THE ORIENTATION HAS BEEN MADE)
			sprintf(buff, "planar_embedding_%d.m", pt);
			planar2Matlab(gext, digraph, m_verbosity, buff);
		}

		// build matrix of mu's
		Matrix<TYPE_MATRIX> *mat_pfaffian = new Matrix<TYPE_MATRIX>(num_vertices(digraph), num_vertices(digraph), TYPE_MATRIX(0));
		ExtDiEdgeIt edi, ede;
		for (tie(edi, ede) = edges(digraph); edi != ede; edi++) {
			int xx = source(*edi, digraph), yy = target(*edi, digraph);

			bool ok = true;
			if (!ok) {
				if (m_verbosity > 2) cout << "rounding " << digraph[*edi].mu << " to zero" << endl;	// round the value (not necessary?)
			}
			else {
				if (m_verbosity > 2) {
					cout << "\t\tM(" << xx+1 << "," << yy+1 << ") = mu(" << digraph[source(*edi,digraph)].id;
					cout << "," << digraph[target(*edi,digraph)].id << ") = " << digraph[*edi].mu << endl;
				}
				(*mat_pfaffian)[xx][yy] = digraph[*edi].mu;
				(*mat_pfaffian)[yy][xx] = -digraph[*edi].mu;
			}
		}

		// compute the 1st. pfaffian O(n^3)
		// pfaffian already destroys the matrix map_pfaffian
		pf = pfaffian<TYPE_MATRIX>(1, num_vertices(digraph)/2, mat_pfaffian);
cerr << "1st. Pfaffian" << endl;

		if (sign_correction) {
			// we only compute the sign correction if all combinations are required.
			Matrix<TYPE_MATRIX> *mat_aux = new Matrix<TYPE_MATRIX>(num_vertices(digraph), num_vertices(digraph), TYPE_MATRIX(0));
			for (tie(edi, ede) = edges(digraph); edi != ede; edi++) {
				int xx = source(*edi, digraph), yy = target(*edi, digraph);
	
				bool ok = true;
				if (!ok) {
					if (m_verbosity > 2) cout << "rounding " << digraph[*edi].mu << " to zero" << endl;	// round the value (not necessary?)
				}
				else {
					if (m_verbosity > 2) {
						cout << "\t\tM(" << xx+1 << "," << yy+1 << ") = mu(" << digraph[source(*edi,digraph)].id;
						cout << "," << digraph[target(*edi,digraph)].id << ") = " << digraph[*edi].mu << endl;
					}
					(*mat_aux)[xx][yy] = +1.;
					(*mat_aux)[yy][xx] = -1.;
				}
			}
	
			// compute the 2nd.. pfaffian O(n^3)
			// pfaffian already destroys the matrix map_pfaffian
	
			Real pf_aux = pfaffian<TYPE_MATRIX>(1, num_vertices(digraph)/2, mat_aux);
	
			if (m_verbosity > 1) {
				cout << "PFAFFIAN AUX IS " << pf_aux << endl;
				cout << "The " << pt << "th Pfaffian is \t" << pf << endl;
			}
			if (pf_aux < Real(0)) pf = -pf;
		}
		else {

			// we assume positive correction
			if (m_verbosity > 1) cout << "Auxiliary PFAFFIAN is not computed (sign might be wrong!)" << endl;
			pf = Abs(pf);
		}

		// end cpu-clock
		if (!pt) {
			final_pfaffian_clock = ttoc() - init_pfaffian_clock;
			if (m_verbosity > 2) cout << endl << "Pfaffian: Spent " << final_pfaffian_clock << " scnds." << endl;
			m_cpu_pfaffian = final_pfaffian_clock;
		}

		if (m_verbosity > 1) { cout << pt << "\t" << vremoved.size() << "\t" << pf << endl; }

	} // endif not null graph

	Real the_term(mu_triplets * pf);
	m_ZPfaffian = m_ZPfaffian + the_term;
	m_contrib_pfaffian.push_back( PfaffianTerm(the_term, nloops ) );
	if (vremoved.size()) {
		m_fcorr << scientific << setprecision(14);

		m_fcorr << "[ ";
		for (uint nr = 0; nr < vremoved.size(); nr++)
			m_fcorr << vremoved[nr] << " ";
		m_fcorr << "]" << "\t";

		m_fcorr << pt																		<< "\t";		// number of pfaffian
		m_fcorr << mu_triplets													<< "\t";    // mu triplet term
		m_fcorr << the_term                             << "\t";		// value of full correction of Z_{\psi}
		m_fcorr << nloops																<< endl;		// number of generalized loops included in the Pfaffian
	}
	else {
		m_ZPfaffian1 = m_ZPfaffian;
	}

	total_loops += nloops;

	if (m_verbosity > 1) cout << scientific << setprecision(14) << " correction is " << the_term << endl;
	if (m_verbosity > 2) cout << "-------------------------------------------------------" << endl;

	// get next pair of triplets
	if (vremoved.size() == 0) {
		if (the_term < Real(0.)) {
			cerr << "z0 IS NEGATIVE!!!!!!!!!" << endl;
		}
		if (m_verbosity) {
			cerr << pt << " Z_0: " << nloops << " loops. correction = " << the_term << endl;
		}
	}
	else {
		if (m_verbosity) {
			cerr << pt;
			cerr << "\t" << the_term << endl;
		}
		gext.clear();
	}
}

void FGraph::makeBiconnected(ExtGraph &gext) {

	// now make it biconnected again
	make_connected(gext);

	property_map<ExtGraph, edge_index_t>::type e_index = get(edge_index, gext);
	graph_traits<ExtGraph>::edges_size_type edge_count = 0;
	graph_traits<ExtGraph>::edge_iterator ei, ei_end;
  for(tie(ei, ei_end) = edges(gext); ei != ei_end; ++ei)
    put(e_index, *ei, edge_count++);

	//Test for planarity; compute the planar embedding as a side-effect
  typedef std::vector< graph_traits<ExtGraph>::edge_descriptor > vec_t;
  std::vector<vec_t> embedding1(num_vertices(gext));
  boyer_myrvold_planarity_test(boyer_myrvold_params::graph = gext, boyer_myrvold_params::embedding = &embedding1[0]);

	typedef std::vector< graph_traits<ExtGraph>::edges_size_type > component_storage_t;
  typedef iterator_property_map < component_storage_t::iterator, property_map<ExtGraph, edge_index_t>::type> component_map_t;
	
	component_storage_t component_storage(num_edges(gext));
  component_map_t component(component_storage.begin(), get(edge_index, gext));
	  
	if (m_verbosity > 2)
		cerr << "\tBefore make_biconnected_planar: " << biconnected_components(gext, component) << " biconnected components" << endl;

	make_biconnected_planar(gext, &embedding1[0]);

  // Re-initialize the edge index, since we just added a few edges
	edge_count = 0;
  for(tie(ei, ei_end) = edges(gext); ei != ei_end; ++ei)
    put(e_index, *ei, edge_count++);

  // Re-size the storage for the biconnected components, since we just added a few edges
  component_storage.resize(num_edges(gext));
  component = component_map_t(component_storage.begin(), get(edge_index,gext));

	if (m_verbosity > 2) {
		cout << "\tAfter make_biconnected_planar : " << biconnected_components(gext, component);
		cout << " biconnected components" << endl;
	}
}


void FGraph::rebuildForneyIndex(FVertexD a)
{
	set<int> vpos;
	PIndex idx = m_forney[a].index;
	if (in_degree(a, m_forney) != (uint)idx.getNVars()) {
//cerr << "size index = " << forney[a].index.getNVars() << " degree = " << degree(a, forney) << endl;
//cerr << "index = " << endl << forney[a].index;

		PIndex new_index;
	  FInEdgeIt ab, ae;
		int new_pos = 0;
    for (tie(ab, ae) = in_edges(a, m_forney); ab != ae; ab++) {
//cerr << "in var " << forney[*ab].id << " " << forney[*ab].pos_index << " -> ";
			new_index.addVar(2);
			vpos.insert(m_forney[*ab].pos_index);
			m_forney[*ab].pos_index = new_pos;
//cerr << forney[*ab].pos_index << endl;
			new_pos++;
		}
		VProb bel(m_forney[a].bel), new_bel(1ul << vpos.size(), 0.);
//cerr << "old belief : " << bel << endl;

		// traverse the old belief by traversing the old index
		// marginalize over the removed vars
		uint p = 0;
		for (vector<vector<uint> >::iterator it = idx.begin(); it != idx.end(); it++) {
			uint bpos = 0, j = 0;
//cerr << "inst : " << endl;					
			for (uint i = 0; i < (*it).size(); i++) {
				bool deleted = (vpos.find(i) == vpos.end());
//cerr << "\t" << i << " is " << (*it)[i] << " deleted? " << deleted << endl; 
				if ((*it)[i] && !deleted) {
					bpos += (1ul << j);
				}
				if (!deleted) j++;
			}
//cerr << "-------> \t" << bpos << endl;
			new_bel[bpos] += bel[p];
			p++;
		}
//cerr << "new index " << endl << new_index << endl;
//cerr << "new belief : " << new_bel << endl;
		m_forney[a].index = new_index;
		m_forney[a].bel = new_bel;

		// we are done because in a Forney graph there the are no pos_factor
	}
}

void FGraph::orientPostOrder(
		BaVertexD u,
		BaVertexD p,
		const BasicGraph &gfaces,
		const vector<vector<ExtEdgeD> > &clockwise_efaces,
		const map<BaEdgeD, ExtEdgeD> &map_faces_edges,
		const ExtGraph &gext,
		ExtDiGraph &digraph,
		map<BaVertexD, bool> &mark)
{
	mark[u] = true;
	BaOutEdgeIt ebb, ebe;
  for (tie(ebb, ebe) = out_edges(u, gfaces); ebb != ebe; ebb++) {
		if ( target(*ebb, gfaces) != p ) {
			if (mark[target(*ebb, gfaces)]) throw "Error! gfaces is NOT a tree!!!";
			orientPostOrder(target(*ebb, gfaces), u, gfaces, clockwise_efaces, map_faces_edges, gext, digraph, mark);
		}
	}

	// Set the orientation of the edge between the faces p and u
	bool odd = false;
	if (m_verbosity > 2) cout << "Setting orientation in Face " << u << endl;
	ExtDiVertexD esource, etarget;
	Real mu;
	int nclock = 0;

	for (uint fedge = 0; fedge < clockwise_efaces[u].size(); fedge++) {

		ExtEdgeD e = clockwise_efaces[u][fedge];																		// edge of the face traversal
		ExtEdgeD ne = clockwise_efaces[u][(fedge+1)%clockwise_efaces[u].size()];		// next edge of the face traversal
		if (m_verbosity>2) cout << "Edge (" << gext[source(e,gext)].id << "," << gext[target(e,gext)].id << ")\t";

		// try to find the edge of gext in digraph in any of the two possible directions
		ExtDiEdgeC edi = edge(source(e, gext), target(e, gext), digraph);
		if (!edi.second) edi = edge(target(e, gext), source(e, gext), digraph);			
		if (!edi.second) {

			// edge not found in digraph (determine clockwise orientation and set mu)
			mu = gext[e].mu;

			if (m_verbosity>2) cout << "not found ";
			if (target(e, gext) == source(ne, gext) ||
					target(e, gext) == target(ne, gext) )  {
	
				esource = source(e, gext);
				etarget = target(e, gext);
				if (m_verbosity>2) cout << "the order is " << gext[esource].id << "~" << gext[etarget].id;
			}
			else {
				etarget = source(e, gext);
				esource = target(e, gext);
				if (m_verbosity>2) cout << "the order is " << gext[esource].id << "~" << gext[etarget].id;
			}
		}
		else {

			// edge found in digraph
			if (m_verbosity>2) cout << "found     ";
			if (target(edi.first, digraph) == source(ne, digraph) ||
					target(edi.first, digraph) == target(ne, digraph) )  {
				odd = !odd;
				nclock++;
				if (m_verbosity>2) cout << "(" << digraph[source(edi.first, digraph)].id << "," << digraph[target(edi.first, digraph)].id << ") clockwise";
			}
			else
				if (m_verbosity>2) cout << "(" << digraph[source(edi.first, digraph)].id << "," << digraph[target(edi.first, digraph)].id << ") counter-clockwise";
		}
		if (m_verbosity>2) cout << endl;
	}

	// if !odd add the edge clockwise oriented
	ExtDiEdgeC ne = (!odd) ? add_edge(esource, etarget, digraph) : add_edge(etarget, esource, digraph);
	digraph[ne.first].mu = mu;
	if (m_verbosity>2) cout << endl << "\t nclock " << nclock << " odd " << odd << endl;
}

void FGraph::makeDirectedDFS(ExtVertexD v, ExtGraph &g, ExtDiGraph &digraph, map<ExtVertexD, bool> &mark)
{
	mark[v] = true;
	ExtOutEdgeIt vi, ve;
  for (tie(vi, ve) = out_edges(v, g); vi != ve; vi++) {
		ExtVertexD t(target(*vi, g));
		if ( !mark[t] ) {
			ExtDiEdgeC ne = add_edge(v, t, digraph);
//			cout << ne.first << " " << v << "~" << t << " " << ne.second << endl;
			digraph[ne.first].mu = g[*vi].mu;
			makeDirectedDFS(t, g, digraph, mark);
		}
	}
}

std::ostream& operator<< (std::ostream& os, FGraph& fg)
{
  // create new file with the tree
  os << fg.m_nfactors << std::endl << std::endl;
  std::pair<VertexIt, VertexIt> iit = vertices(*fg.g);
  while (iit.first != iit.second) {
    Vertex a = *iit.first;
    if ((*fg.g)[a].type == FACTOR) {
      os << degree(a, *fg.g) << std::endl;
      OutEdgeIt j_begin, j_end;
      for (tie(j_begin, j_end) = out_edges(a, (*fg.g)); j_begin != j_end; j_begin++)
				os << (*fg.g)[target(*j_begin, *fg.g)].label << " ";
			os << std::endl;
      for (tie(j_begin, j_end) = out_edges(a, (*fg.g)); j_begin != j_end; j_begin++)
				os << (*fg.g)[target(*j_begin, *fg.g)].size << " ";
      os << std::endl;
      os << (*fg.g)[a].size << std::endl;
      for (int i=0; i<(*fg.g)[a].size; i++)
				os << i << "\t" << (*fg.g)[a].psi[i] << std::endl;
      os << std::endl;
    }
    iit.first++;
  }
  return os;
}

void FGraph::addEdge(const Vertex &s,
		     const Vertex &t,
		     BGraph &_g,
		     const uint &size,
		     const uint &pos)
{
  std::pair<Edge, bool> ce = add_edge(s, t, _g);
  VProb prob(_g[t].size, 1./_g[t].size);

  _g[ce.first].m_old = prob; _g[ce.first].m_new = prob;
  _g[ce.first].n_old = prob; _g[ce.first].n_new = prob;
  _g[ce.first].pos_index = pos;

	if (m_random_init) {
		_g[ce.first].m_old[0] = gsl_rng_uniform(m_r);
		_g[ce.first].m_old[1] = 1-gsl_rng_uniform(m_r);
		_g[ce.first].m_new[0] = gsl_rng_uniform(m_r);
		_g[ce.first].m_new[1] = 1-gsl_rng_uniform(m_r);
		_g[ce.first].n_old[0] = gsl_rng_uniform(m_r);
		_g[ce.first].n_old[1] = 1-gsl_rng_uniform(m_r);
		_g[ce.first].n_new[0] = gsl_rng_uniform(m_r);
		_g[ce.first].n_new[1] = 1-gsl_rng_uniform(m_r);
	}
}

void FGraph::buildFactors()
{
  std::pair<VertexIt, VertexIt> itf = vertices(*g);
  for (uint f = 0; (f < m_nfactors); itf.first++, f++)
    m_factors.push_back(*itf.first);
}

void FGraph::buildVarsSizes()
{
  // sort the variables by size
  std::pair<VertexIt, VertexIt> vpv = vertices(*g); vpv.first += m_nfactors;
  while (vpv.first != vpv.second) {
    VectorVertexIt ii = m_vars_sizes.begin();
    bool fnd = false;
    while (!fnd && (ii != m_vars_sizes.end())) {
      if ( (*g)[*ii].size > (*g)[*vpv.first].size ) fnd = true;
      else ii++;
    }
    m_vars_sizes.insert(ii, *vpv.first);
    vpv.first++;
  }
}

void FGraph::buildVarsPos()
{
  // sort the variables by label
  std::pair<VertexIt, VertexIt> vpv = vertices(*g); vpv.first += m_nfactors;
  int pos = 0;
  while (vpv.first != vpv.second) {
    m_vars_labels.push_back(*vpv.first);
    m_vars_pos[*vpv.first] = pos;

	  OutEdgeIt b_begin, b_end;
		int posf = 0;
    for (tie(b_begin, b_end) = out_edges(*vpv.first, (*g)); b_begin != b_end; b_begin++) {
      (*g)[*b_begin].pos_factor = posf;
			posf++;
    }
    vpv.first++;
    pos++;
  }
	for (int ii = 0; ii < (int)m_vars_labels.size(); ii++) {
		m_label2idx[(*g)[m_vars_labels[ii]].label] = ii;
	}

/*  for (std::map<int, int>::iterator it = m_label2idx.begin(); it != m_label2idx.end(); it++)
			std::cout << (*it).first << "\t" << (*it).second << std::endl;
*/
/*  sort(m_vars_labels.begin(), m_vars_labels.end());

  for (uint var = 0; var < m_vars_labels.size(); var++)
    std::cout << m_vars_labels[var] << " ";
  std::cout << std::endl;*/
}

void FGraph::exact() {

  uint q = 2, total_vals = 1;
  PIndex jidx;
  for (uint vv = 0; vv < m_nvars; vv ++) {
    total_vals *= q;
    jidx.addVar(q);
  }

  // compute joint by brute-force
  VProb joint(total_vals, 1.);
  VProb joint_log(total_vals, 0.);
  std::vector<bool> sign(total_vals, true);
  for (VectorVertexIt f = m_factors.begin(); f != m_factors.end(); f++) {

    // get the factor
    VProb factor = (*g)[*f].psi;
    PIndex idxf = (*g)[*f].index;

    if (degree(*f, *g) == 1) {
			AdjIt ab, ae;
      tie(ab, ae) = adjacent_vertices(*f, *g);
      uint pos_var = m_vars_pos[*ab]; //(*g)[*ab].label-1;
      for (uint i = 0; i < total_vals; i++) {
				joint[i] *= factor[jidx[i][pos_var]];
				joint_log[i] += Log(Abs(factor[jidx[i][pos_var]]));
				if (factor[jidx[i][pos_var]] < Real(0.))
					sign[i] = !sign[i];
      }
    }
    else if (degree(*f, *g) == 2) {
 
      AdjIt ab, ae;
      tie(ab, ae) = adjacent_vertices(*f, *g);
      uint pos_var1 = m_vars_pos[*ab]; //(*g)[*ab].label-1;
      ab++;
      uint pos_var2 = m_vars_pos[*ab]; //(*g)[*ab].label-1;
      if (pos_var1 > pos_var2) {
				throw "Unexpected error computing FGraph::exact";
			}

//      cout << "vars " << pos_var1 << " " << pos_var2 << endl;
      VectorInt midx = jidx.getIndex2(pos_var1, pos_var2);
//      for (int iu = 0; iu < midx.size(); iu++) cout << midx[iu] << " "; cout << endl;
      
      for (uint i = 0; i < total_vals; i++) {
				joint[i] *= factor[midx[i]];
				joint_log[i] += Log(Abs(factor[midx[i]]));
				if (factor[midx[i]] < Real(0.)) sign[i] = (!sign[i]);
      }
    }
    else {
      throw "EXACT only supports pairwise interactions (sorry)... Use libDAI";
    }
  }

	m_logZ_exact = Log(joint.normalizeL1());

	if (m_verbosity > 1) cout << "m_logZ_exact = " << m_logZ_exact << std::endl;

  // single node exact marginals
  std::vector<VProb> probs;
  for (uint i=0; i<m_nvars; i++) {
    VProb p(2,0.);
    for (uint j = 0; j < total_vals; j++) {
      p[jidx[j][i]] += joint[j];
    }
    probs.push_back(p);
    if (m_verbosity > 1) std::cout << "p[" << i << "] = " << p << " m = " << p[0] - p[1] << std::endl;
  }

  // pairwise exact marginals
  std::vector<std::vector<VProb> > prob2;
  for (uint i=0; i<m_nvars; i++) {
    std::vector<VProb> prob2i;
    for (uint j=i; j<m_nvars; j++) {
      VProb p2(q*q, 0.);
      VectorInt midx = jidx.getIndex2(i, j);
      for (uint k = 0; k < total_vals; k++) {
				p2[midx[k]] += joint[k];
      }
      //      std::cout << "p2[" << i << "," << j << "] = " << p2 << std::endl;
      prob2i.push_back(p2);
    }
    prob2.push_back(prob2i);
  }

  // exact covariances
  std::vector<std::vector<VProb> > covs;
  for (uint i = 0; i < m_nvars; i++) {
    std::vector<VProb> covi;
    for (uint j = i; j < m_nvars; j++) {
      VProb scov(q*q, 0.);
      for (uint k = 0, l = 0; k < q*q; k++) {
				if (k && (k%2 == 0)) l++;
				scov[k] = prob2[i][j-i][k] - probs[i][k%2]*probs[j][l];
      }
      //std::cout << "cov[" << i << "," << j << "] = " << scov << std::endl;
      covi.push_back(scov);
    }
    covs.push_back(covi);
  }

  // insert probabilities and covariances in the graph
  uint i = 0;
  for (VectorVertexIt itv = m_vars_labels.begin(); itv != m_vars_labels.end(); itv++){

    //    std::cout << (*g)[*itv].label << std::endl;
    (*g)[*itv].p = probs[i];
    for (uint j = 0; j < m_nvars; j++) {
      uint x = (i<j) ? i : j;
      uint y = (i>j) ? i : j;
      //      std::cout << "\t" << i << "," << j << " " << x << "," << y << std::endl;

      (*g)[*itv].m_cov.push_back(covs[x][y-x]);
    }
    i++;
  }
	if (m_verbosity > 2) std::cout << "EXACT]" << std::endl;
  m_method = EXACT1;
	m_exact = true;
}

VProb FGraph::getCov(int vari, int varj)
{
  // get the varj position of the m_cov of vari node
  // note that g is indexed by Vertex (obtained using m_vars_labels)
  // and m_cov is indexed by an index
  return (*g)[m_vars_labels[vari]].m_cov[varj];
}

FGraph::~FGraph()
{
  // delete random seed
  gsl_rng_free(m_r);

  // delete boost graph
  delete g;
}

// variable -> factor
VProb FGraph::getBPn(const Edge& e, bool old)
{
  OutEdgeIt b_begin, b_end;
  Vertex a = source(e, (*g));       // factor a
  Vertex i = target(e, (*g));       // variable i
  VProb n_msg((*g)[i].size,1.);

  // for each neighbor b of the variable j except a
  for (tie(b_begin, b_end) = out_edges(i, (*g)); b_begin != b_end; b_begin++) {
    Vertex b = target(*b_begin, (*g));  // the factor b
    if (b != a)
      n_msg *= (old) ? (*g)[*b_begin].m_old : (*g)[*b_begin].m_new;
  }

//  if (m_negpot)
//		n_msg.normalizeLInf();
//	else
//		n_msg.normalizeL1();

  return n_msg;
}

// factor -> variable
VProb FGraph::getBPm(const Edge& e, bool old)
{
  // declare variables outside the loops for efficiency
  OutEdgeIt j_begin, j_end;
  Vertex a = source(e, (*g));       // factor a
  Vertex i = target(e, (*g));       // variable i
  Vertex j;
  VProb nmsg;

  // temporal factor with the values of the potential and its index
  VProb factor((*g)[a].psi), prob;
  PIndex *index = &(*g)[a].index;

  // for each neighbor j of the factor a except i
  for (tie(j_begin, j_end) = out_edges(a, (*g)); j_begin != j_end; j_begin++) {

    j = target(*j_begin, (*g));  // the variable j
    if (j != i) {

      // iterate over all factor values indexing the n msg
      nmsg = getBPn(*j_begin, old);
      for (uint itf=0; itf<factor.size(); itf++)
        factor[itf] *= nmsg[index->get(itf,(*g)[*j_begin].pos_index)];
    }
  }

  // SUM STEP : marginalize to obtain the message
  prob.reset((*g)[i].size, 0.);
  for (uint itf = 0; itf < factor.size(); itf++)
    prob[index->get(itf, (*g)[e].pos_index)] += factor[itf];

  // normalize m messages
//  #ifdef NORM_BP
	if (m_negpot)
		prob.normalizeLInf();//NORM_BP);
	else
		prob.normalizeL1();//NORM_BP);
//  #endif

//	if ( (*g)[i].label == 6345) { 

//		std::cout << "Mmesg from "<<  a <<" is "<< prob;
//		std::cout << " size " << a << " is " << factor.size() << std::endl;
//  } 

  return prob;
}

bool FGraph::doBP(Real tol, uint max_iter, bool output)
{
  m_tol = tol;
  m_max_iter = max_iter;
	m_output = output;

  double init_clock, end_clock;
  init_clock = ttoc();

  // Initializations
  EdgeIt ei, ei_end;
  VProb msg;
  switch (m_bp_variant) {

  case BP_MAXRES:

    // do the first pass
    for (tie(ei, ei_end) = edges(*g); ei != ei_end; ++ei)
      (*g)[*ei].m_new = getBPm(*ei, true);

    // Initialize the residuals in a first iteration
    for (tie(ei, ei_end) = edges(*g); ei != ei_end; ++ei) {
      VProb msg = getBPm(*ei, true);
      Real residual = msg.mdiff((*g)[*ei].m_old);
      (*g)[*ei].it_residual = m_map_res.insert(MapResElem(residual, *ei));
    }
    break;

  case BP_SEQRND:

    // Initialize the random vector of edges
    for (tie(ei, ei_end) = edges(*g); ei != ei_end; ++ei)
      m_rnd_edges.push_back(*ei);
    break;

  default:
    break;
  }

  // Main loop
  OutEdgeIt b_begin, b_end, j_begin, j_end;
  Vertex a, i, j, b;
  uint step = 0;
  bool convergence = false;
  while ( (!convergence) && (step < max_iter) ) {

    step++;
    m_maxdif = -1.;
    switch (m_bp_variant) {

    case BP_MAXRES:

      for (uint ed = 0; ed < m_nedges; ed++) {
				
				// Koller et al
				// select the edge with highest residual O(1)
				MapResIt it_map = m_map_res.end();
				it_map--;
				Edge edge_maxr = (*it_map).second;

        //for (MapResIt it_map2 = m_map_res.begin(); it_map2 != m_map_res.end(); it_map2++) {
        //  std::cout << setprecision(5) << (*it_map2).first << " ";
        //}
        //std::cout << std::endl;

        // get new msg and replace the residual with 0
       	//	(*g)[edge_maxr].n_new = getBPn(edge_maxr);

        // this is just to reproduce Joris output...
				bool use_old = (!step) ? true : false;
				(*g)[edge_maxr].m_new = getBPm(edge_maxr, use_old);

				//std::cout << "ma" << (*g)[edge_maxr].m_new << std::endl;
				m_map_res.erase((*g)[edge_maxr].it_residual);
				(*g)[edge_maxr].it_residual = m_map_res.insert(MapResElem( 0., edge_maxr ));

				// update the affected messages
				a = source(edge_maxr, (*g));
				i = target(edge_maxr, (*g));
	
				for (tie(b_begin, b_end) = out_edges(i, (*g)); b_begin != b_end; b_begin++) {
					b = target(*b_begin, (*g));
					if (b != a) {
						for (tie(j_begin, j_end) = out_edges(b, (*g)); j_begin != j_end; j_begin++) {
							j = target(*j_begin, (*g));
							if (j != i) {
								VProb hip_msg = getBPm(*j_begin, false);
								m_map_res.erase((*g)[*j_begin].it_residual);
								Real residual = hip_msg.mdiff((*g)[*j_begin].m_new);
								//std::cout << "\t\t hip msg " << hip_msg << " new residual! " << residual << std::endl;
								(*g)[*j_begin].it_residual = m_map_res.insert(MapResElem( residual, *j_begin ));
							}
						}
					}
				}
      }
      break;

    case BP_SEQRND:

      // shuffle the ordering of the edges
      random_shuffle( m_rnd_edges.begin(), m_rnd_edges.end() );
      for (std::vector<Edge>::iterator it = m_rnd_edges.begin(); it != m_rnd_edges.end(); it++) {
        // (*g)[*it].n_new = getBPn(*it, false);
        (*g)[*it].m_new = getBPm(*it, false);
      }
      break;

    case BP_PARAL:

      // set the new msgs of each edge
      for (tie(ei, ei_end) = edges(*g); ei != ei_end; ++ei)
				(*g)[*ei].m_new = getBPm(*ei, true);

      // replace the old with new msgs
      for (tie(ei, ei_end) = edges(*g); ei != ei_end; ++ei)
				(*g)[*ei].m_old = (*g)[*ei].m_new;

      break;

    default:

      for (tie(ei, ei_end) = edges(*g); ei != ei_end; ++ei) {
				//	(*g)[*ei].n_new = getBPn(*ei);
				(*g)[*ei].m_new = getBPm(*ei, false);
      }
      break;
    }

    // get the beliefs
    OutEdgeIt out_b, out_e;
    for (VectorVertexIt itv = m_vars_labels.begin(); itv != m_vars_labels.end(); itv++) {
      Vertex v = *itv;
      VProb bel((*g)[v].size, Real(1.));
      
      // compute beliefs for each edge to the var v
      for (tie(out_b, out_e) = out_edges(v, *g); out_b != out_e; out_b++) {
//cout << *out_b << "\t" << (*g)[*out_b].m_new << endl;
        bel *= (*g)[*out_b].m_new;
      }

			bel.normalizeL1();					

			// compute differences with previous iteration and update the maximum
//cout << "comp"; cout.flush();
      Real bel_dif = (*g)[v].bel.mdiff(bel);
//cout << "]"; cout.flush();
      m_maxdif = (bel_dif > m_maxdif) ? bel_dif : m_maxdif;

      // set the belief
      (*g)[v].bel = bel;
    }

    convergence = (m_maxdif < m_tol);
    if (m_verbosity > 3) {
      std::cout << step << " -BP--------------------------- " << m_maxdif << std::endl;
    }
  }

	m_BP_steps = step;
  end_clock = ttoc();
  m_cpu_bp = end_clock - init_clock;

	if (!convergence) return false;
	else if (m_verbosity) {
    cerr << "BP converged after " << step << " iterations using " << getBPVariantString();
    cerr << " maxdiff = " << setprecision(14) << m_maxdif << std::endl;
  }

	// output number of iterations and cpu-time
	if ( m_output ) {
		std::ofstream fstats;
		char sfbel[100];
		sprintf(sfbel, "%s.stats", m_filename.substr(m_filename.find_last_of('/')+1, m_filename.size()-m_filename.find_last_of('/')).c_str());
		if (m_verbosity > 1) std::cout << "Opening file " << sfbel << " for writing" << std::endl;
		fstats.open(sfbel, ios::app);
		if (!fstats.good())
			throw "... cannot open the file " + std::string(sfbel);
		fstats << step << "\t" << m_cpu_bp << std::endl;
		fstats.close();
	}
 
  // compute last n msgs if required
//  if ( (m_bp_variant == BP_MAXRES) || (m_bp_variant == BP_PARAL) )
  for (tie(ei, ei_end) = edges(*g); ei != ei_end; ++ei) {
    (*g)[*ei].n_new = getBPn(*ei, false);
  }

  vector<Real> ms;
  if (convergence && (m_verbosity > 1)) { std::cout << "bp beliefs:" << std::endl; }

  int nmone = 0;

	// output beliefs
  std::ofstream fbel;
	if ( m_output ) {
	char sfbel[100];
  sprintf(sfbel, "%s.bel", m_filename.substr(m_filename.find_last_of('/')+1, m_filename.size()-m_filename.find_last_of('/')).c_str());
	if (m_verbosity > 1) std::cout << "Opening file " << sfbel << " for writing" << std::endl;
  fbel.open(sfbel);
		if (!fbel.good())
			throw "... cannot open the file " + std::string(sfbel);
	}
	Real bmax = 0.; Vertex vmax = -1;
	for (VectorVertexIt itv = m_vars_labels.begin(); itv != m_vars_labels.end(); itv++) {

    if (convergence && (m_verbosity > 1)) {
			std::cout << *itv << " -> ["<< (*g)[*itv].bel[0] << " " << (*g)[*itv].bel[1] << "]" << std::endl;
    }
    (*g)[*itv].m = (*g)[*itv].bel[0]-(*g)[*itv].bel[1];

    // warn about those beliefs with extreme values (less than machine precission)
    Real thresh = 1e-14;
    if ( ( ((*g)[*itv].bel[0] < thresh) || ((*g)[*itv].bel[1] < thresh) ) && m_verbosity ) {
      nmone++;
    }

		// output beliefs of non-dummy vars to a file
  	if ( m_output  ) {
			if ((*g)[*itv].bel[1] > bmax) { bmax = (*g)[*itv].bel[1]; vmax = *itv; }
	    fbel << (*g)[*itv].label << "\t" << setprecision(14) << (*g)[*itv].bel[0] << "\t" << (*g)[*itv].bel[1] << std::endl;
		}
  }
	if ( m_output ) fbel.close();

	// compute pairwise marginals
  if (convergence && (m_verbosity > 1)) std::cout << "bp (factor) beliefs:" << endl;
  for (VectorVertexIt f = m_factors.begin(); f != m_factors.end(); f++) {

    Vertex a = *f;
    VProb bel = (*g)[a].psi;
    PIndex *index = &(*g)[a].index;

    // multiply the factor with the n messages
    for (tie(j_begin, j_end) = out_edges(a, (*g)); j_begin != j_end; j_begin++)
      for (uint itf = 0; itf < bel.size(); itf++)
				bel[itf] *= (*g)[*j_begin].n_new[index->get(itf,(*g)[*j_begin].pos_index)];

		if (m_negpot)
			bel.normalizeLInf();
		else
			bel.normalizeL1();

    // set the belief to the factor node
    (*g)[a].bel = bel;
		if (m_verbosity > 1) cout << a << " -> " << bel << endl;

    // bel is the marginal of the vars within the factor
    // marginalize over all pairs in the factor
    OutEdgeIt j1_begin, j1_end, j2_begin, j2_end;
    for (tie(j1_begin, j1_end) = out_edges(a, (*g)); j1_begin != j1_end; j1_begin++) {

      Vertex j1 = target(*j1_begin, (*g));

      // check single node marginals
      VProb sm((*g)[j1].size,0.);
      for (uint itf = 0; itf < bel.size(); itf++)
        sm[index->get(itf,(*g)[*j1_begin].pos_index)] += bel[itf];

			for (tie(j2_begin, j2_end) = out_edges(a, (*g)); j2_begin != j2_end; j2_begin++) {
        Vertex j2 = target(*j2_begin, (*g));
        if (j2 >= j1) {
          VProb marg((*g)[j1].size*(*g)[j2].size,0.);
          int var1 = (*g)[*j1_begin].pos_index;
          int var2 = (*g)[*j2_begin].pos_index;
          VectorInt index2 = index->getIndex2(var1, var2);
          for (uint itf = 0; itf < bel.size(); itf++)
            marg[index2[itf]] += bel[itf];
        }
      }
    } // end pairwise marginals
  } // end factors

  // Calculate the stationary point of the Bethe free energy to get the Z
	// It is computed according the Yedidia paper "constructing free energy ..."
	// First, traverse all the factors and get the U_b and H_b subterms
  Real Ubethe = 0., Hbethe = 0.;
  for (VectorVertexIt f = m_factors.begin(); f != m_factors.end(); f++) {
    VProb *factor = &((*g)[*f].psi), *bel = &((*g)[*f].bel);
    for (uint i = 0; i < (*factor).size(); i++) {
			if ( ((*factor)[i]<Real(0)) || ((*bel)[i]<Real(0)) ) {
	 	    std::cerr << "negative beliefs " << (*factor)[i] << " " << (*bel)[i] << std::endl;
			} else {
        if (((*factor)[i] > Real(0.)) && ((*bel)[i] > Real(0.))) {
          Ubethe += (*bel)[i] * Log((*factor)[i]);
          Hbethe += (*bel)[i] * Log((*bel)[i]);
					if ( isnan(Ubethe) || isnan(Hbethe) ) {
						std::cerr << *f << " " << (*factor)[i] << " " << (*bel)[i] << std::endl;
						throw "isnan";
					}
        }
      }
    }
  }
  Hbethe = -Hbethe;

	// Second, traverse all the variables to get the H_b remaining terms
	Real term(0.);
  for (VectorVertexIt itv = m_vars_labels.begin(); itv != m_vars_labels.end(); itv++) {
    VProb *bel = &((*g)[*itv].bel);
    Real h;
    if ( ((*bel)[0]>Real(0.)) && ((*bel)[1]>Real(0.)) )
			h = ( (*bel)[0] * Log((*bel)[0]) ) + ( (*bel)[1] * Log((*bel)[1]) );
		else
			h = 0;
    Hbethe += (degree(*itv, *g)-1)*h;
		term += (degree(*itv, *g)-1)*h;
		if (isnan(Hbethe)) throw "isnan";
  }
  m_logZ_bp = (Ubethe + Hbethe);
  if (m_verbosity > 1) {
		std::cout << "U = " << Ubethe << std::endl;
	  std::cout << "S = " << Hbethe << std::endl;
	}

  return true;
}
          
std::map<Vertex, VProb> FGraph::clamp(int var, uint val)
{
  std::map<Vertex, VProb> map_factors;
  Vertex i = m_vars_labels[var];

	if (m_verbosity > 1) cout << "clamping " << (*g)[m_vars_labels[var]].label << std::endl;
  // traverse all its factors
  OutEdgeIt b_begin, b_end;
  for (tie(b_begin, b_end) = out_edges(i, (*g)); b_begin != b_end; b_begin++) {
    Vertex b = target(*b_begin, (*g));  // the factor b
    if ( (*g)[b].type != FACTOR ) {
      std::cerr << "Error! not a factor!" << std::endl;
      return map_factors;
    }

		VProb factor((*g)[b].psi);
    PIndex *index = &(*g)[b].index;
    map_factors[b] = factor;
    for (uint itf=0; itf<factor.size(); itf++) {
      if (val != index->get(itf,(*g)[*b_begin].pos_index))
        factor[itf] = 0;
    }
    (*g)[b].psi = factor;
  }

  // check clampings
/*  std::cout << "[new factors" << std::endl;
  for (VectorVertexIt f = m_factors.begin(); f != m_factors.end(); f++) {
    std::cout << (*g)[*f].psi << std::endl;
  }
  std::cout << "new factors]" << std::endl;
*/
  return map_factors;
}

void FGraph::unclamp(int var, std::map<Vertex, VProb> factors)
{
  for (std::map<Vertex, VProb>::iterator it = factors.begin(); it != factors.end(); it++)
    (*g)[(*it).first].psi = (*it).second;
}

void FGraph::reset()
{
  EdgeIt ei, ei_end;
  for (tie(ei, ei_end) = edges(*g); ei != ei_end; ++ei) {
    (*g)[*ei].m_new.fill(1.); (*g)[*ei].m_old.fill(1.);
    (*g)[*ei].n_new.fill(1.); (*g)[*ei].n_old.fill(1.);
  }
}


bool FGraph::doBPForney(Real tol, uint maxiter)
{
	VertexIt vi, ve;
	if (m_verbosity) cout << "performing BP in the Forney graph (SEQRANDOM)" << endl;
	
	// start pfaffian cpu-clock
  double init_bp_forney_clock = ttoc(), final_bp_forney_clock;

	// output graph
	if (m_verbosity > 2) {
	  for (tie(vi, ve) = vertices(m_forney); vi != ve; vi++) {
	    FVertexD a = *vi;
			cout << "Vertex " << a << endl;
			FInEdgeIt i_begin, i_end;
			FOutEdgeIt o_begin, o_end;
			cout << "\tindex " << endl; cout << m_forney[a].index << endl;
			for (tie(o_begin, o_end) = out_edges(a, m_forney); o_begin != o_end; o_begin++)
				cout << "\t" << *o_begin << " " << m_forney[*o_begin].pos_index << "\t";
			cout << endl;
			for (tie(i_begin, i_end) = in_edges(a, m_forney); i_begin != i_end; i_begin++)
				cout << "\t" << *i_begin << " " << m_forney[*i_begin].pos_index << "\t";
			cout << endl;
		}
	}

	if (m_verbosity > 1) forney2Dot("forney.dot", m_forney);
	
	// initialize messages
  FEdgeIt fei, fee;
	for (tie(fei, fee) = edges(m_forney); fei != fee; ++fei)
		m_forney[*fei].msg = VProb(2, .5);

	bool conv = false, end = false;
	uint nit = 0;

  if (m_verbosity > 2) cout << "num edges Forney is " << num_edges(m_forney) << endl;

	// random shuffling of the edges
	for (tie(fei, fee) = edges(m_forney); fei != fee; ++fei) 
		m_rnd_edges_forney.push_back(*fei);

	while (!conv && !end) {
		conv = true;
		if (m_verbosity > 3) { cout << "IT " << nit; cout.flush(); }
		m_maxdif = -1;

		if (m_bp_variant == BP_SEQRND) { 
      random_shuffle( m_rnd_edges_forney.begin(), m_rnd_edges_forney.end() );
		}
 
		for (vector<FEdgeD>::iterator fei = m_rnd_edges_forney.begin(); fei != m_rnd_edges_forney.end(); ++fei) {
			FVertexD i = source(*fei, m_forney), j = target(*fei, m_forney);
			VProb phi = m_forney[i].psi;
			if (m_verbosity > 3) { cout << "\tphi = " << phi << endl; }
			FInEdgeIt i_begin, i_end;
      for (tie(i_begin, i_end) = in_edges(i, m_forney); i_begin != i_end; i_begin++) {
				if (source(*i_begin, m_forney) != j) {
					if (m_verbosity > 3) cout << "\t\tmsg " << *i_begin << " = " << m_forney[*i_begin].msg << endl;
					for (uint p = 0; p < phi.size(); p++) {
	
						// multiply message with factor
						// we need to know which position of the message corresponds to the
						// value p of the factor
						phi[p] *= m_forney[*i_begin].msg[ m_forney[i].index.get(p, m_forney[*i_begin].pos_index) ];
					}
				}
			}
			phi.normalizeL1();
			if (m_verbosity > 3) { cout << "\tphi = " << phi << endl; }

			// marginalize
			VProb msg = VProb(2, 0.);
			FEdgeC et = edge(j, i, m_forney);
			for (uint p=0; p<phi.size(); p++) {

				// we need to know which position of the 
				msg[m_forney[i].index.get(p, m_forney[et.first].pos_index)] += phi[p];
			}
			if (m_forney[*fei].msg.mdiff(msg) > m_maxdif) m_maxdif = m_forney[*fei].msg.mdiff(msg);
			m_forney[*fei].msg = msg;
			if (m_verbosity > 3) {
				cout << "\t msg " << *fei << " = " << msg << endl;
				cout << "-------------------------------------------" << endl;
			}
		}
		conv = (tol > m_maxdif);
		end = (nit == maxiter);
		if (m_verbosity > 3) cout << "---- it " << nit << " -> " << m_maxdif << endl;
		nit++;
	}

	// BP clock
	final_bp_forney_clock = ttoc() - init_bp_forney_clock;

	if (m_verbosity > 1) {
    cout << endl << "BP : Spent " << final_bp_forney_clock << " scnds). " << endl;
	}
	m_cpu_bp_forney = final_bp_forney_clock;
	
	if (end) 
		cerr << "Forney: NO BP CONVERGENCE after " << nit << " iterations " << endl;
	else {
		if (m_verbosity) cerr << "convergence in " << nit << " iterations" << endl;

		// compute beliefs
		// single variable beliefs
		// on traversing edges remember that they are doubled...
		map<int, VProb> m_bels;
		map<int, int> m_vars_count;
		for (tie(fei, fee) = edges(m_forney); fei != fee; ++fei) {
			FVertexD i = source(*fei, m_forney), j = target(*fei, m_forney);
			int id = m_forney[*fei].id;
			FEdgeC e1 = edge(i, j, m_forney), e2 = edge(j, i, m_forney);
			VProb m1 = m_forney[e1.first].msg, m2 = m_forney[e2.first].msg; 
			m1 *= m2;
			VProb bel = m1;
			bel.normalizeL1();
			Real th = (tol > Real(1e-10)) ? (tol*Real(10)) : Real(1e-10);
	    if (m_bels.find(id) == m_bels.end()) {
				m_bels[id] = bel;
				m_vars_count[id] = 1;
			}
			else m_vars_count[id]++;
			if ( Abs(m_bels[id][0]-bel[0]) >  th) {
				cerr << m_bels[id][0] << " - " << bel[0] << " = " << Abs(m_bels[id][0]-bel[0]);
				cerr << " > " << tol << "\t";
				throw "Error BP Forney!";
			}
			m_forney[e1.first].bel = bel;
			m_forney[e2.first].bel = bel;
		}
	
		// irreducible correlations (nodes)
	  for (tie(vi, ve) = vertices(m_forney); vi != ve; vi++) {
	    FVertexD a = *vi;
			VProb bel = m_forney[a].psi;
			FInEdgeIt i_begin, i_end;
	 		for (tie(i_begin, i_end) = in_edges(a, m_forney); i_begin != i_end; i_begin++) 
				for (uint p = 0; p < bel.size(); p++) 
					bel[p] *= m_forney[*i_begin].msg[ m_forney[a].index.get(p, m_forney[*i_begin].pos_index) ];
	
			bel.normalizeL1();
			m_forney[a].bel = bel;
		} 
		if (m_verbosity > 1) { 
			cout << "single variable beliefs" << endl;
			for (map<int, VProb>::iterator it = m_bels.begin(); it != m_bels.end(); it++)
				cout << (*it).first << " -> " << (*it).second << endl;
			cout << "factor beliefs" << endl;
			for (tie(vi, ve) = vertices(m_forney); vi != ve; vi++) 
				cout << *vi << " -> " << m_forney[*vi].bel << endl;
		}

// TO COMPARE A POSSIBLE DISCREPANCY BETWEEN THE BELIEFS OF BP FACTOR GRAPH AND FORNEY GRAPH		
//ofstream fout("bels_bp_forney.txt");
//for (tie(fei, fee) = edges(m_forney); fei != fee; ++fei)
//  fout << *fei << " \t" << m_forney[*fei].bel << endl;
//for (tie(vi, ve) = vertices(m_forney); vi != ve; vi++) 
//	fout << *vi << " \t" << m_forney[*vi].bel << endl;
//fout.close();

		// compute Z_BP forney
	  // Calculate the stationary point of the Bethe free energy to get the Z
		// It is computed according the Yedidia paper "constructing free energy ..."
		// First, traverse all the factors and get the U_b and H_b subterms
	  Real Ubethe = 0., Hbethe = 0.;
		for (tie(vi, ve) = vertices(m_forney); vi != ve; vi++) {
			FVertexD a = *vi;
	    VProb *factor = &(m_forney[a].psi), *bel = &(m_forney[a].bel);
	    for (uint i = 0; i < (*factor).size(); i++) {
				if ( ((*factor)[i]<Real(0)) || ((*bel)[i]<Real(0)) ) {
		 	    cerr << "negative beliefs " << (*factor)[i] << " " << (*bel)[i] << endl;
				} else {

	        if (((*factor)[i] > Real(0.)) && ((*bel)[i] > Real(0.))) {
	          Ubethe += (*bel)[i] * Log((*factor)[i]);
						Hbethe += (*bel)[i] * Log((*bel)[i]);
						
						if ( isnan(Ubethe) || isnan(Hbethe) ) {
							cerr << a << " " << (*factor)[i] << " " << (*bel)[i] << endl;
							throw "isnan";
						}
	        }
	      }
	    }
	  }
	  Hbethe = -Hbethe;
		Real term(0.);
		for (map<int, VProb>::iterator itv = m_bels.begin(); itv != m_bels.end(); itv++) {
	    VProb *bel = &(m_bels[(*itv).first]);
	    Real h;
	    if ( ((*bel)[0]>Real(0.)) && ((*bel)[1]>Real(0.)) )
				h = ( (*bel)[0] * Log((*bel)[0]) ) + ( (*bel)[1] * Log((*bel)[1]) );
			else
				h = 0;
			m_vars_count[(*itv).first] = m_vars_count[(*itv).first]/2;
	    Hbethe += m_vars_count[(*itv).first]*h;
			term += m_vars_count[(*itv).first]*h;
			
			if ( isnan(Hbethe) ) throw "isnan";

		}
		m_logZ_bp_forney = (Ubethe + Hbethe);
	  if (m_verbosity > 1) {
			cout << "U = " << Ubethe << endl;
		  cout << "S = " << Hbethe << endl;
		}
	}
	return (!end);
}

std::ostream& operator<< (std::ostream& os, PfTerm& t)
{
	os.unsetf(ios::scientific);
	os.setf(ios::fixed);
	os << "[";
	os << (double)t.mu;
	os << "|";
	if ( !t.sign ) os << "f";
	else os << "t";
	os << "]";
	
	return os;
}
 
