/*  Copyright (C) 2006  Joris Mooij  [j dot mooij at science dot ru dot nl]
    and Bastian Wemmenhove [b dot wemmenhove at science dot ru dot nl]
    Radboud University Nijmegen, The Netherlands

    This file is part of libDAI.

    libDAI is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    libDAI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with libDAI; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <iostream>
#include <sstream>
#include <map>
#include <set>
#include <algorithm>

#include "gmr.h"
#include "diffs.h"
#include "properties.h"
#include "alldai.h"
#include "util.h"
#include "index.h"
#include "varset.h"



// To check: Does Joris allow for single node potentials? Then I should incorporate them.


using namespace std;

const char *GMR::Name = "GMR";

void GMR::Regenerate(){
  DAIAlgFG::Regenerate();

  // calculate the ijI
  if(Verbose()>1) cout << "the ijI: " << endl;
  for( size_t i=0; i < nrVars(); i++ ) {
    for( _nb_cit I = nbV(i).begin(); I != nbV(i).end(); I++ )     
      for( _nb_cit j = nbF(*I).begin(); j != nbF(*I).end(); j++ )
	if( *j != i ) {
	  _ijI_type _ijI_entry;
	  _ijI_entry.i = i;
	  _ijI_entry.j = *j; 
	  _ijI_entry.I = *I;
	  _ijI.push_back(_ijI_entry);
	  if(Verbose()>1) cout << "(" << i << "," << *j << "," << *I << ")" << endl;
	}
  }
  cout << endl;
  // calculate the iIJ
  for( size_t i=0; i < nrVars(); i++ ) {
    for( _nb_cit I = nbV(i).begin(); I != nbV(i).end()-1; I++ )     
      for( _nb_cit J = I+1; J != nbV(i).end(); J++ ){
	  _iIJ_type _iIJ_entry;
	  _iIJ_entry.i = i;
	  _iIJ_entry.I = *I;
	  _iIJ_entry.J = *J;
	  _iIJ.push_back(_iIJ_entry);
	  }
  }
    // calc ijJ
  for( size_t i=0; i < nrVars(); i++ ) {
    for( _nb_cit I = nbV(i).begin(); I != nbV(i).end(); I++ )
      for( _nb_cit j = nbF(*I).begin(); j != nbF(*I).end(); j++ )
	if( *j != i ) {
	  _ijJ_type _ijJ_entry;
	  _ijJ_entry.i = i;
	  _ijJ_entry.j = *j;
	  for( _nb_cit J = nbV(*j).begin(); J != nbV(*j).end(); J++ )
	    if( *J != *I) {
	      _ijJ_entry.J = *J;	  
	      _ijJ.push_back(_ijJ_entry);
	    }
	}
  }
  
  // have to calculate _jik


  for( size_t i=0; i < nrVars(); i++){
    _jik_type _jik_entry;
    _jik_entry.i = i;
    for( _nb_cit I = nbV(i).begin(); I != nbV(i).end(); I++)
      for( _nb_cit j = nbF(*I).begin(); j != nbF(*I).end(); j++)
	if( *j != i ){
	  _jik_entry.j = *j;
	  for( _nb_cit J = nbV(i).begin(); J != nbV(i).end(); J++)
	    for( _nb_cit k = nbF(*J).begin(); k != nbF(*J).end(); k++){
	      if( (*k != i )&&(*k != *j) ){
		_jik_entry.k = *k;
		_jik.push_back(_jik_entry);	
	      }
	      if( (*k==*j) && (*J!=*I) )
		cout << "Warning, we have 'short loops'.." << endl;
	    }
	}
  } 
		   
  // calc _EijI_ind

  _EijI.clear();
  vector<size_t> cola(nrVars(),-1);
  _EijI.assign(nrVars(),cola);
  for(size_t ind = 0; ind < _ijI.size(); ind++ )   
    _EijI[_ijI[ind].i][_ijI[ind].j] = ind;

  _EiIJ.clear();
  vector<size_t> colb(nrFactors(),-1);
  _EiIJ.assign(edges().size(),colb);
  for(size_t ind = 0; ind < _iIJ.size(); ind++ )
    _EiIJ[VV2E(_iIJ[ind].i,_iIJ[ind].I)][_iIJ[ind].J]=ind;

  _EijJ.clear();
  vector<size_t> colc(nrFactors(), -1);
  _EijJ.assign(_ijI.size(),colc);
  for(size_t ind = 0; ind < _ijJ.size(); ind++ )
    _EijJ[_EijI[_ijJ[ind].i][_ijJ[ind].j]][_ijJ[ind].J] = ind;

  _Ejik.clear();
  vector<size_t> cold(nrVars(), -1);
  _Ejik.assign(_ijI.size(),cold);
  if(Verbose()>1) cout << " the jik: " << endl;
  for(size_t ind = 0; ind < _jik.size(); ind++ ){
    _Ejik[_EijI[_jik[ind].i][_jik[ind].j]][_jik[ind].k] = ind;            // this is all rather suboptimal, I am using
  // loads of unnecessary memory. I could try to store things in sparse structures instead.
    if(Verbose()>1) cout << "(" << _jik[ind].i << "," << _jik[ind].j << "," << _jik[ind].k << ")" << endl;
  }
  _Pmess.clear(); 
  _Pmess.reserve(nr_ijI());

  _mumess.clear();
  _mumess.reserve(nrEdges());
  
  _lambdamess.clear();
  _lambdamess.reserve(nrEdges());

  _rhomess.clear();
  _rhomess.reserve(nr_iIJ());
  
  _numess.clear();
  _numess.reserve(nr_ijJ());

  _concorrel.clear();
  _concorrel.reserve(nr_jik());

  _beliefs.clear();
  _beliefs.reserve(nrVars());


  for(size_t ij=0; ij < nr_ijI();  ij++){
    _Pmess.push_back(Factor(var(_ijI[ij].i)));
  }
  if(Verbose()>1) cout << "the iI: " << endl;
  for(vector<_edge_t>::const_iterator iI=edges().begin(); iI!=edges().end(); 
      iI++){
    _mumess.push_back(Factor(var(iI->first)) );
    if(Verbose()>1) cout << "i,I = " << iI->first << "," << iI->second << endl;
  }
  for(vector<_edge_t>::const_iterator iI=edges().begin(); iI!=edges().end(); 
      iI++){
    _lambdamess.push_back(Factor(var(iI->first)) );
  }
  for(size_t iIJ=0; iIJ < nr_iIJ(); iIJ++){
    _rhomess.push_back(Factor(var(_iIJ[iIJ].i)));
  }
  for(size_t ijJ=0; ijJ<nr_ijJ(); ijJ++){
    _numess.push_back(Factor(var(_ijJ[ijJ].i)|var(_ijJ[ijJ].j)));
  }
  for(size_t jik=0; jik<nr_jik(); jik++){
    _concorrel.push_back(Factor(var(_jik[jik].j)|var(_jik[jik].k)));
  }
  for(size_t i=0; i<nrVars(); i++)
    _beliefs.push_back(Factor(var(i)));

}


bool GMR::checkProperties(){



    if( !HasProperty("tol") )
        return false;
    if (!HasProperty("maxiter") )
        return false;
    if (!HasProperty("verbose") )
      return false;
    if (!HasProperty("epsilon") )
     return false;
    ConvertPropertyTo<double>("epsilon");

    ConvertPropertyTo<double>("tol");
    ConvertPropertyTo<size_t>("maxiter");
    ConvertPropertyTo<size_t>("verbose");

  if (HasProperty("cavainame") )
    ConvertPropertyTo<string>("cavainame");
  if (HasProperty("cavaiopts") )
    ConvertPropertyTo<Properties>("cavaiopts");
  if( HasProperty("reinit") )
    ConvertPropertyTo<bool>("reinit");
  
  return true;

}

void GMR::init(){
  for(size_t iI=0; iI < nrEdges(); iI++){
    _mumess[iI].randomize();
    //   cout << " mu messages: " << Mumes(edge(iI).first,edge(iI).second) << endl;    
    if( ( nbF(edge(iI).second).size()) > 2 )
      _lambdamess[iI].randomize();
    else
      _lambdamess[iI].fill(0.0);
    //  cout << " lambda messages: " << Lammes(edge(iI).first,edge(iI).second) << endl;
  }
  for(size_t iIJ=0; iIJ < nr_iIJ(); iIJ++){
    _rhomess[iIJ].randomize();
    // cout << " rho messages: " << Rhomes(_iIJ[iIJ].i,_iIJ[iIJ].I, _iIJ[iIJ].J) << endl;
  }
  for(size_t ijJ=0; ijJ < nr_ijJ(); ijJ++){
    _numess[ijJ].randomize();
    // cout << "nu messages: " << Numes(_ijJ[ijJ].i, _ijJ[ijJ].j, _ijJ[ijJ].J) << endl;
  }
  for(size_t ijI=0; ijI < nr_ijI(); ijI++){
    _Pmess[ijI].randomize();
    _Pmess[ijI] = _Pmess[ijI]/_Pmess[ijI].totalSum();
    // cout << "P messages: " << Pmes(_ijI[ijI].i,_ijI[ijI].j) << endl;
  }
  
}  

double GMR::run(){
  double tic = toc();
  Diffs diffs(nrVars(), 1.0);
  // init();
  vector<Factor> old_beliefs;    
  vector<size_t> edge_seq;
  edge_seq.reserve( nr_ijI() );
  for( size_t i = 0; i < nr_ijI(); i++ )
    edge_seq.push_back( i );

  old_beliefs.reserve( nrVars() ); 
  for( size_t i = 0; i < nrVars(); i++ ) 
    old_beliefs.push_back(belief1(i));

  Initcavs( GetPropertyAs<string>("cavainame"),GetPropertyAs<Properties>("cavaiopts") );    
// We might have to inherit from LCBP to make all names and options be
    // allright. Otherwise, we have to introduce new aliases anyway.

    // cavainame is probably going to be "BP", cavaiopts is e.g. the 
    // updatetype.
  cout << "cavity initialisation finished " << endl;
  cout << nr_ijI() << "," << nr_ijJ() << "," << nr_jik() << "," << nr_iIJ() << endl;
  size_t iter = 0;
  for( iter=0; iter < MaxIter() && diffs.max() > Tol(); iter++ ) {
      cout << "iteration " << iter << ", maxdiff = " << diffs.max() << endl;

       UpdateOtherMessages();  
    if(Verbose()>1)   cout << "other messages updated .. " << endl;

    //  random_shuffle( edge_seq.begin(), edge_seq.end() );
    //    for( size_t t = 0; t < nr_ijI(); t++ ) {
    //size_t ijI = edge_seq[t];
    for(size_t ijI = 0; ijI< nr_ijI(); ijI++){
      //UpdatePmessNew(ijI);
      UpdatePmess(ijI);
      if(Verbose()>1)  cout << "Pmess " << ijI << " updated " << endl;
     
    }
    for( size_t i = 0; i < nrVars(); i++ ) {
      Factor nb( belief1(i) );
      diffs.push( dist( nb, old_beliefs[i], Prob::DISTLINF ) ); //is allright, 
      // dist is just some probability distribution, defined in factor.h 
      old_beliefs[i] = nb;
    }  
  
 
  }
  Iterations( iter );

  if( Verbose() >= 1 ) {
    if( diffs.max() > Tol() ) {
      if( Verbose() == 1 )
	cout << endl;
      cout << "GMR::run:  WARNING: not converged within " << MaxIter()
	   << " passes (" << toc() - tic << " clocks)...final maxdiff:" << diffs.max() << endl;
    } else {
      if( Verbose() >= 3 )
	cout << "GMR::run:  ";
      cout << "converged in " << iter << " passes (" << toc() - tic << " clocks)." << endl;
    }
  } 
  if(Verbose()>1)  cout << "calculating beliefs "<< endl;
  calcbeliefs();
  //  cout << "allright" << endl;
  return diffs.max();
}

void GMR::UpdateOtherMessages(){
  if(Verbose()>1) cout << "Mu, Lam.." << endl;
  for(size_t iI=0; iI < nrEdges(); iI++){
    UpdateMumes(iI);
    if( ( nbF(edge(iI).second).size()) > 2 )
    UpdateLammes(iI);
  }   
  if(Verbose()>1) cout  << "Rho.. " << endl;
  for(size_t iIJ=0; iIJ < nr_iIJ(); iIJ++){  
    UpdateRhomes(iIJ);
  }    
  if(Verbose()>1)  cout << "Nu.. " << endl;
  for(size_t ijJ=0; ijJ < nr_ijJ(); ijJ++){
    UpdateNumes(ijJ);
  }
}

void GMR::UpdateMumes(size_t iI){
  size_t i = edge(iI).first;
  size_t I = edge(iI).second; 
  Factor prod( factor( I ) );
  for( _nb_cit j = nbF(I).begin(); j!=nbF(I).end(); j++){
    if( *j != i ){
      prod *= Pmes(*j,i);
    }
  }
  _mumess[iI] = _mumess[iI]*(1.-Eps())+prod.partSum(var(i))*Eps(); 
  if(Verbose()>2) cout << "mu message " << iI << " is " << _mumess[iI] << endl;
}

void GMR::UpdateLammes(size_t iI){
  size_t i = edge(iI).first;
  size_t I = edge(iI).second;
  vector<Factor> prod;
  Factor Lam(var(i),0.0);
  prod.reserve((nbF(I).size()*(nbF(I).size()-1))/2);
  size_t m = 0;
  for( _nb_cit j = nbF(I).begin(); j!=nbF(I).end(); j++)
    if( *j != i )
      for( _nb_cit k = j+1; k!=nbF(I).end(); k++)
        if( *k != i)
	  prod.push_back( factor( I ) );

  for( _nb_cit j = nbF(I).begin(); j!=nbF(I).end()-1; j++)
    if( *j != i )
      for( _nb_cit k = j+1; k!=nbF(I).end(); k++)
        if( *k != i){
	  for( _nb_cit l = nbF(I).begin(); l!=nbF(I).end(); l++)
	    if( (*l != i)&&( (*l !=*j )&&( *l!=*k )))
	      prod[m] *= Pmes(*l,i); // is *= P^{(i)}(x_l)
	  
	  prod[m] *= Concor(i,*j,*k);  // is *= c^{(i)}(x_j,x_k)
	  m++;
	}
  for(size_t r = 0; r<m; r++){
    Lam += prod[r].partSum(var(i));
  }
  _lambdamess[iI] = _lambdamess[iI]*(1.-Eps())+ Lam*Eps();
  if(Verbose()>2) cout << "lambda message " << iI << " is " << _lambdamess[iI] << endl;
}

void GMR::UpdateRhomes(size_t iIJ){
  size_t i= _iIJ[iIJ].i;
  size_t I= _iIJ[iIJ].I;
  size_t J= _iIJ[iIJ].J;
  size_t m1=0;
  size_t m2=0;
  size_t m=0;

  VarSet boys;
  vector<Factor> prod1;
  vector<Factor> prod2;
  vector<Factor> bla;
  Factor Rho(var(i),0.0);
  prod1.reserve(nbF(I).size()-1);
  prod2.reserve(nbF(J).size()-1);
  bla.reserve((nbF(I).size()-1)*(nbF(J).size()-1));
  
  for( _nb_cit j = nbF(I).begin(); j!= nbF(I).end(); j++)
    if( *j != i)
      prod1.push_back( factor( I ) );
  for( _nb_cit l = nbF(I).begin(); l!= nbF(I).end(); l++)
    for( _nb_cit k = nbF(J).begin(); k!= nbF(J).end(); k++)
      if( (*l != i)&&(*k != i) ) 
	prod2.push_back( factor( J ) );
 
  for( _nb_cit j = nbF(I).begin(); j!= nbF(I).end(); j++)
    if( *j != i){
      for( _nb_cit k = nbF(I).begin(); k!= nbF(I).end(); k++)
	if(( *k != i)&&( *k!=*j ))
	  prod1[m1] *= Pmes(*k,i);  
      prod1[m1]=prod1[m1].partSum(var(*j)|var(i));
      m1++;
    }
  for( _nb_cit l = nbF(J).begin(); l!= nbF(J).end(); l++)
    if( *l != i){
      for( _nb_cit k = nbF(J).begin(); k!= nbF(J).end(); k++)
	if( ( *k != i )&&( *k != *l ) )
	  prod2[m2] *= Pmes(*k,i);
      prod2[m2]=prod2[m2].partSum(var(*l)|var(i));
      m2++;
    }
  m1=0;
  for( _nb_cit j = nbF(I).begin(); j!= nbF(I).end(); j++)
    if( *j != i){
      m2=0;
      for( _nb_cit l = nbF(J).begin(); l!= nbF(J).end(); l++)
	if( *l != i){
	  bla.push_back(prod1[m1]*(prod2[m2]*Concor(i,*j,*l)));
	  
	  m2++;
	  m++;
	}
      m1++;
    }
  for(size_t r=0; r<m; r++)
    Rho += bla[r].partSum(var(i));	
  _rhomess[iIJ] = _rhomess[iIJ]*(1.-Eps())+Rho*Eps();
  if(Verbose()>0) cout << "rho message " << iIJ << " is " << _rhomess[iIJ] << endl;
}

void GMR::UpdateNumes(size_t ijJ){
  size_t i=_ijJ[ijJ].i;
  size_t j=_ijJ[ijJ].j;
  size_t J=_ijJ[ijJ].J;
  vector<Factor> prod;
  Factor Nu(var(j)|var(i),0.0);
  prod.reserve(nbF(J).size()-1);
  size_t m=0;
  for(_nb_cit k=nbF(J).begin(); k!=nbF(J).end() ; k++)
    if(*k != j)
      prod.push_back( factor( J ) );
  for(_nb_cit k=nbF(J).begin(); k!=nbF(J).end() ; k++)
    if(*k != j){
      for( _nb_cit l=nbF(J).begin(); l!=nbF(J).end(); l++)
	if( (*l != j)&&(*l != *k) ){
	  prod[m] *= Pmes(*l,j);
	}
      prod[m] *= Concor(j,i,*k); 
      m++;
    }
  for(size_t r=0; r<m; r++)
    Nu += prod[r].partSum(var(j)|var(i));
  _numess[ijJ] = _numess[ijJ]*(1.-Eps())+Nu*Eps();
  if(Verbose()>2) cout << "nu message " << ijJ << " is " << _numess[ijJ] << endl;
}


void GMR::Initcavs( const string & name, const Properties & opts){
  vector<Factor> concor;
  VarSet delt;
  vector<Var> deltvar;
  size_t N;
  for(size_t i=0; i<nrVars(); i++){
    InfAlg *cav = newInfAlg( name, *this, opts ); 
    cav->makeCavity( var(i) ); // factorgraph function.
    delt=cav->delta(var(i));
    concor=calcConcorrel(*cav, delt, reInit());  
    N=delt.size();    
    deltvar.clear();                 // perhaps this is not the most 
    deltvar.reserve( N );               // efficient way...
    for( VarSet::const_iterator n = delt.begin() ; n != delt.end(); n++)
      deltvar.push_back( *n );
    for(size_t j=0 ; j<N; j++)
      for(size_t k=j+1 ; k<N; k++){
	_concorrel[_Ejik[_EijI[i][findVar(deltvar[j])]][findVar(deltvar[k])]] = concor[N*j-(j+2)*(j+1)/2+k];
	_concorrel[_Ejik[_EijI[i][findVar(deltvar[k])]][findVar(deltvar[j])]] = concor[N*j-(j+2)*(j+1)/2+k];
        if(Verbose()>0) cout << Concor(i,findVar(deltvar[j]),findVar(deltvar[k])) << endl;
      }
  }

}


vector<Factor> GMR::calcConcorrel( const InfAlg & obj, const VarSet& ns, bool reInit ){
    // convert ns to vector<VarSet>
    size_t N = ns.size();
    vector<Var> vns;
    vns.reserve( N );
    for( VarSet::const_iterator n = ns.begin(); n != ns.end(); n++ )
        vns.push_back( *n );

    vector<Factor> cavbeliefs;
    cavbeliefs.reserve(N);
    for( size_t j = 0; j < N; j++ )
      cavbeliefs.push_back(Factor(vns[j]));

    vector<Factor> pairbeliefs;
    pairbeliefs.reserve( N * N );
    for( size_t j = 0; j < N; j++ )
        for( size_t k = 0; k < N; k++ )
            if( j == k )
                pairbeliefs.push_back(Factor());
            else
                pairbeliefs.push_back(Factor(vns[j] | vns[k]));

    InfAlg *clamped = obj.clone();
    
    clamped->init();
    clamped->run();
    for( size_t j = 0; j < N; j++ ) {
      cavbeliefs[j] = clamped->belief(vns[j]);
 
    }
    if( !reInit )
      clamped->init();
    Complex logZ0; 
 
    for( size_t j = 0; j < N; j++ ) {
  
        // clamp Var j to its possible values
        for( size_t j_val = 0; j_val < vns[j].states(); j_val++ ) {
	  	  if( obj.Verbose() >= 2 )                                 
		                   cout << j << "/" << N-1 << " (" << j_val << "/" << vns[j].states()
<< "): ";

            // save unclamped factors connected to ns
	
            clamped->saveProbs( ns );

            clamped->clamp( vns[j], j_val );
            if( reInit )
                clamped->init();
            clamped->run();

            //if( j == 0 )
            //  logZ0 = obj.logZ();
          double Z_xj = 1.0;
            if( j == 0 && j_val == 0 ) {
                logZ0 = clamped->logZ();
            } else {
                // subtract logZ0 to avoid very large numbers
                Complex Z = exp(clamped->logZ() - logZ0);
                if( fabs(imag(Z)) > 1e-5 )
		              cout << "calcPairBelief::  Warning: complex Z: " << Z << endl;
                Z_xj = real(Z);
            }

            for( size_t k = 0; k < N; k++ )
                if( k != j ) {
                    Factor b_k = clamped->belief(vns[k]);
                    for( size_t k_val = 0; k_val < vns[k].states(); k_val++ )
                        if( vns[j].label() < vns[k].label() )
                            pairbeliefs[j * N + k][j_val + (k_val * vns[j].states())] = Z_xj * b_k[k_val];
                        else
                            pairbeliefs[j * N + k][k_val + (j_val * vns[k].states())] = Z_xj * b_k[k_val];
                }
           // restore clamped factors
            clamped->undoProbs( ns );
        }
    }
    delete clamped;
    // Calculate result by taking the geometric average
    vector<Factor> result;
    result.reserve( N * (N - 1) / 2 );
    for( size_t j = 0; j < N; j++ )
      for( size_t k = j+1; k < N; k++ ){
	Factor bla=(pairbeliefs[j*N+k]*pairbeliefs[k*N+j])^0.5;
	result.push_back(bla/bla.totalSum() - cavbeliefs[j]*cavbeliefs[k]);
      }
    //    for(size_t i=0 ; i<result.size(); i++)
    //      result[i].fill(0./result[i].p().size());

    return result;
   
}


Factor GMR::belief1( size_t i ) const{
  Factor Gfac(var(i));
  Factor Gfac1(var(i),1.0);
  Factor Gfac2(var(i));
  for(_nb_cit J = nbV(i).begin(); J != nbV(i).end(); J++ )
    Gfac1.p() *= Mumes(i,*J).p();
  Gfac = Gfac1;
  for(_nb_cit J = nbV(i).begin(); J != nbV(i).end(); J++ ){
    Gfac2.p() = Gfac1.p()* Lammes(i,*J).p();
    Gfac += Gfac2.divided_by(Mumes(i,*J));
  }

  for(_nb_cit J = nbV(i).begin(); J != nbV(i).end()-1; J++ )
    for(_nb_cit K = J+1; K != nbV(i).end(); K++){
      Gfac2.p() = Gfac1.p()*Rhomes(i,*J,*K).p();
      Gfac += Gfac2.divided_by(Factor(var(i),Mumes(i,*J).p()*Mumes(i,*K).p()));
  }

  return( Gfac.marginal(var(i)) );
}

Factor GMR::belief (const Var &n) const{
  return( _beliefs[findVar( n )] );
}
 
void GMR::calcbeliefs(){
  cout << " nr of vars is " << nrVars() << endl;
  for(size_t i=0; i<nrVars(); i++){

    _beliefs[i] = belief1(i);
    cout << "belief " << i << " is " << _beliefs[i] << endl;
  }
}

void GMR::UpdatePmess( size_t ijI){
  double little = Eps();
  size_t i = _ijI[ijI].i; // i is the central variable, we are going to update
                   // P^{(j)}(x_i).
  size_t j = _ijI[ijI].j;
  size_t I = _ijI[ijI].I;
  //  cout << "(" << i << " " << j << " " << I << ")" << endl;
  Factor Gfac(var(i));
  Factor Gfac1(var(i),1.0);
  Factor Gfac2(var(i));
  Factor Gfac3(var(j),1.0);
  Factor nufac(var(i) + var(j),0.0); // this is possibly a bit dodgy
  if(nbV(i).size()>1){
    for(_nb_cit J=nbV(i).begin(); J!= nbV(i).end(); J++)
      if(*J != I)
	Gfac1.p() *= Mumes(i,*J).p();
    Gfac = Gfac1;

    for(_nb_cit J = nbV(i).begin(); J != nbV(i).end(); J++ )
      if(*J != I){
	Gfac2.p() = Gfac1.p()*Lammes(i,*J).p();
	Gfac += Gfac2.divided_by(Mumes(i,*J));
      }

    for(_nb_cit J = nbV(i).begin(); J != nbV(i).end()-1; J++ )
      for(_nb_cit K = J+1; K != nbV(i).end(); K++)
	if((*J != I)&&(*K != I)){
	  Gfac2.p() = Gfac1.p()*Rhomes(i,*J,*K).p();
	  Gfac += Gfac2.divided_by(Factor(var(i),Mumes(i,*J).p()*Mumes(i,*K).p()));
	}
    Gfac2 = Gfac.marginal(var(i));
  }
  else{
    Gfac2.fill(1.0/(Gfac2.p().size()));
  }
  if(nbV(j).size()>1){
    for(_nb_cit J = nbV(j).begin(); J != nbV(j).end(); J++)
      if(*J != I)
	Gfac3.p() *= Mumes(j,*J).p();
    for(_nb_cit J = nbV(j).begin(); J != nbV(j).end(); J++)
      if(*J != I)
	nufac += Numes(i,j,*J)*Gfac3.divided_by(Mumes(j,*J));
    double x= Gfac3.totalSum();
    Gfac3 = nufac.partSum(var(i)); 

    if(x>1e-6)
    Gfac2 -=  Gfac3/x;  
  }

  _Pmess[ijI] = _Pmess[ijI]*(1.-little) + Gfac2.marginal(var(i))*little;
  if(Verbose()>2)  cout << "P message " << ijI << " is " << Pmes(i,j) << endl;
}


void GMR::UpdatePmessNew( size_t ijI){
  double little = Eps();
  size_t i = _ijI[ijI].i; // i is the central variable, we are going to update
                   // P^{(j)}(x_i).
  size_t j = _ijI[ijI].j;
  size_t I = _ijI[ijI].I;
  //  cout << "(" << i << " " << j << " " << I << ")" << endl;
  Factor Gfac(var(i));
  Factor Gfac1(var(i),1.0);
  Factor Gfac2(var(i));
  Factor Gfac3(var(j),1.0);
  Factor nufac(var(i)+var(j),0.0); // this is possibly a bit dodgy
  if(nbV(i).size()>1){
    for(_nb_cit J=nbV(i).begin(); J!= nbV(i).end(); J++)
      if(*J != I){
	UpdateMumes(VV2E(i,*J));
	Gfac1.p() *= Mumes(i,*J).p();
      }
    Gfac = Gfac1;

    for(_nb_cit J = nbV(i).begin(); J != nbV(i).end(); J++ )
      if(*J != I){
	UpdateLammes(VV2E(i,*J));
	Gfac2.p() = Gfac1.p()*Lammes(i,*J).p();
	Gfac += Gfac2.divided_by(Mumes(i,*J));
      }

    for(_nb_cit J = nbV(i).begin(); J != nbV(i).end()-1; J++ )
      for(_nb_cit K = J+1; K != nbV(i).end(); K++)
	if((*J != I)&&(*K != I)){
	  UpdateRhomes(_EiIJ[VV2E(i,*J)][*K]);
	  Gfac2.p() = Gfac1.p()*Rhomes(i,*J,*K).p();
	  Gfac += Gfac2.divided_by(Factor(var(i),Mumes(i,*J).p()*Mumes(i,*K).p()));
	}
    Gfac2 = Gfac.marginal(var(i));
  }
  else{
    Gfac2.fill(1.0/(Gfac2.p().size()));
  }
  if(nbV(j).size()>1){
    for(_nb_cit J = nbV(j).begin(); J != nbV(j).end(); J++)
      if(*J != I){
	UpdateMumes(VV2E(j,*J));
	Gfac3.p() *= Mumes(j,*J).p();
      }
    for(_nb_cit J = nbV(j).begin(); J != nbV(j).end(); J++)
      if(*J != I){
	UpdateNumes(_EijJ[_EijI[i][j]][*J]);
	nufac += Numes(i,j,*J)*Gfac3.divided_by(Mumes(j,*J));
      }
    double x= Gfac3.totalSum();
    Gfac3 = nufac.partSum(var(i)); 

    if(x>1e-6)
    Gfac2 -=  Gfac3/x;  
  }

  _Pmess[ijI] = _Pmess[ijI]*(1.-little) + Gfac2.marginal(var(i))*little;
  if(Verbose()>2)  cout << "P message " << ijI << " is " << Pmes(i,j) << endl;
}
string GMR::identify() const {
    stringstream result (stringstream::out);
    result << Name << GetProperties();
    return result.str();
}



// then propagate updates until convergence, looking very much like 
// bp stuff

// The classes to inherit from are at least Factor (better: Daialg) 
// and perhaps lcbp

// We need a number of private functions:

// 1. the P^{(i)}(x_j), which will be something like messages, but they will
// not be from factor to node, since they explicitly go from node to node.

// 2. the mu_{\alpha \to i}(x_i), which will be the same as the messages

// 3. the \lambda_{\alpha \to i}(x_i), having same representation as the
// messages

// 4. the \rho_{(\alpha,\beta)\to i}(x_i) messages, that have quite a 
// difficult representation, since they involve more than one neighbouring
// factor if i

// 5. the \nu_{\alpha \to i}(x_i,x_j) which depends on factor alpha and 
// node j, both of which are in the cavity of i.


// Todo: make constructors, add comments
//       optimize speed (perhaps introduce phi functions)
//       reserve some more memory here and there
