/*  Copyright (C) 2006  Joris Mooij  [j dot mooij at science dot ru dot nl]
    Radboud University Nijmegen, The Netherlands

    This file is part of libDAI.

    libDAI is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    libDAI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with libDAI; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


#ifndef __LCBP3_H__
#define __LCBP3_H__


#include "daialg.h"
#include "enum.h"
#include "factorgraph.h"


using namespace std;


class LCBP3 : public DAIAlgFG {
    protected:
        vector<Factor>      _pancakes;
        vector<Factor>      _cavitydists;
        vector<Factor>      _phis;

        /// Single variable beliefs
        vector<Factor>      _beliefs;

    public:
        ENUM(TypeType,PLUS,MINUS,STAR,HOIMINUS,HOIPLUS)
        ENUM(CavityType,FULL,PAIR,UNIFORM,RANDOM)
        ENUM(UpdateType,SEQFIX,SEQRND,NONE)

        struct {
            TypeType   type;
            UpdateType updates;
            CavityType cavity;
            bool       reinit;
        } _properties;
        TypeType Type() const { return _properties.type; }
        CavityType Cavity() const { return _properties.cavity; }
        UpdateType Updates() const { return _properties.updates; }
        bool reInit() const { return _properties.reinit; }
        
        /// Default constructor
        LCBP3() : DAIAlgFG() {};
        /// Copy constructor
        LCBP3(const LCBP3 & x) : DAIAlgFG(x), _pancakes(x._pancakes), _cavitydists(x._cavitydists), _phis(x._phis), _beliefs(x._beliefs) {};
        /// Clone function
        LCBP3* clone() const { return new LCBP3(*this); }
        /// Create function
        LCBP3* create() const { return new LCBP3(); }
        /// Construct LCBP3 object from a FactorGraph and parameters
        LCBP3(const FactorGraph & fg, const Properties &opts);
        /// Assignment operator
        LCBP3& operator=(const LCBP3 & x) {
            if( this != &x ) {
                DAIAlgFG::operator=(x);
                _pancakes       = x._pancakes;
                _cavitydists    = x._cavitydists;
                _phis           = x._phis;
                _beliefs        = x._beliefs;
            }
            return *this;
        }

        static const char *Name;
        double CalcCavityDist (size_t I, const string &name, const Properties &opts);
        double InitCavityDists (const string &name, const Properties &opts);

        void init();
        Factor NewPancake (size_t iI, bool & hasNaNs);
        Factor Pancake (size_t I);
        double run();

        string identify() const;
        Factor belief (const Var &n) const;
        Factor belief (const VarSet &ns) const { assert( 0 == 1 ); }
        vector<Factor> beliefs() const;
        Complex logZ() const { return NAN; }
        void CalcBelief (size_t i);
        const Factor & belief (size_t i) const { return _beliefs[i]; };
        const Factor & pancake (size_t i) const { return _pancakes[i]; };
        const Factor & cavitydist (size_t i) const { return _cavitydists[i]; };

        void clamp( const Var &n, size_t i ) { assert( 0 == 1 ); }
        void undoProbs( const VarSet &ns ) { assert( 0 == 1 ); }
        void saveProbs( const VarSet &ns ) { assert( 0 == 1 ); }
        void makeFactorCavity(size_t I) { assert( 0 == 1 ); }
        virtual void makeCavity(const Var & n) { assert( 0 == 1 ); }
        bool checkProperties();
};


#endif
