extern "C" {
#include "glpk.h"
}

#include <iostream>
#include <fstream>
#include "factor.h"
#include "box.h"
#include "bp.h"
#include <CImg.h>
#include <cmath>

using namespace std;
using namespace libDAI;
using namespace cimg_library;

size_t nr_states = 2;

double potstrength( const Factor & psi ) {
    double strength = 0.0;
    for( size_t a = 0; a < nr_states; a++ )
        for( size_t b = 0; b < nr_states; b++ )
            for( size_t c = 0; c < nr_states; c++ )
                for( size_t d = 0; d < nr_states; d++ ) {
                    double x = (psi[a*nr_states+b] * psi[c*nr_states+d]) / (psi[a*nr_states+d] * psi[c*nr_states+b]);
                    if( x > strength )
                        strength = x;
                }
    return pow(strength,0.25);
}

double potstrength_old( const Factor & psi ) {
    double strength = 0.0;
    for( size_t a = 0; a < nr_states * nr_states; a++ )
        for( size_t b = 0; b < nr_states * nr_states; b++ ) {
            double x = psi[a] / psi[b];
            if( x > strength )
                strength = x;
        }
    return pow(strength,0.5);
}

double Ihlerrec( const Factor & psi, double delta ) {
    double E = pow(potstrength(psi),2.0);
    return (E * delta + 1.0) / (E + delta);
}

double Ihlerdist( const Factor & psi1, const Factor & psi2 ) {
    double dist = 0.0;
    for( size_t a = 0; a < psi1.stateSpace(); a++ )
        for( size_t b = 0; b < psi2.stateSpace(); b++ ) {
            double x = (psi1[a] / psi2[a]) / (psi1[b] / psi2[b]);
            if( x > dist )
                dist = x;
        }
    return pow(dist,0.5);
}

double theta( const Factor & psi ) {
    return 0.5 * log( psi[1] / psi[0] );
}

double J( const Factor & psi ) {
    return 0.25 * (log(psi[3]) + log(psi[0]) - log(psi[2]) - log(psi[1]));
}

int main() {

/*
{
    Var x0(0,nr_states);
    Var x1(1,nr_states);
    double psival[4] = {1.0, 1.0, 2.0, 1.0};
    double psi1val[2] = {1.0/3.0,2.0/3.0};
    double psi1bval[2] = {2.0/3.0,1.0/3.0};
    Factor psi01(x0|x1,psival);
    Factor psi1(x1,psi1val);
    Factor psi1b(x1,psi1bval);
    vector<Prob> psi1s; psi1s.push_back(psi1.p()); psi1s.push_back(psi1b.p());

    cout << theta(psi1) << ", " << theta(psi1b) << endl;
    cout << J(psi01) << endl;
    cout << exp(fabs(theta(psi1) - theta(psi1b))) << " = " << Ihlerdist(psi1, psi1b) << endl;
    cout << tanh(log(potstrength_old(psi01))) << " = " << tanh(log(potstrength(psi01))) << " = " << psi01.strength(x0,x1) << " = " << abs(tanh(J(psi01))) << endl;

    double r = Ihlerdist( psi1, psi1b );
    cout << r << endl;
    Factor psi0 = (psi1 * psi01).marginal(x0);
    Factor psi0b = (psi1b * psi01).marginal(x0);
    cout << psi0 << endl;
    cout << psi0b << endl;
    cout << Ihlerdist( psi0, psi0b ) << endl;
    double R = Ihlerrec( psi01, Ihlerdist( psi1, psi1b ) );
    cout << R << endl;
    cout << psi0.divided_by(psi0 * (1.0-R*R) + (R*R)) << " <= " << psi0b << " <= " << (psi0.divided_by(psi0 * (R*R-1.0) + 1.0) * (R*R)) << endl;
    cout << psi0b.divided_by(psi0b * (1.0-R*R) + (R*R)) << " <= " << psi0 << " <= " << (psi0b.divided_by(psi0b * (R*R-1.0) + 1.0) * (R*R)) << endl;
    cout << endl;

    Box b1 (x1, psi1s);
    cout << b1 << endl;
    Box b0 = b1.encloseNormedSumProd(psi01,x0);
    cout << b0 << endl;
    vector<Prob> ep_b0 = b0.extremePoints();
    for( size_t pa = 0; pa < ep_b0.size(); pa++ )
        for( size_t pb = 0; pb < ep_b0.size(); pb++ )
            cout << "d(" << pa << ", " << pb << ") = " << Ihlerdist( Factor(x0,ep_b0[pa]), Factor(x0,ep_b0[pb]) ) << endl;
    cout << Ihlerdist( Factor(x0,b0.max), Factor(x0,b0.min) ) << endl;
    cout << (psi1 * psi01).marginal(x0) << endl;
    cout << (psi1b * psi01).marginal(x0) << endl;
    double mes1 = atanh(tanh(J(psi01)) * tanh(theta(psi1)));
    double mes1b = atanh(tanh(J(psi01)) * tanh(theta(psi1b)));
    cout << exp(-mes1)/(2*cosh(mes1)) << "," << exp(mes1)/(2*cosh(mes1)) << endl;
    cout << exp(-mes1b)/(2*cosh(mes1b)) << "," << exp(mes1b)/(2*cosh(mes1b)) << endl;
}
    return 1;
*/

    Var x0(0,nr_states);
    Var x1(1,nr_states);
    Var x2(3,nr_states);
    Var x1b(2,nr_states);
    Var x2b(4,nr_states);


    double psival[4] = {1.0, 2.0, 2.0, 1.0};
    Factor psi01(x0|x1,psival);
    Factor psi02(x0|x2,psival);
    Factor psi12(x1|x2,psival);
/*    Factor psi01(x0|x1);
    Factor psi02(x0|x2);
    Factor psi12(x1|x2);*/


    vector<double> BPerror;
    vector<double> easyBPerror;

    size_t dimx = 800;
    size_t dimy = 600;
    CImg<unsigned char> image(dimx,dimy,1,3);

    for( size_t iter = 0; iter < 100; iter++ ) {

        psi01.randomize();
        psi02.randomize();
        psi12.randomize();

        cout << tanh(log(potstrength(psi01))) << " = " << psi01.strength(x0,x1) << " =? " << (Ihlerrec(psi01,1.001) - 1.0) / 0.001 << endl;

        Factor psi = psi01 * psi02;
        Var dest = x0;

    //    Box box1(x1, psi12.marginal(x1).P(), psi12.marginal(x1).P());
    //    Box box2(x2, psi12.marginal(x2).P(), psi12.marginal(x2).P());
        Factor psi12_marg1 = psi12.marginal(x1);
        Factor psi12_marg2 = psi12.marginal(x2);

        cout << "  psi12_marg1 = " << psi12_marg1 << ", psi12_marg2 = " << psi12_marg2 << endl;

        Box bound(dest);
        bool failed = false;
        for( size_t direction = 0; direction < 2; direction++ ) {
            for( size_t x_dest = 0; x_dest < dest.states(); x_dest++ ) {
                // for each state of dest
                Factor psi_slice = psi.slice( dest, x_dest );

                LPX *lp = lpx_create_prob();

                if( direction == 0 )
                    lpx_set_obj_dir(lp, LPX_MIN);
                else
                    lpx_set_obj_dir(lp, LPX_MAX);

                size_t n = psi_slice.stateSpace();           // number of columns (structural variables <-> probability distribution on psi_slice)
                size_t m = x1.states() + x2.states();        // number of rows (auxiliary variables <-> constraints imposed by known marginals)

                lpx_add_cols(lp,n);
                lpx_add_rows(lp,m);

                // set objective function and col bounds
                lpx_set_obj_coef(lp,0,0.0);
                for( size_t N = 0; N < n; N++ ) {
                    lpx_set_obj_coef(lp,N+1,psi_slice[N]);
                    lpx_set_col_bnds(lp,N+1,LPX_LO,0.0,1.0);
                }

                multind mi(x1|x2);
                // set row bounds and constraint matrix rows corresponding to the known marginal of x1
                for( size_t i = 0; i < x1.states(); i++ ) {
                    lpx_set_row_bnds(lp,i+1,LPX_FX,psi12_marg1[i],psi12_marg1[i]);
//                    lpx_set_row_bnds(lp,i+1,LPX_DB,0.0,psi12_marg1[i]);

                    vector<int> ind;
                    vector<double> val;
                    ind.push_back(0); val.push_back(0.0);
                    for( size_t j = 0; j < mi.max(); j++ ) {
                        std::vector<size_t> vi = mi.vi(j);
                        if( vi[0] == i ) {
                            ind.push_back( j+1 );
                            val.push_back( 1.0 );
                        }
                    }
                    lpx_set_mat_row(lp,i+1,ind.size()-1,&(ind[0]),&(val[0]));
                }

                // set row bounds and constraint matrix rows corresponding to the known marginal of x2
                for( size_t i = 0; i < x2.states(); i++ ) {
                    lpx_set_row_bnds(lp,x1.states()+i+1,LPX_FX,psi12_marg2[i],psi12_marg2[i]);
//                    lpx_set_row_bnds(lp,x1.states()+i+1,LPX_DB,0.0,psi12_marg2[i]);

                    vector<int> ind;
                    vector<double> val;
                    ind.push_back(0); val.push_back(0.0);
                    for( size_t j = 0; j < mi.max(); j++ ) {
                        std::vector<size_t> vi = mi.vi(j);
                        if( vi[1] == i ) {
                            ind.push_back( j+1 );
                            val.push_back( 1.0 );
                        }
                    }
                    lpx_set_mat_row(lp,x1.states()+i+1,ind.size()-1,&(ind[0]),&(val[0]));
                }

                double result;
                if( 1 ) {
                    lpx_set_int_parm(lp, LPX_K_PRESOL, 1);      // use pre-solver
                    lpx_set_int_parm(lp, LPX_K_MSGLEV, 0);      // quiet
                    cout << "Simplex result: " << lpx_simplex(lp) << endl;
                    result = lpx_get_obj_val(lp);
                    if( lpx_get_status(lp) != LPX_OPT )
                        failed = true;
                } else {
                    lpx_set_int_parm(lp, LPX_K_MSGLEV, 0);      // quiet
                    cout << "Interior point method result: " << lpx_interior(lp) << endl;
                    result = lpx_ipt_obj_val(lp);
                    if( lpx_ipt_status(lp) != LPX_T_OPT )
                        failed = true;
                }

                cout << "  that is:      " << result << endl;

                if( direction == 0 )
                    bound.min[x_dest] = result;
                else
                    bound.max[x_dest] = result;

                lpx_delete_prob(lp);
            }
        }


        // Real marginal
        Factor marg_x0 = (psi01 * psi02 * psi12).marginal(x0);
        cout << endl;
        cout << "Real marginal:   " << marg_x0 << endl;
        cout << endl;


        // LP bound
        Box b_0 = bound.normalized();
        cout << "LP bound:        " << b_0 << "(success = " << !failed << ")" << endl;
/*        if( !failed ) */
        assert( !failed );
        assert( b_0.contains( marg_x0, 1e-9 ) );
        cout << endl;

        for( size_t x = size_t(50 + b_0.min[0] * 700); x < size_t(50 + b_0.max[0] * 700); x++ )
            for( size_t y = size_t(50 + iter * 5 - 1); y <= size_t(50 + iter * 5 + 1); y++ )
                image(x, y, 0) = 255;


        // Example I
        Box b_1(x1);
        Box b_2(x2);
        Box b_10 = b_1.encloseNormedSumProd(psi01, x0);
        Box b_20 = b_2.encloseNormedSumProd(psi02, x0);
        b_0 = (b_10 * b_20).normalized();
        cout << "Example I bound: " << b_0 << endl;
        assert( b_0.contains( marg_x0, 1e-9 ) );


        for( size_t x = size_t(50 + b_0.min[0] * 700); x < size_t(50 + b_0.max[0] * 700); x++ )
            for( size_t y = size_t(50 + iter * 5 - 1); y <= size_t(50 + iter * 5 + 1); y++ )
                image(x, y, 2) = 100;


        // Example II
        Factor psi1b2(x1b|x2,&(psi12[0]));
        Box b_1b(x1b);
        Box b_1b_2 = b_1b.encloseNormedSumProd(psi1b2, x2);
        b_20 = b_1b_2.encloseNormedSumProd(psi02, x0);
        b_0 = (b_10 * b_20).normalized();
        cout << "Example II bound " << b_0 << endl;
        assert( b_0.contains( marg_x0, 1e-9 ) );


        for( size_t x = size_t(50 + b_0.min[0] * 700); x < size_t(50 + b_0.max[0] * 700); x++ )
            for( size_t y = size_t(50 + iter * 5 - 1); y <= size_t(50 + iter * 5 + 1); y++ )
                image(x, y, 2) = 150;


        // Example III (does this work?)
        Factor psi12b(x1|x2b,&(psi12[0]));
        Box b_2b(x2b);
        Box b_2b_1 = b_2b.encloseNormedSumProd(psi12b, x1);
        b_10 = b_2b_1.encloseNormedSumProd(psi01, x0);
        b_0 = (b_10 * b_20).normalized();
        cout << "Example III bound" << b_0 << endl;
        assert( b_0.contains( marg_x0, 1e-9 ) );


        for( size_t x = size_t(50 + b_0.min[0] * 700); x < size_t(50 + b_0.max[0] * 700); x++ )
            for( size_t y = size_t(50 + iter * 5 - 1); y <= size_t(50 + iter * 5 + 1); y++ )
                image(x, y, 2) = 255;


        // BP marginal
        vector<Factor> factors;
        factors.push_back( psi01 );
        factors.push_back( psi02 );
        factors.push_back( psi12 );
        FactorGraph fg(factors);

        BP bp(fg, Properties()("maxiter",string("100UL"))("tol",string("1e-9"))("verbose",string("0UL"))("updates",string("SEQFIX")));
        bp.init();
        bp.run();
        cout << endl;
        cout << "BP marginal:     " << bp.belief(x0) << endl;

        for( size_t y = 50 + iter * 5 - 1; y <= 50 + iter * 5 + 1; y++ ) {
            image( 50 + int(bp.belief(x0)[0] * 700), y, 0 ) = 255;
            image( 50 + int(bp.belief(x0)[0] * 700), y, 2 ) = 255;
        }

        for( size_t y = 50 + iter * 5 - 1; y <= 50 + iter * 5 + 1; y++ ) {
            image( 50 + int(marg_x0[0] * 700), y, 1 ) = 255;
            image( 50 + int(marg_x0[0] * 700), y, 2 ) = 255;
        }


        // Ihler's bound
        double delta_2b = exp(pow(potstrength(psi02),2.0));
        double delta_1b = exp(pow(potstrength(psi01),2.0));
        double delta_1  = Ihlerrec(psi12,delta_2b);
        double delta_2  = Ihlerrec(psi12,delta_1b);
        double delta_0  = Ihlerrec(psi01,delta_1) * Ihlerrec(psi02,delta_2);
        Box Ihler(x0);
        for( size_t state = 0; state < nr_states; state++ ) {
            Ihler.min[state] = bp.belief(x0)[state] / (delta_0*delta_0 + (1.0-delta_0*delta_0)*bp.belief(x0)[state]);
            Ihler.max[state] = delta_0*delta_0 * bp.belief(x0)[state] / (1.0 - (1.0-delta_0*delta_0)*bp.belief(x0)[state]);
        }
        assert( Ihler.contains( bp.belief(x0), 1e-9 ) );
        assert( Ihler.contains( marg_x0, 1e-9 ) );
        cout << "Ihler bound:     " << Ihler << endl;
        for( size_t x = size_t(50 + Ihler.min[0] * 700); x < size_t(50 + Ihler.max[0] * 700); x++ )
            for( size_t y = size_t(50 + iter * 5 + 2); y <= size_t(50 + iter * 5 + 2); y++ )
                image(x, y, 1) = 255;
        

        // Uniform messages instead of BP
//        Factor easyBP_x0 = (psi01.marginal(x0) * (psi12.marginal(x2) * psi02).marginal(x0)).normalized();
        Factor easyBP_x0 = ((psi12.marginal(x1) * psi01).marginal(x0) * (psi12.marginal(x2) * psi02).marginal(x0)).normalized();
        cout << "easyBP marginal: " << easyBP_x0 << endl;
        cout << endl;

        for( size_t y = 50 + iter * 5 - 1; y <= 50 + iter * 5 + 1; y++ ) {
            image( 50 + int(easyBP_x0[0] * 700), y, 0 ) = 255;
            image( 50 + int(easyBP_x0[0] * 700), y, 1 ) = 255;
            image( 50 + int(easyBP_x0[0] * 700), y, 2 ) = 0;
        }
        
        // Self avoiding walks
        for( size_t state_x0b = 0; state_x0b < x0.states(); state_x0b++ )
            for( size_t state_x0c = 0; state_x0c < x0.states(); state_x0c++ ) {
                Factor lalala = (((psi02.slice(x0,state_x0b) * psi12).marginal(x1) * psi01).marginal(x0) * ((psi01.slice(x0,state_x0c) * psi12).marginal(x2) * psi02).marginal(x0)).normalized();
                cout << "SAW " << state_x0b << state_x0c << ":          " << lalala << endl;
                image( 50 + int(lalala[0] * 700), 50 + iter*5, 1 ) = 255;
            }

        // Next one...
        cout << "*************************************************************************************************************************************" << endl;

        BPerror.push_back( bp.belief(x0)[0] - marg_x0[0] );
        easyBPerror.push_back( easyBP_x0[0] - marg_x0[0] );
    }

    CImgDisplay main_disp(image,"Image",0);

//  image4.save_jpeg(file_p,100);

    while (!main_disp.is_closed)
        cimg::wait(40);

    ofstream outfile;
    outfile.open ("exp.dat");
    for( size_t iter = 0; iter < BPerror.size(); iter++ ) {
        outfile << iter << "\t" << BPerror[iter] << "\t" << easyBPerror[iter] << endl;
    }
    outfile.close();

};
