/*  Copyright (C) 2006,2007  Joris Mooij  [joris at jorismooij dot nl]
    Radboud University Nijmegen, The Netherlands
    
    This file is part of libDAI.

    libDAI is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    libDAI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with libDAI; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


#include <iostream>
#include "alldai.h"
#include "bounds.h"


using namespace std;
using namespace libDAI;


Factor BinaryFactor( const Var &n, double field ) {
    assert( n.states() == 2 );
    double buf[2];
    buf[0] = exp(-field); 
    buf[1] = exp(field);
    return Factor(n, &buf[0]);
}

Factor BinaryFactor( const Var &n1, const Var &n2, double coupling ) {
    assert( n1.states() == 2 );
    assert( n2.states() == 2 );
    assert( n1 != n2 );
    double buf[4];
    buf[0] = (buf[3] = exp(coupling));
    buf[1] = (buf[2] = exp(-coupling));
    return Factor(n1 | n2, &buf[0]);
}

Factor BinaryFactor( const Var &n1, const Var &n2, const Var &n3, double coupling ) {
    assert( n1.states() == 2 );
    assert( n2.states() == 2 );
    assert( n3.states() == 2 );
    assert( n1 != n2 );
    assert( n1 != n3 );
    assert( n2 != n3 );
    double buf[8];
    buf[0] = (buf[3] = exp(-coupling));
    buf[1] = (buf[2] = exp(coupling));
    buf[4] = (buf[7] = exp(coupling));
    buf[5] = (buf[6] = exp(-coupling));
    return Factor(n1 | n2 | n3, &buf[0]);
}

vector<Factor> triple2pair( const Factor & A ) {
    assert( A.vars().size() == 3 );
    Prob logP = A.p().log();
    double J0   = ( logP[0] + logP[1] + logP[2] + logP[3] + logP[4] + logP[5] + logP[6] + logP[7]) / 8.0;
    double J1   = (-logP[0] + logP[1] - logP[2] + logP[3] - logP[4] + logP[5] - logP[6] + logP[7]) / 8.0;
    double J2   = (-logP[0] - logP[1] + logP[2] + logP[3] - logP[4] - logP[5] + logP[6] + logP[7]) / 8.0;
    double J3   = (-logP[0] - logP[1] - logP[2] - logP[3] + logP[4] + logP[5] + logP[6] + logP[7]) / 8.0;
    double J12  = ( logP[0] - logP[1] - logP[2] + logP[3] + logP[4] - logP[5] - logP[6] + logP[7]) / 8.0;
    double J13  = ( logP[0] - logP[1] + logP[2] - logP[3] - logP[4] + logP[5] - logP[6] + logP[7]) / 8.0;
    double J23  = ( logP[0] + logP[1] - logP[2] - logP[3] - logP[4] - logP[5] + logP[6] + logP[7]) / 8.0;
    double J123 = (-logP[0] + logP[1] + logP[2] - logP[3] + logP[4] - logP[5] - logP[6] + logP[7]) / 8.0;

/*    size_t i = 0;
    for( int s3 = -1; s3 <= 1; s3 += 2 )
        for( int s2 = -1; s2 <= 1; s2 += 2 )
            for( int s1 = -1; s1 <= 1; s1 += 2 ) {
                cout << s1 << " " << s2 << " " << s3 << "  " << logP[i] << "  " << (J0 + J1*s1 + J2*s2 + J3*s3 + J12*s1*s2 + J13*s1*s3 + J23*s2*s3 + J123*s1*s2*s3) << endl;
                i++;
            }
*/

    VarSet::const_iterator s = A.vars().begin();
    Var s1 = *s;
    s++;
    Var s2 = *s;
    s++;
    Var s3 = *s;
    Var s0 = Var( 0, 2 );

    vector<Factor> result;
//    result.push_back( BinaryFactor( s1, J1 ) );
//    result.push_back( BinaryFactor( s2, J2 ) );
//    result.push_back( BinaryFactor( s3, J3 ) );
//    result.push_back( BinaryFactor( s1, s2, J12 ) );
//    result.push_back( BinaryFactor( s1, s3, J13 ) );
//    result.push_back( BinaryFactor( s2, s3, J23 ) );
//    result.push_back( BinaryFactor( s1, s2, s3, J123 ) );

    double b = -J123 / (4.0 * fabs(J123)) * acosh( exp( 4 * fabs(J123) ) );
    double a = -J123 / (4.0 * fabs(J123)) * acosh( exp(  8* fabs(b) ) );
    result.push_back( BinaryFactor( s0, fabs(a) ) );
    result.push_back( BinaryFactor( s0, s1, a ) );
    result.push_back( BinaryFactor( s0, s2, a ) );
    result.push_back( BinaryFactor( s0, s3, a ) );
    result.push_back( BinaryFactor( s1, J1-b ) );
    result.push_back( BinaryFactor( s2, J2-b ) );
    result.push_back( BinaryFactor( s3, J3-b ) );
    result.push_back( BinaryFactor( s1, s2, J12 - fabs(b) ) );
    result.push_back( BinaryFactor( s1, s3, J13 - fabs(b) ) );
    result.push_back( BinaryFactor( s2, s3, J23 - fabs(b) ) );

    Factor prod;
    for( size_t I = 0; I < result.size(); I++ ) {
//        cout << result[I] << " " << result[I].p().log() << endl;
        prod *= result[I];
    }
    prod *= exp(J0) * cosh(J123) / (cosh(2.0*a) + exp(2*fabs(b)));

    cout << prod.partSum(s1|s2|s3) / A << endl;
    
    return result;
}


int main( int argc, char *argv[] ) {
    Var x1(1,2);
    Var x2(2,2);
    Var x3(3,2);
    Var x4(4,2);
    Factor f12(x1|x2);
    Factor f13(x1|x3);
    Factor f23(x2|x3);
    Factor f24(x2|x4);
    Factor f34(x3|x4);
    Factor A(x1|x2|x3);
    Factor B(x2|x3|x4);

    Var x0(0,2);
    Factor f10(x1|x0);
    Factor f20(x2|x0);
    Factor f30(x3|x0);

/*    double k = 1.0;
    double b = -k / (4.0 * fabs(k)) * acosh( exp( 4 * fabs(k) ) );
    double a = -k / (4.0 * fabs(k)) * acosh( exp(  8* fabs(b) ) );
    for( int s1 = -1; s1 <= 1; s1 += 2 )
        for( int s2 = -1; s2 <= 1; s2 += 2 )
            for( int s3 = -1; s3 <= 1; s3 += 2 ) {
                double ks1s2s3 = k * s1 * s2 * s3;
                double bla = log(cosh(k)) - log(cosh(2.0*a)+exp(2*fabs(b))) - b * (s1 + s2 + s3) - fabs(b) * (s1*s2 + s1*s3 + s2*s3) + log( 2.0 * cosh(-(fabs(a) + a*(s1 + s2 + s3))) );
                cout << s1 << " " << s2 << " " << s3 << "  " << ks1s2s3 << "  " << bla << endl;
            }

    return 0;
*/

    size_t nriter = 100000;
    for( size_t iter = 0; iter < nriter; iter++ ) {
/*        A.randomize();
        B.randomize();*/
        f12.randomize();
        f13.randomize();
        f23.randomize();
        f24.randomize();
        f34.randomize();
//        A = f12 * f13 * f23;
        A.randomize();
        f10.randomize();
        f20.randomize();
        f30.randomize();
//        A = ((f10 * f20 * f30 * f12 * f13 * f23)^2.0).partSum(x1|x2|x3);
//        B = f24 * f34;

        vector<Factor> factors;
        factors.push_back( A );
//        factors.push_back( B );
//        factors.push_back( f12 );
//        factors.push_back( f13 );
        factors.push_back( f24 );
        factors.push_back( f34 );
        FactorGraph fg(factors);

        JTree jt( fg, Properties()("verbose",(size_t)0)("updates",string("HUGIN")) );
        jt.init();
        jt.run();

        Box box;
        if( false ) {
            BoundTree tr;
            tr.buildTree( fg, x1, 1000000, true );
            size_t verbose = 0;
            box = tr.BoxProp( fg, false, verbose );
        } else {
            Box b3(x3);
            Box b34 = b3.boundSumProd(f34, x4, true);
            Box b42 = b34.boundSumProd(f24, x2, true);

            Box b2(x2);
            Box b24 = b2.boundSumProd(f24, x4, true);
            Box b43 = b24.boundSumProd(f34, x3, true);

            vector<Box> boxes;
            boxes.push_back( b42 );
            boxes.push_back( b43 );
            Box b1 = boundSumProd(A, boxes, x1, false, true);

            cout << b1 << endl;
            Box b23 = boundProd( boxes );
            if( !b23.contains( (f24*f34).marginal(x2|x3), 1e-9 ) )
                cout << "OOPS" << endl;
            box = b23.boundSumProd( A, x1, true );
            cout << box << endl;

            box = b1;
        }

        cout << box.contains( jt.belief(x1), 1e-9 ) << endl;
        if( !box.contains( jt.belief(x1), 1e-5 ) ) {
/*            vector<Factor> equiv = triple2pair( A );
            equiv.push_back( f24 );
            equiv.push_back( f34 ); 
            FactorGraph fg_equiv(equiv);

            fg_equiv.WriteToFile( "/tmp/bug_equiv.fg" );
            fg.WriteToFile( "/tmp/bug.fg" );
            Box box = tr.BoxProp( fg, false, 1 );*/
            cout << box << endl;
            cout << jt.belief(x1) << endl;
            //tr.display( cout, fg );
            break;
        }
//        cout << box << endl;
//        cout << jt.belief(x1) << endl;

//        if( iter == (nriter-1) )
//            tr.display( cout, fg );
    }

    return 0;
}
