#include <iostream>
#include "enum.h"

using namespace std;


ENUM3(color,red,green,blue)


int main() {
        color b(color::red);
        color c(color::blue);
        color d("green");
        cout << (const char *)b << endl;
        cout << (char const *)c << endl;
        cout << (string)d << endl;
        cout << (int)d << endl;

        return 0;
}


// file      : alt2.cpp
// author    : Boris Kolpackov <boris@kolpackov.net>
// copyright : Copyright (c) 2003 Boris Kolpackov
// license   : http://kolpackov.net/license.html


class Color
{
public:
  enum Value
  {
    red, green, blue
  };

  Color (Value v)
      : v_ (v)
  {
  }

  operator Value () const
  {
    return v_;
  }

  operator char const* () const;

private:
  Value v_;
};

#include <iostream>

namespace
{
  char const* color_labels_[] = {"red", "green", "blue"};
}

Color::
operator char const* () const
{
  return color_labels_[v_];
}

std::ostream&
operator<< (std::ostream& o, Color c)
{
  char const* str = c;
  return o << str;
}

using std::cerr;
using std::endl;

void f (Color)
{
  cerr << "color" << endl;
}

void f (int)
{
  cerr << "int" << endl;
}

int
main ()
{
  // 1
  {
    // Color c (red);   // error: red undeclared
    Color c (Color::red);
  }

  // 2
  {
    // Color c;         // error: no default c-tor
    Color c (Color::red);
    c = Color::blue;
  }

  // 3
  {
    Color c (Color::red);
    cerr << c << " " << Color::green << endl; // prints "red 1"!
  }

  // 4
  {
    Color c (Color::red);

    f (c);
    f (Color::red); // calls f (int)

    // c + Color::red;  // ambigous operator+ overloading

    if (c) {}        // ok

    if (c == Color::red && c != Color::blue)
    {
      cerr << "ok" << endl;
    }
  }

  // 5
  {
    Color c (Color::red);

    switch (c)
    {
    case Color::red:
      cerr << "r" << endl;
      break;

    case Color::green:
      cerr << "g" << endl;
      break;

    case Color::blue:
      cerr << "b" << endl;
      break;
    }
  }
}

/*

This software or documentation is provided 'as-is', without any express or
implied warranty. In no event will the author or contributors be held liable
for any damages arising from the use of this software or documentation.

Permission is granted to anyone to use this software or documentation for any
purpose, including commercial applications, and to alter it and redistribute it
freely, subject to the following restrictions:

    1. The origin of this software or documentation must not be misrepresented;
    you must not claim that you wrote the original software or documentation.
    If you use this software or documentation in a product, an acknowledgment
    in the product documentation would be appreciated but is not required.

    2. Altered source versions must be plainly marked as such, and must not be
    misrepresented as being the original software or documentation. Altered
    source versions may be contributed back to the author to be integrated into
    the original software or documentation at the author's discretion.

    3. Neither this notice or reference to it nor any copyright, author or
    contributor clause may be removed from or altered in any software source
    distribution or documentation covered by this license.

    4. Neither the name of the author, nor the names of contributors may be
    used to endorse or promote products derived from this software or
    documentation without specific prior written permission.
*/
