#include <boost/any.hpp>
#include <string>
#include <list>
#include <map>
#include <iostream>
#include "properties.h"


// METHOD 1

class OProperty {
	public:
    	std::string name;
    	boost::any value;

	public:
		OProperty();
		OProperty(const std::string & n, const boost::any & v) : name(n), value(v) {};

};

std::ostream& operator<< (std::ostream & os, const OProperty& p) {
	os << "(" << p.name << ": ";
	if( p.value.type() == typeid(int) )
		os << boost::any_cast<int>(p.value);
	else if( p.value.type() == typeid(std::string) )
		os << boost::any_cast<std::string>(p.value);
	os << ")";
	return( os );
}

typedef std::list<OProperty> properties;


// METHOD 2

typedef std::pair<std::string, boost::any> property;
typedef std::map<std::string, boost::any>  property_map;

std::ostream& operator<< (std::ostream & os, const property & p) {
	os << "(" << p.first << ": ";
	if( p.second.type() == typeid(int) )
		os << boost::any_cast<int>(p.second);
	else if( p.second.type() == typeid(std::string) )
		os << boost::any_cast<std::string>(p.second);
	os << ")";
	return( os );
}


// METHOD 3

void printOpts(const Properties &opts) {
    for( Properties::const_iterator p = opts.begin(); p != opts.end(); p++ )
        std::cout << *p << std::endl;
}

int main() {
	int a = 5;
	properties prop;

	prop.push_back( OProperty("a", a) );
	prop.push_back( OProperty("b", std::string("boer")) );

	for( properties::const_iterator p = prop.begin(); p != prop.end(); p++ )
		std::cout << *p << std::endl;

	property_map prop2;

	prop2["a"] = a;
	prop2["b"] = std::string("boer");

	for( property_map::const_iterator p = prop2.begin(); p != prop2.end(); p++ )
		std::cout << *p << std::endl;

    Properties prop3;
    prop3.Set("a",a);
    prop3.Set("b",std::string("boer"));
    printOpts( prop3 );
    Properties prop4 = prop3;
    printOpts( prop4("c",a) );
    printOpts( Properties()("a",a)("b",std::string("boer")) );

	return 0;
}
