/*  Copyright (C) 2006,2007  Joris Mooij  [joris at jorismooij dot nl]
    Copyright (C) 2002  Martijn Leisink  <martijn@mbfys.kun.nl>
    Radboud University Nijmegen, The Netherlands
    
    This file is part of libDAI.

    libDAI is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    libDAI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with libDAI; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


#ifndef __defined_libdai_boundprop_h
#define __defined_libdai_boundprop_h

#define USE_LPSOLVE

#ifdef USE_LPSOLVE
#ifdef LPSOLVE_NEW
#include <lpsolve/lp_lib.h>
#else
#include "../../../../code/bound_propagation_1.0/lp_solve_3.2/lpkit.h"
#endif
#endif


#ifdef USE_GLPK
extern "C" { 
    #include "glpk.h"
}
#endif


#include "box.h"
#include "factor.h"


namespace libDAI {


    typedef std::set<Factor> Cluster;

    class BoundSet: public std::set<Box> {
        private:
#ifdef USE_LPSOLVE
            ::lprec *lp;
#endif
#ifdef USE_GLPK
            ::LPX *lpx;
#endif

        public:
            BoundSet() {};
            BoundSet( const std::set<Box> &bs ) : std::set<Box>(bs) {};
            BoundSet( const Cluster& C ) {
                for( Cluster::const_iterator i = C.begin(); i != C.end(); i++ )
                    insert( Box( i->vars() ) );
            };
            void make_lp( const VarSet& sep );
            void delete_lp() { 
#ifdef USE_LPSOLVE
                ::delete_lp(lp); 
#endif
#ifdef USE_GLPK
                ::lpx_delete_prob(lpx);
#endif
            };
            void newbound( Box& b, const Factor& Pjoint, const Factor& Psep ) const;
            BoundSet up( const FactorGraph& fg );
            BoundSet up( const FactorGraph& fg, const Cluster& base );
            size_t Verbose() const { return 0; }
    };

    void BoundProp_solve( const FactorGraph &net, long maxsize, double tol, size_t maxiter, BoundSet &bounds );
    void BoundProp_setup( const FactorGraph &net, long maxsize, BoundSet &B, Cluster &Bmain );
    void BoundProp_tile( const FactorGraph &net, long maxsize, Cluster &Cex, Cluster &Cin, size_t Verbose = 0 );


}


#endif
