/*  Copyright (C) 2006,2007  Joris Mooij  [joris at jorismooij dot nl]
    Radboud University Nijmegen, The Netherlands
    
    This file is part of libDAI.

    libDAI is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    libDAI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with libDAI; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


#include <iostream>
#include "alldai.h"
#include "box.h"

#include "boundprop.h"



using namespace std;
using namespace libDAI;


double min( vector<double> & x ) {
    return *min_element( x.begin(), x.end() );
}
double max( vector<double> & x ) {
    return *max_element( x.begin(), x.end() );
}
double avg( vector<double> & x ) {
    double sum = 0.0;
    for( size_t i = 0; i < x.size(); i++ )
        sum += x[i];
    return sum / x.size();
}
double stddev( vector<double> & x ) {
    double sumsq = 0.0;
    for( size_t i = 0; i < x.size(); i++ )
        sumsq += x[i] * x[i];
    double _avg = avg( x );
    double result = sumsq / x.size() - _avg * _avg;
    if( result < 0.0 )
        return 0.0;
    else
        return sqrt( result );
}





int main( int argc, char *argv[] ) {
    if ( argc != 5 ) {
        cout << "Usage: " << argv[0] << " <filename.fg> <fast> <brief> <tol>" << endl << endl;
        cout << "Reads factor graph <filename.fg> and runs" << endl;
        cout << "BoundPropagation on it." << endl << endl;
        return 1;
    } else {
        char *filename = argv[1];
        bool fast = atoi( argv[2] );
        bool brief = atoi( argv[3] );
        double tol = atof( argv[4] );
        size_t maxiter = 10000;

        double time = toc();

        FactorGraph fg;
        fg.ReadFromFile( filename );

//        Network net;
//        fg2net( fg, net );

        size_t maxcl = 1;
        for( size_t i = 0; i < fg.nrVars(); i++ ) {
            VarSet Di = fg.Delta( fg.var( i ) );
            if( Di.stateSpace() > maxcl )
                maxcl = Di.stateSpace();
        }

        BoundSet bounds;
        BoundProp_solve( fg, fast ? 0 : maxcl, tol, maxiter, bounds );

        vector<double> gaps( fg.nrVars(), 1.0 );
        for( size_t i = 0; i < fg.nrVars(); i++ ) {
            BoundSet::iterator b = find( bounds.begin(), bounds.end(), Box(fg.var(i)) );
            if( b != bounds.end() ) {
                gaps[i] = b->gap();
            } else {
                cerr << "Not found" << endl;
                return 1;
            }
        }

        time = toc() - time;

        if( brief )
            cout << "      " << min(gaps) << " " << max(gaps) << " " << avg(gaps) << " " << stddev(gaps) << " " <<  time << endl;
        else {
            for( size_t i = 0; i < fg.nrVars(); i++ )
                printf( "7 BoundProp %d %.8f\n", i, gaps[i] );
//                cout << "7 BoundProp " << i << " " << gaps[i] << " " << time << endl;
        }
    }

    return 0;
}
