/*  Copyright (C) 2006,2007  Joris Mooij  [joris at jorismooij dot nl]
    Radboud University Nijmegen, The Netherlands
    
    This file is part of libDAI.

    libDAI is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    libDAI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with libDAI; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


#ifndef __defined_libdai_bounds_other_h
#define __defined_libdai_bounds_other_h


#include <boost/numeric/ublas/matrix_sparse.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include "factorgraph.h"
#include <vector>
#include <string>
#include "boundprop/boundprop.h"


namespace libDAI {


    typedef boost::numeric::ublas::compressed_matrix<double> matrix;
    typedef boost::numeric::ublas::vector<double> ublasvector;

    class BinaryPairwiseGM {
        public:
            size_t N;
            matrix w;
            ublasvector th;
            double logZ0;

            BinaryPairwiseGM() {}
            BinaryPairwiseGM( const FactorGraph &fg );
            BinaryPairwiseGM( size_t _N, const matrix &_w, const ublasvector &_th, double _logZ0 ) : N(_N), w(_w), th(_th), logZ0(_logZ0) {}
            BinaryPairwiseGM( const BinaryPairwiseGM &x ) : N(x.N), w(x.w), th(x.th), logZ0(x.logZ0) {};
            BinaryPairwiseGM & operator=( const BinaryPairwiseGM &x ) {
                if( this != &x ) {
                    N  = x.N;
                    w  = x.w;
                    th = x.th;
                    logZ0 = x.logZ0;
                }
                return *this;
            }
            double doBP( size_t maxiter, double tol, size_t verbose, ublasvector &m );
    };

    double BoundJaakkolaJordanUpper( const BinaryPairwiseGM &net, ublasvector &mu );
    double BoundJaakkolaJordanLower( const BinaryPairwiseGM &net, ublasvector &mu );
    double OptimizeBound( const BinaryPairwiseGM &net, ublasvector &mu, std::string method );
    double BoundLeisink3Lower( const BinaryPairwiseGM &net, ublasvector &mu );
    double CalcDobrushinBinary( const BinaryPairwiseGM &net, matrix &C );
    void   findLoopClustersPairwise( const FactorGraph & fg, std::set<VarSet> &allcl, VarSet newcl, const Var & root, size_t length, VarSet vars );
    size_t countLoopsPairwise( const FactorGraph & fg, size_t loopdepth );
    double CalcGirthPairwise( const FactorGraph &fg );
    double BoundTatikonda( double c, double M );
    double BoundTagaMase( const matrix &C, size_t i );
    double BoundTagaMaseImproved( const FactorGraph &fg, const matrix &C, size_t i );
    double BoundlogZ( const FactorGraph &fg, std::string method, bool debug_bounds, size_t verbose );
    void   BoundTRP( const FactorGraph &fg, std::vector<Box> &results );

}


#endif
