/*  Copyright (C) 2006,2007  Joris Mooij  [joris at jorismooij dot nl]
    Radboud University Nijmegen, The Netherlands
    
    This file is part of libDAI.

    libDAI is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    libDAI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with libDAI; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


#include <iostream>
#include "alldai.h"


using namespace std;
using namespace libDAI;


int main( int argc, char *argv[] ) {
    if ( argc != 2 ) {
        cout << "Usage: " << argv[0] << " <filename.fg>" << endl << endl;
        cout << "Reads factor graph <filename.fg> and compares" << endl;
        cout << "several methods of calculating Z on clamped versions." << endl << endl;
        return 1;
    } else {
        FactorGraph fg;
        fg.ReadFromFile(argv[1]);

        size_t  maxiter = 10000;
        double  tol = 1e-9;
        size_t  verb = 1;


        Properties opts;
        opts.Set("maxiter",maxiter);
        opts.Set("tol",tol);
        opts.Set("verbose",verb);

        JTree jt( fg, opts("updates",string("HUGIN")) );
        jt.init();
        jt.run();

        BP bp(fg, opts("updates",string("SEQFIX")));
        bp.init();
        bp.run();

        TRW trw(fg, opts("updates",string("SEQFIX"))("outertol",1e-9)("maxouteriter",0UL));
        trw.init();
        trw.runTRW();


        cout << "Exact log partition sum: " << real( jt.logZ() ) << endl;
        cout << "LBP log partition sum: " << real( bp.logZ() ) << endl;
        cout << "TRW log partition sum: " << real( trw.logZ() ) << endl;


        opts.Set("verbose",0UL);
        double Z_ex = exp(real(jt.logZ()));
        MF mf( fg, opts );
        mf.init();
        mf.run();
        double Z_mf = exp(real(mf.logZ()));
        TRW trw_(fg, opts("updates",string("SEQFIX"))("outertol",1e-9)("maxouteriter",0UL));
        trw_.init();
        trw_.run();
        double Z_trw = exp(real(trw_.logZ()));
        cout << endl << endl;
        cout << "\t" << Z_ex << "\t" << Z_mf << "\t" << Z_mf - Z_ex << "\t" << Z_trw << "\t" << Z_trw - Z_ex << endl << endl;
        for( size_t i = 0; i < fg.nrVars(); i++ ) {
            Var v_i = fg.var(i);

            vector<double> Z_ex_i(2,0.0);
            vector<double> Z_mf_i(2,0.0);
            vector<double> Z_trw_i(2,0.0);

            for( size_t x_i = 0; x_i < 2; x_i++ ) {
                FactorGraph clamped = fg.clamped( v_i, x_i );

                JTree jtcl( clamped, opts("updates",string("HUGIN")) );
                jtcl.init();
                jtcl.run();
                Z_ex_i[x_i] = exp(real(jtcl.logZ()));

                MF mfcl( clamped, opts );
                mfcl.init();
                mfcl.run();
                Z_mf_i[x_i] = exp(real(mfcl.logZ()));

                TRW trwcl(clamped, opts("updates",string("SEQFIX"))("outertol",1e-9)("maxouteriter",0UL));
                trwcl.init();
                trwcl.run();
                Z_trw_i[x_i] = exp(real(trwcl.logZ()));
            }

            cout << i << "\t" << (Z_ex_i[0] + Z_ex_i[1]) << "\t" << (Z_mf_i[0] + Z_mf_i[1]) << "\t" << (Z_mf_i[0] + Z_mf_i[1]) - (Z_ex_i[0] + Z_ex_i[1]) << "\t" << (Z_trw_i[0] + Z_trw_i[1]) << "\t" << (Z_trw_i[0] + Z_trw_i[1]) - (Z_ex_i[0] + Z_ex_i[1]) << endl;
        }
    }

    return 0;
}
