/*  Copyright (C) 2006,2007  Joris Mooij  [joris at jorismooij dot nl]
    Radboud University Nijmegen, The Netherlands
    
    This file is part of libDAI.

    libDAI is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    libDAI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with libDAI; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/


#include <iostream>
#include <stdlib.h>
#include <algorithm>
#include "bounds_other.h"
#include "alldai.h"
#include "box.h"
#include "bounds.h"
#include "diffs.h"
#include "cmath"
#include <boost/numeric/ublas/matrix_sparse.hpp>
#include <boost/numeric/ublas/matrix_proxy.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/io.hpp>

#include <gsl/gsl_errno.h>
#include <gsl/gsl_math.h>
#include <gsl/gsl_multimin.h>


using namespace std;
using namespace libDAI;


int main( int argc, char *argv[] ) {
    if ( argc != 5 ) {
        cout << "Usage: " << argv[0] << " <filename.fg> <method> <debug> <verbose>" << endl << endl;
        cout << "Reads factor graph <filename.fg> and calculates (bounds on) logZ." << endl;
        cout << "Method can be EXACT,MF,LK,LKOPT,TRW,JJ,JJOPT,BP." << endl;
        return 1;
    } else {
        FactorGraph fg;
        char *infile = argv[1];
        string method = argv[2];
        bool debug = atoi(argv[3]);
        size_t verbose = atoi(argv[4]);

        fg.ReadFromFile( infile );

        double time = 0;
        cout.precision(10);

        time = toc();
        double logZ = BoundlogZ( fg, method, debug, verbose );
        time = toc() - time;
        cout << logZ << " " << time << endl;

/*          ATTEMPT FOR IMPROVED TAGA-MASE BOUND
        size_t N = fg.nrVars();
        BinaryPairwiseGM net( fg );
        matrix C_Dobrushin;
        double c = CalcDobrushinBinary( net, C_Dobrushin );
        vector<double> a(N, 1.0);
        size_t root = 0;
        for( size_t iter = 0; iter < 10; iter++ ) {
            for( size_t i = 0; i < N; i++ ) {
                double c_i = 0.0;
//                    if( i != root )
//                        c_i += 1.0;
                for( size_t j = 0; j < N; j++ ) {
                    c_i += C_Dobrushin(i,j) * a[j];
                }
                if( c_i < a[i] )
                    a[i] = c_i;
                cout << i << " " << c_i << " " << a[i] << endl;
            }
            cout << endl;
        }
*/
        return 0;
    }
}
