# need a gnuplot 4.2 or higher (~/bin/gnuplot)
set terminal postscript eps enhanced monochrome


set size 0.8,0.8
unset ytics
set ytics
#set xtics auto
#set ytics rotate
set logscale x
unset logscale y
set xrange [0.5e-9:2]
set xlabel 'Gaps'
set ylabel ''
set key off
set yrange [0:13]

# Gaps for rectangular, binary Ising grid

set output 'eps/bounds_1E-2.eps'
set title '{/Symbol b} = 10^{-2}'
plot 'dat/bounds_1E-2.dat' u ($4+1e-9):1:yticlabel(2) pt 5

set output 'eps/bounds_1E-1.eps'
set title '{/Symbol b} = 10^{-1}'
plot 'dat/bounds_1E-1.dat' u ($4+1e-9):1:yticlabel(2) pt 5

set output 'eps/bounds_1E0.eps'
set title '{/Symbol b} = 10^{0}'
plot 'dat/bounds_1E0.dat' u ($4+1e-9):1:yticlabel(2) pt 5

set output 'eps/bounds_1E1.eps'
set title '{/Symbol b} = 10^{1}'
plot 'dat/bounds_1E1.dat' u ($4+1e-9):1:yticlabel(2) pt 5

# Gaps for rectangular, ternary Ising grid

unset ytics
set yrange [0:8]
set ytics

set output 'eps/bounds3_1E-2.eps'
set title '{/Symbol b} = 10^{-2}'
plot 'dat/bounds3_1E-2.dat' u ($4+1e-9):1:yticlabel(2) pt 5

set output 'eps/bounds3_1E-1.eps'
set title '{/Symbol b} = 10^{-1}'
plot 'dat/bounds3_1E-1.dat' u ($4+1e-9):1:yticlabel(2) pt 5

set output 'eps/bounds3_1E0.eps'
set title '{/Symbol b} = 10^{0}'
plot 'dat/bounds3_1E0.dat' u ($4+1e-9):1:yticlabel(2) pt 5

set output 'eps/bounds3_1E1.eps'
set title '{/Symbol b} = 10^{1}'
plot 'dat/bounds3_1E1.dat' u ($4+1e-9):1:yticlabel(2) pt 5

# Gaps alarm

unset ytics
set yrange [0:5]
set ytics

set output 'eps/alarm.eps'
set title 'ALARM'
plot 'dat/alarm.dat' u ($4+1e-9):1:yticlabel(2) pt 5

# Gaps for PROMEDAS

set yrange [0:10]
unset ytics
set ytics auto
set ylabel 'instance'

set output 'eps/promedas_bp.eps'
set title 'Belief Propagation'
plot 'dat/promedas_bp.dat' u ($5+1e-9):1 pt 5

set output 'eps/promedas_cav.eps'
set title 'BoxProp-SubT'
plot 'dat/promedas_cav.dat' u ($5+1e-9):1 pt 5

set output 'eps/promedas_saw.eps'
set title 'BoxProp-SAWT'
plot 'dat/promedas_saw.dat' u ($5+1e-9):1 pt 5

set output 'eps/promedas_boundprop.eps'
set title 'BoundProp'
plot 'dat/promedas_boundprop.dat' u ($5+1e-9):1 pt 5

# Computation times for rectangular, binary Ising grid

set size 1.4,0.7
unset xtics
set xtics rotate
set ytics auto
unset logscale x
set logscale y
set xrange [0.4:13]
set yrange [0.5e-2:*]
set xlabel ''
set ylabel 'Computation time (s)'
set title ''
set key left box height 1
set boxwidth 0.2
set output 'eps/bounds_time.eps'
plot 'dat/bounds_1E-2.dat' u ($1):(($5+1)/100.0) every 25 with boxes fill solid 0 lt 1 title '{/Symbol b} = 10^{-2}', \
     'dat/bounds_1E-1.dat' u ($1+0.2):(($5+1)/100.0):xticlabel(2) every 25 with boxes fill solid 0.3 lt 1 title '{/Symbol b} = 10^{-1}', \
     'dat/bounds_1E0.dat'  u ($1+0.4):(($5+1)/100.0) every 25 with boxes fill solid 0.6 lt 1 title '{/Symbol b} = 10^{0}', \
     'dat/bounds_1E1.dat'  u ($1+0.6):(($5+1)/100.0) every 25 with boxes fill solid 1 lt 1 title '{/Symbol b} = 10^{1}'

# Computation times for rectangular, ternary Ising grid

set size 0.9,0.7
unset xtics
set xtics rotate
set xrange [0.4:8.0]
set output 'eps/bounds3_time.eps'
plot 'dat/bounds3_1E-2.dat' u ($1):(($5+1)/100.0) every 25 with boxes fill solid 0 lt 1 title '{/Symbol b} = 10^{-2}', \
     'dat/bounds3_1E-1.dat' u ($1+0.2):(($5+1)/100.0):xticlabel(2) every 25 with boxes fill solid 0.3 lt 1 title '{/Symbol b} = 10^{-1}', \
     'dat/bounds3_1E0.dat'  u ($1+0.4):(($5+1)/100.0) every 25 with boxes fill solid 0.6 lt 1 title '{/Symbol b} = 10^{0}', \
     'dat/bounds3_1E1.dat'  u ($1+0.6):(($5+1)/100.0) every 25 with boxes fill solid 1 lt 1 title '{/Symbol b} = 10^{1}'

# Computation times for ALARM
set size 0.5,0.7
unset xtics
set xtics rotate
set xrange [0.0:5.0]
set boxwidth 0.8
set title 'ALARM'
set key off
set output 'eps/alarm_time.eps'
plot 'dat/alarm.dat' u ($1):(($5+1)/100.0):xticlabel(2) every 25 with boxes fill solid 0.5 lt 1

# Scatter plots
set xtics auto
set ytics auto
set xrange [0.5e-10:2]
set yrange [0.5e-10:2]
set logscale x
set logscale y
set ylabel 'BoxProp-SAWT'
set title ''
set xlabel 'BoundProp'
set output 'eps/bounds_1E1_BoundProp.eps'
plot 'dat/bounds_1E1_BoundProp.dat' u ($1+1e-9):($2+1e-9) pt 5, x

set xlabel 'MF-TRW'
set output 'eps/bounds_1E1_MF-TRW.eps'
plot 'dat/bounds_1E1_MF-TRW.dat' u ($1+1e-9):($2+1e-9) pt 5, x

set xlabel 'LK-TRW'
set output 'eps/bounds_1E1_LK-TRW.eps'
plot 'dat/bounds_1E1_LK-TRW.dat' u ($1+1e-9):($2+1e-9) pt 5, x

exit
