//
// loopgraph.cc
//  
// Made by Vicen Gomez
// Login   <vicen@localhost.localdomain>
// 
// Started on  Tue May 16 11:04:13 2006 Vicen Gomez
//

#include "loopgraph.hh"

int vpow2[7] = {1, 2, 4, 8, 16, 32, 48};
////////////////////////////////
// static member initializations

using namespace std;

BGraph LoopGraph::gcore = '\0';
Mu LoopGraph::m_mus = Mu();

////////////////////
// LoopGraph methods

LoopGraph::~LoopGraph() {}


// writes the loop (assumes the representation used in the combinatiorial search)
void LoopGraph::loop2Dot(const BitSet &loop)
{
  stringstream svloop, seloop;
	pair<VertexIt, VertexIt> vit = vertices(gcore);
	while (vit.first != vit.second) {
		string str_type = (gcore[*vit.first].type == FACTOR) ? "shape=box," : "shape=circle,";
		if (loop.test(*vit.first)) svloop << "\t" << *vit.first << "[label=\"" << *vit.first << "\"," << str_type  << "fillcolor=black];" << std::endl;
		else svloop << "\t" << *vit.first << "[label=\"" << *vit.first << "\"," << str_type << "fillcolor=gray];" << std::endl;
    OutEdgeIt b_begin, b_end;
    for (tie(b_begin, b_end) = out_edges(*vit.first, gcore); b_begin != b_end; b_begin++) {
		//	cerr << m_node_pos[*vit.first] << "+" << j << "!=" << gcore[*b_begin].pos_factor << endl;
			Vertex u = target(*b_begin, gcore);
			if (u > *vit.first) {
  			if ( loop.test(u) && loop.test(*vit.first) ) seloop << "\t" << *vit.first << " -- " << u << "[color=black,weight=15];" << std::endl;
				else seloop << "\t" << *vit.first << " -- " << u << "[color=grey,weight=10];" << std::endl;
			}
		}
    vit.first++;
	} 

	string str_loop;
	to_string(loop, str_loop);
	str_loop += ".dot";
	ofstream os(str_loop.c_str());
	os << "graph G {" << std::endl;
	os << "graph[size=\"100,100\"];" << std::endl;
			  
  os << "node[shape=circle,width=0.2,height=0.2,fixedsize=true,fontsize=0,style=filled,fillcolor=gray];" << std::endl;
  //os << "node[shape=circle,width=0.2,height=0.2,fixedsize=true,fontsize=0,style=filled,fillcolor=gray];" << std::endl;
	os << svloop.rdbuf()->str();
	os << "edge[];" << std::endl;
	os << seloop.rdbuf()->str();
	os << "}" << std::endl;
	os.close();
}

void LoopGraph::preprocessGraph()
{
  // remove nodes with degree one 
  if (m_verbosity > 1) cerr << "preprocessing graph ...";
  std::pair<VertexIt, VertexIt> it = vertices(gcore);

  uint nvertices = num_vertices(gcore);
  uint rem = 0, remp;
  do {
    remp = rem;
    std::pair<VertexIt, VertexIt> it = vertices(gcore);
    while ( (it.first != it.second) && (rem == remp) ) {
      if (degree(*it.first, gcore) == 1) {
        if (gcore[*it.first].type == FACTOR) m_nfactors--;
          clear_vertex(*it.first, gcore);

        // remove_vertex invalidates iterator
       	remove_vertex(*it.first, gcore);
       	rem++;
      }
      it.first++;
    }
  } while (rem - remp);

	// rebuild indices and beliefs of the factors
	// NOTE that if other members of PVertex or PEdge are used they also should be rebuilt
	m_nfactors = 0;
	VertexIt vi, ve;
	for (tie(vi, ve) = vertices(gcore); vi != ve; vi++) {
		Vertex a = *vi;
		set<int> vpos;
		if (gcore[a].type == FACTOR) {
			m_nfactors++;
	     PIndex idx = (gcore)[a].index;
			if (degree(a, gcore) != (uint)idx.getNVars()) {
				PIndex new_index;
				OutEdgeIt ab, ae;
				int new_pos = 0;
			  for (tie(ab, ae) = out_edges(a, gcore); ab != ae; ab++) {
					new_index.addVar(2);
					vpos.insert(gcore[*ab].pos_index);
					gcore[*ab].pos_index = new_pos;
					new_pos++;
				}
				VProb bel(gcore[a].bel), new_bel(1ul << vpos.size(), 0.);
	
				// traverse the old belief by traversing the old index
				// marginalize over the removed vars
				uint p = 0;
				for (vector<vector<uint> >::iterator it = idx.begin(); it != idx.end(); it++) {
					uint bpos = 0, j = 0;
					for (uint i = 0; i < (*it).size(); i++) {
						bool deleted = (vpos.find(i) == vpos.end());
						if ((*it)[i] && !deleted) {
							bpos += (1ul << j);
						}
						if (!deleted) j++;
					}
					new_bel[bpos] += bel[p];
					p++;
				}
				gcore[a].index = new_index;
				gcore[a].bel = new_bel;
			}
		}
	}		  

	
  // set the proper pos_factor in the edges
  std::pair<VertexIt, VertexIt> vpv = vertices(gcore); vpv.first += m_nfactors;
  while (vpv.first != vpv.second) {
	  OutEdgeIt b_begin, b_end;
		int posf = 0;
    for (tie(b_begin, b_end) = out_edges(*vpv.first, (gcore)); b_begin != b_end; b_begin++) {
      (gcore)[*b_begin].pos_factor = posf;
			posf++;
    }
    vpv.first++;
  }

  if (m_verbosity) {
		cerr << " " << rem << "/" << nvertices;
		cerr << " vertices removed during preprocessing" << std::endl;
	}
}

void LoopGraph::buildCoreMap()
{
  m_g2core.resize(num_vertices(*g), -1);
  m_core2g.resize(num_vertices(gcore), -1);
  std::pair<VertexIt, VertexIt> itc = vertices(gcore); 

  // forall vertices in the 2-core
  while (itc.first != itc.second) {
    PVertex vc = gcore[*itc.first];
    bool found = false;
    std::pair<VertexIt, VertexIt> itg = vertices(*g);

    // forall vertices in the original graph
    while ( (itg.first != itg.second) && !found ) {
      PVertex vg = (*g)[*itg.first];
      if ( (vg.type == vc.type) && (vg.label == vc.label) ) {
       found = true;
       m_g2core[*itg.first] = *itc.first;
       m_core2g[*itc.first] = *itg.first;
      }
      itg.first++;
    }

    if (!found) throw "error building vertices correspondences...";
    itc.first++;
  }
}

void LoopGraph::computeMusForneyGraph()
{
	if (m_verbosity > 2) cerr << "Precomputing mus" << endl;
 	m_mus.resize( num_vertices(m_forney) );
	FVertexIt vi, ve;
  for (tie(vi, ve) = vertices(m_forney); vi != ve; vi++) {

		FVertexD a = *vi;
    PIndex *idx = &(m_forney)[a].index;
		vector<Real> vmagn, vsqrt;
		FInEdgeIt i_begin, i_end;
		for (tie(i_begin, i_end) = in_edges(a, m_forney); i_begin != i_end; i_begin++) {

			// store magnetizations and denominators of the mus
			Real m(m_forney[*i_begin].bel[0] - m_forney[*i_begin].bel[1]);
			vmagn.push_back( m );
			vsqrt.push_back( Sqrt( Real(1.) - m*m ) );
		}

		// forall combinations of variables in factor a
		// we traverse the index...
		m_mus[a].resize((*idx).size(), Real(0.));
		for (vector<vector<uint> >::iterator it = (*idx).begin(); it != (*idx).end(); it++) {

			// forall variables in a particular combination
			VProb bela(m_forney[a].bel);
			int pos = 0, nvars = 0;
			for (uint i = 0; i < (*it).size(); i++) {

				// if variable appears in that combination
        if ((*it)[i]) {

					nvars++;
					pos += (1ul<<i);

					// For general interactions set here the permutation variables
					// traverse all the factor and multiply with (x_i - m_i) 
          for (uint itf = 0; itf < bela.size(); itf++) {
            bela[itf] *= (idx->get(itf, i) == 0) ? (Real(+1.) - vmagn[i]) : (Real(-1.) - vmagn[i]);
            bela[itf] /= vsqrt[i]; 
					}
				}
			} // end for one particular combination of vars

			// only if two or more vars
			Real term = 0.;
			if (nvars > 1) {
				for (uint itf = 0; itf < bela.size(); itf++) 
					term += bela[itf];
				m_mus[a][pos] = term;
			}
		} // end forall values of the index
	}

	if (m_verbosity > 1) {
		ofstream fmus("mus_forney.txt");
		fmus << scientific << setprecision(20);
		fmus << "MUS : " << endl;
		for (uint mi = 0; mi < m_mus.size(); mi++) {
		  for (uint mj = 0; mj < m_mus[mi].size(); mj++) 
        fmus << m_mus[mi][mj] << " ";
			fmus << endl;
		}
		fmus.close();
	}

}

void LoopGraph::computeMusFactorGraph()
{
	if (m_verbosity > 2) cerr << "Precomputing mus" << endl;

	// forall vertices in the 2-core
	// one term for each var and for each possible degree
 	m_mus.resize( num_vertices(gcore) );
	m_sfactors.resize( num_vertices(gcore) );
	std::pair<VertexIt, VertexIt> itc = vertices(gcore);
  while (itc.first != itc.second) {
		if (gcore[*itc.first].type == VAR) {
		  if (m_verbosity > 3) cerr << "var term" << endl;

      // precompute variable terms
      LVertex i = *itc.first;
			Real m = gcore[i].bel[0] - gcore[i].bel[1];
      m_mus[i].resize(1+degree(i, gcore), 1.);
      for (uint p = 2; p <= degree(i, gcore); p++) {
 	      Real a = 1., b = 1., c = 1., m2 = m*m;
        for (uint d = 0; d < (p-1); d++) {
          a *= (Real(1.) - m);
          b *= (Real(-1.) -m);
          c *= (Real(1.) - m2);
				}
 				b *= Real(-1.);
        m_mus[i][p] = (a + b)/(Real(2.)*c);
			}
		} // end if var

		itc.first++;
	}

	// one term for each factors and for each possible config of two or more vars
  itc = vertices(gcore);
  OutEdgeIt j_begin, j_end;
	while (itc.first != itc.second) {
	  if (gcore[*itc.first].type == FACTOR) {
		  if (m_verbosity > 3) cerr << "factor term" << endl;

  		// precompute factor terms
      LVertex a = *itc.first;
      PIndex *idx = &(gcore)[a].index;
			vector<LVertex> vneigh, vpos_index, vpos_factor;
      for (tie(j_begin, j_end) = out_edges(a, gcore); j_begin != j_end; j_begin++) {
				vneigh.push_back( target(*j_begin, gcore) );              // vneigh contains the adjacent vars
				vpos_index.push_back( gcore[*j_begin].pos_index );        // contains the pos of the vars in the index of the factor 
				vpos_factor.push_back( gcore[*j_begin].pos_factor );      // contains the pos of the factor for the adj vars
			}

			// forall combinations of variables in factor a
			int p = 0;
			m_mus[a].resize((*idx).size());
			for (vector<vector<uint> >::iterator it = (*idx).begin(); it != (*idx).end(); it++) {

				// forall variables in a particular combination
				VProb bela(gcore[a].bel);
				for (uint i = 0; i < (*it).size(); i++) {

					// if variable appears in that combination
          if ((*it)[i]) {

						// For general interactios set here the permutation variables

						// traverse all the factor and multiply with (x_i - m_i) 
            for (uint itf = 0; itf < bela.size(); itf++) {
							Real m = gcore[vneigh[i]].m;
	            bela[itf] *= (idx->get(itf, vpos_index[i]) == 0) ? (Real(+1.) - m) : (Real(-1.) - m);
						}
					}
				} // end for one particular combination of vars

				Real term = 0.;
        for (uint itf = 0; itf < bela.size(); itf++) 
          term += bela[itf];
				m_mus[a][p] = term;

				// NOTE that this imples more than 2 bits could be marked for general interactions
  			p++;
			} // end forall index values

			// for pairwise interactions there is only one pfactor for each factor

      // traverse the other pfactors and set them as neighbors
			BitSet padjacent(m_nfactors, 0), psubgraph(num_vertices(gcore), 0);
			for (uint pf = 0; pf < m_pfactors.size(); pf++) {
				if ( (m_pfactors[pf].i == vneigh[0]) || (m_pfactors[pf].i == vneigh[1]) ||
						 (m_pfactors[pf].j == vneigh[0]) || (m_pfactors[pf].j == vneigh[1]) )
				{
					// pfactors
					m_pfactors[pf].adjacent.set( m_pfactors.size() );	// new pfactor ~> previous
          padjacent.set(pf);                          			// previous ~> new pfactor
				}
			}

			// pfactor
			if (m_verbosity > 3) cerr << "before " <<endl;
			psubgraph.set(a); psubgraph.set(vneigh[0]);	psubgraph.set(vneigh[1]);
	  	PFactor phi( vneigh[0], vneigh[1], m_mus[a][3], psubgraph, padjacent );
	    m_pfactors.push_back( phi );
			m_mus[a][0] = 1.;

			// sfactor
      uint vi = m_node_pos[vneigh[0]]+vpos_factor[0], vj = m_node_pos[vneigh[1]]+vpos_factor[1];
			uint vk = m_node_pos[a], vl = m_node_pos[a]+1; 
			m_sfactors[a] = SFactor( vi, vj, vk, vl ); 
			if (m_verbosity > 3) cerr << "before22 " <<endl;

			// IMPORTANT: For factors the second index is not the degree
			//            but for pairwise interactions we make an ex
			m_mus[a][2] = m_mus[a][3];
		} // end if factor

		itc.first++;
	} // end forall factors

	if (m_verbosity > 1) {
		ofstream fmus("mus.txt");
		fmus << scientific << setprecision(20);
		fmus << "MUS : " << endl;
		for (uint mi = 0; mi < m_mus.size(); mi++) {
			if (gcore[mi].type == VAR) fmus << "VAR\t";
			else fmus << "FACT\t";
		  for (uint mj = 0; mj < m_mus[mi].size(); mj++) {
        fmus << m_mus[mi][mj] << " ";
			}
			fmus << endl;
		}
		fmus << "PFACTORS : " << endl;
  	for (uint pf = 0; pf < m_pfactors.size(); pf++) fmus << m_pfactors[pf] << endl;	

		fmus << "SFACTORS : " << endl;
  	for (uint pf = 0; pf < m_sfactors.size(); pf++) fmus << pf << " " << m_sfactors[pf];	
		fmus << endl;
		fmus.close();
	}
}

bool LoopGraph::doLoopSeries(bool output)
{
  m_logZ_tlsbp = Real( m_logZ_bp );
	m_sumR = Real(1.);
	m_sum_single = Real(1.);
  m_error_bp = getBPError();

	// lets output the cumulated error of the pfaffian here ...
	if (m_verbosity > 1) {
		char buff[400];
		sprintf(buff, "cum_err_pfaffian_%s.m", m_filename.c_str());
		ofstream fcum(buff);
		Real acum(0.), logZ_pfaffian = m_logZ_bp_forney;
		fcum << setprecision(15) << Abs(m_logZ_exact - logZ_pfaffian) << endl;
		for (uint p = 0; p < m_contrib_pfaffian.size(); p++) {
			acum += m_contrib_pfaffian[p].first;
			logZ_pfaffian = m_logZ_bp_forney + Log(acum);
			fcum << setprecision(15) << Abs(m_logZ_exact - logZ_pfaffian) << endl;
		}
		fcum.close();
	}

	// -------------------------

	if ( m_verbosity ) {
		cerr << scientific << setprecision(14) << "m_ZPfaffian " << m_ZPfaffian << endl;
		if (Abs( m_ZPfaffian ) > Real(0)) {
		  cerr << "\tsqrt(det(P)) = " << m_ZPfaffian << endl;
			cerr << "\tsumR         = " << m_sumR << endl << endl;
		}

		cerr << "\tlogZ_exact   = " << m_logZ_exact << " " << expl(m_logZ_exact) << endl;
		cerr << "\tlogZ_bp      = " << m_logZ_bp << " " << expl(m_logZ_bp) << endl;

		if (Abs( m_ZPfaffian ) >  Real(0)) {
			cerr << "\tlogZ_single  = " << getLogZSingle() << " " << Exp( getLogZSingle() ) << endl;
			cerr << "\tlogZ_Pfaff   = " << getLogZPfaffian() << " " << Exp( getLogZPfaffian() ) << endl;
		}
		cerr << endl;
    cerr << "\terr_Z_bp     = " << Abs(m_logZ_bp - m_logZ_exact)/m_logZ_exact << endl;

		if (Abs( m_ZPfaffian ) > Real(0)) {
			cerr << "\terr_Z0_Pfaff = " << Abs(m_logZ_exact - getLogZPfaffian1())/m_logZ_exact << endl;
			cerr << "\terr_Z_Pfaff  = " << Abs(m_logZ_exact - getLogZPfaffian())/m_logZ_exact << endl;
		}
    cerr << endl;
	}
	else {

	}
	return true;
}

std::ostream& operator<< (std::ostream& os, const vector<BitSet>& sg)
{
	for (vector<BitSet>::const_iterator it = sg.begin(); it != sg.end(); it++)
		os << (*it) << " ";
	os << endl;
	return os;
}

std::ostream& operator<< (std::ostream& os, const LoopGraph::PFactor& p)
{
	os << "vars  : " << p.i << " " << p.j << endl;
	os << "bits  : " << p.subgraph << " " << endl;
	os << "muf   : " << p.muf << endl;
	os << "padj  : " << p.adjacent << endl;
	os << endl;
	return os;
}

std::ostream& operator<< (std::ostream& os, const LoopGraph::SFactor& p)
{
	os << "bits  : " << p.i << " " << p.j << " " << p.k << " " << p.l << endl;
//	os << endl;
	return os;
}



std::ostream& operator<< (std::ostream& os, const LoopGraph::Disconnected& disc)
{
  os << setprecision(15);
  os << "subgra = " << disc.loop << endl;
  os << "miss   = " << disc.nmiss << endl;
  os << "term   = " << disc.term << endl;
  os << "nterms = " << disc.nterms << endl;
  os << "---------------------" << endl;
	return os;
}



//						if (uedges == the_loop) {
//							cerr << (*it1).second << endl;
//							cerr << (*it2).second << endl;
//						for (uint ii = 0; ii < (*it1).first.size(); ii++) {
//              if ((*it1).first.test(ii))
//								cerr << ii << "-";
//						}
//					  cerr << " " << (*it1).first.count() << endl;
//						for (uint ii = 0; ii < (*it2).first.size(); ii++) {
//              if ((*it2).first.test(ii))
//								cerr << ii << "-";
//						}
//						cerr << " " << (*it2).first.count() << endl;
//						}

//  while (it.first != it.second)  {
//		cerr << (*it.first) << endl;
//	  OutEdgeIt j_begin, j_end;
//    for (tie(j_begin, j_end) = out_edges(*it.first, gcore); j_begin != j_end; j_begin++) { 
//			cerr << *j_begin << " " << endl;
//    }
//		it.first++;
//	}

//		BitSet the_loop(num_edges(gcore));
//		the_loop.set(4); the_loop.set(5); the_loop.set(6); the_loop.set(7);
//		the_loop.set(8); the_loop.set(9); the_loop.set(14); the_loop.set(15);
//		the_loop.set(24); the_loop.set(25); the_loop.set(26); the_loop.set(27);
//		the_loop.set(28); the_loop.set(29); the_loop.set(32); the_loop.set(33);
//		the_loop.set(40); the_loop.set(41); the_loop.set(44); the_loop.set(45);


