#include <iostream>
#include <fstream>
#include <map>
#include <numeric>
#include <cmath>
#include <cstdlib>

#include "util.h"
#include "alldai.h"
#include "exceptions.h"
#include <boost/program_options.hpp>

using namespace std;
using namespace libDAI;
namespace po = boost::program_options;

class TestDAI {
protected:
	InferenceAlgorithm          *obj;
  string          name;
  vector<double>  err;

public:
  vector<Factor>  q;
  double          logZ;
  double          maxdiff;
  clock_t         time;
  size_t          iters;

  TestDAI( const FactorGraph &fg, const string &_name, const Properties &opts ) : obj(NULL), name(_name), err(), q(), logZ(0.0), maxdiff(0.0), time(0) {
      double tic = toc();
      if( name == "LDPC" ) {
          double zero[2] = {1.0, 0.0};
          q.clear();
          for( size_t i = 0; i < fg.nrVars(); i++ )
              q.push_back( Factor(Var(i,2), zero) );
          logZ = NAN;
          maxdiff = 0.0;
          iters = 1;
      } else
          obj = newInfAlg( name, fg, opts );

      time += toc() - tic;
  }


//  TestDAI( const FactorGraph &fg, const string &_name, const Properties &opts ) : obj(NULL), name(_name), err(), q(), logZ(0.0), maxdiff(0.0), time(0)	{
//    clock_t tic = toc();
//    obj = newInfAlg( name, fg, opts );
//    time += toc() - tic;
//	}

  ~TestDAI() { 
    if( obj != NULL )
        delete obj;
  }

	InferenceAlgorithm *getInfAlg() { return obj; }

  string identify() { 
    if( obj != NULL )
      return obj->identify(); 
    else
      return "NULL";
  }

  vector<Factor> allBeliefs(size_t limit) {
      vector<Factor> result;
      for( size_t i = 0; i < obj->grm().nrVars(); i++ ) {
          result.push_back( obj->belief( obj->grm().var(i) ) );
          if( result.size() == limit )
              break;
      }
      return result;
  }

//  vector<Factor> allBeliefs() {
//    vector<Factor> result;
//    for( size_t i = 0; i < obj->grm().nrVars(); i++ )
//      result.push_back( obj->belief( obj->grm().var(i) ) );
//    return result;
//  }

  void doDAI(size_t limit = 0) {
      double tic = toc();
      if( obj != NULL ) {
          obj->init();
          obj->run();
          time += toc() - tic;
          logZ = real(obj->logZ());
          maxdiff = obj->maxDiff();
          iters = obj->Iterations();
          q = allBeliefs(limit);
      };
  }

//  void doDAI() {
//   double tic = toc();
//   if( obj != NULL ) {
//       obj->init();
//       obj->run();
//       time += toc() - tic;
//       logZ = real(obj->logZ());
//       maxdiff = obj->maxDiff();
//       iters = obj->Iterations();
//       q = allBeliefs(limit);
//   };
// }

  void calcErrs( const TestDAI &x ) {
    err.clear();
    err.reserve( q.size() );
    for( size_t i = 0; i < q.size(); i++ )
      err.push_back( dist( q[i], x.q[i], Prob::DISTLINF ) );
    //err.push_back( dist( q[i], x.q[i], Prob::DISTTV ) );
  }

  void calcErrs( const vector<Factor> &x ) {
    err.clear();
    err.reserve( q.size() );
    for( size_t i = 0; i < q.size(); i++ )
      err.push_back( dist( q[i], x[i], Prob::DISTLINF ) );
    //err.push_back( dist( q[i], x[i], Prob::DISTTV ) );
  }

  double maxErr() { 
    return( *max_element( err.begin(), err.end() ) );
  }
        
  double avgErr() { 
    return( accumulate( err.begin(), err.end(), 0.0 ) / err.size() );
  }
};

pair<string, Properties> parseMethod( const string &_s, const map<string,string> & aliases ) {
  string s = _s;
  if( aliases.find(_s) != aliases.end() )
    s = aliases.find(_s)->second;

  pair<string, Properties> result;
  string & name = result.first;
  Properties & opts = result.second;

  string::size_type pos = s.find_first_of('[');
  name = s.substr( 0, pos );
  if( pos == string::npos )
    throw "Malformed method";
  size_t n = 0;
  for( ; n < sizeof(DAINames) / sizeof(string); n++ )
    if( name == DAINames[n] )
      break;
//  if( n == sizeof(DAINames[) / sizeof(string) ) {
//    cout << sizeof(DAINames) << " / " <<  sizeof(string) << endl;
//    throw "Unknown inference algorithm";
//	}

  stringstream ss;
  ss << s.substr(pos,s.length());
  ss >> opts;

  return result;
}

double clipdouble( double x, double minabs ) {
  if( fabs(x) < minabs )
    return minabs;
  else
    return x;
}


